package edu.toronto.cs.cxmlparser;

import java.io.*;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.clang.*;
import edu.toronto.cs.yasm.cfg.*;
import edu.toronto.cs.yasm.cfg.CFGraph.*;
import java.util.*;

import static edu.toronto.cs.expr.CTypeDeclOp.*;
import static edu.toronto.cs.expr.CTypeOp.*;
import static edu.toronto.cs.expr.CNamedDeclOp.*;
import static edu.toronto.cs.expr.CLabelDeclOp.*;
import static edu.toronto.cs.expr.CGlobalOp.*;
import static edu.toronto.cs.expr.CRecordOp.*;
import static edu.toronto.cs.expr.CFunctionDeclOp.*;
import static edu.toronto.cs.expr.CBasicBlockOp.*;
import static edu.toronto.cs.expr.CStmtOp.*;
import static edu.toronto.cs.expr.NumericOp.*;
import static java.lang.Integer.parseInt;
import static java.lang.Boolean.parseBoolean;

/**
 * A class to parse XML representation of C programs produced by GCC
 *
 *
 * Created: Fri Feb 16 16:17:32 2007
 *
 * @author <a href="mailto:arie@pcawf.sei.cmu.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public class CXMLParser extends DefaultHandler
{


  /** 
   * Factory to create expressions
   */
  ExprFactory efac;


  /**
   * A set of tag names that are skipped by the parser
   */
  Set<String> skippedTags;

  /**
   * A mapping between type ids (tuid) and types 
   */
  Map<Integer, Expr> typeTable; // tuid -> CType
  /**
   * A mapping between declaration ids (uid) and declarations 
   */
  Map<Integer, Expr> declTable; // uid -> CVarDecl
  /**
   * A mapping between function uid and its control flow graph.
   */
  Map<Integer, CFGraph> cfgTable; // uid -> CFGraph


  /**
   * Mapping between tag names and constant Operators
   * XXX only include tagName if waiting does not need to be set for it
   */
  Map<String, Operator> ops;

  /**
   * A CProgram containing the results of the last parse.
   */
  CProgram cProgram;


  /*** 
   *** Variables that are used during parsing 
   ***/

  /**
   * The parsing stack. Each entry is a list of expressions.
   */
  private StackOfList<Expr> parseStack;


  /**
   * flag indicating whether the next endElement event should be ignored
   */
  private boolean waiting = false;


  /**
   * tuid of the currently processed type declaration
   */
  private int curTypeUID;

  /**
   * Current control flow graph 
   */
  private CFGraph cfg;
  
  /** 
   * Current basic block
   */
  private BasicBlock bb;

  /**
   * Current context for declarations 
   */
  private Expr globalContext;
  private Expr funcContext;
  private Expr recordContext;


  /**
   * Creates a new <code>CXMLParser</code> instance.
   *
   * @param _efac an <code>ExprFactory</code> value
   */
  public CXMLParser (ExprFactory _efac) 
  {
    efac = _efac;
    skippedTags = new HashSet<String> ();
    typeTable = new TreeMap<Integer, Expr> ();
    declTable = new TreeMap<Integer, Expr> ();
    cfgTable = new TreeMap<Integer, CFGraph> ();
    ops = new HashMap<String, Operator> ();

    globalContext = null;
    funcContext = null;
    recordContext = null;

    initOps ();
  }


  /**********************************************************************/
  /* IMPLEMENTATION OF DefaultHandler */ 
  /**********************************************************************/
  
  public void startDocument () throws SAXException
  {
    parseStack = new StackOfList<Expr> ();
    
    parseStack.pushEmpty ();

  }

  public void endDocument () throws SAXException 
  {
    // -- the size of the parseStack at the end should be the same as
    // -- at the beginning, which is currently 1
    assert parseStack.size () == 1;

    cProgram =
      new CProgram (cfgTable, new CSymbolTable (typeTable, declTable));
  }


  public void startElement (String namespaceURI, String sName, 
      String qName, Attributes attrs)
    throws SAXException
    {
      String eName = tagName (sName, qName);

      Expr e = exprForTag (eName, attrs);

      if (e != null)
	{
	  parseStack.top ().add (e);
	  if (!waiting)
	    parseStack.pushEmpty ();
	}
      else
	{
	  skippedTags.add (eName);
	  // -- make sure that waiting is false since we are going to
	  // -- skip the end tag anyways
	  waiting = false;
	}
    }

  public void endElement (String namespaceURI, String sName, String qName)
    throws SAXException
    {
      String eName = tagName (sName, qName);

      if (skippedTags.contains (eName)) return;

      // -- if someone is waiting for an end tag (has no children)
      // --   skip the end tag, and switch the waiting flag
      if (waiting)
      {
	waiting = false;
	return;
      }

      // -- get kids
      LinkedList<Expr> kids = parseStack.pop ();
      // -- remove the last expression that is waiting for its children
      Expr last = parseStack.top ().removeLast ();
      // -- apply, which possibly returns a cannonical expression
      last = last.naryApply (kids);
      // -- put the expression back in
      parseStack.top ().add (last);

      // -- do any form of post-processing required
      postProcess (last);

    }


  /**********************************************************************/
  /* EXPRESSION CREATION METHODS */
  /**********************************************************************/
  

  /**
   * Extracts an unqualified name of an element
   *
   * @param sName a <code>String</code> value
   * @param qName a <code>String</code> value
   * @return a <code>String</code> value
   */
  private static String tagName (String sName, String qName)
  {
    // -- if sName is non empty use that, otherwise use qName
    return "".equals (sName) ? qName : sName;
  }


  /**
   * Returns the current context for variable declarations.
   *
   * @return an <code>Expr</code> value
   */
  private Expr curVarContext ()
  {
    return funcContext != null ? funcContext : globalContext;
  }

  /**
   * Returns a functionDecl with the specified uid. Constructs the
   * functionDecl if it was not created and logs it in the decl table.
   *
   * @param name a <code>String</code> value
   * @param uid an <code>int</code> value
   * @param context an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  private Expr buildFuncDecl (String name, int uid, Expr context)
  {
    Expr decl = declTable.get (uid);
    if (decl == null)
      {
	decl = efac.op (new CFunctionDeclOp (name, uid, context));
	declTable.put (uid, decl);
      }
    return decl;
  }
  

  /**
   * Constructs an open expression for a tag.  Side effects: may set
   * {@code waiting} to true to indicate that this tag is not
   * expecting any children and is simply waiting for its closure;
   * will set variable declaration context. The method should return
   * {@code null} for any tag that does not have an expression.
   *
   * @param eName a <code>String</code> value
   * @param attrs an <code>Attributes</code> value
   * @return an <code>Expr</code> value corresponding to tag 
   * @{code eName} or null
   */
  Expr exprForTag (String eName, Attributes attrs)
  {
    if (eName.equals ("tu"))
      // -- inititalize global context
      return (globalContext = efac.op (CGlobalOp.GLOBAL));

    else if (eName.equals ("typeDecl"))
    {
      // -- store current tuid for the following type definition
      curTypeUID = parseInt (attrs.getValue ("tuid"));

      int uid = -1;
      // -- extract the uid of this declaration. Only some typeDecls
      // -- have a uid
      if (attrs.getValue ("uid") != null)
	{
	  uid = parseInt (attrs.getValue ("uid"));
	  // -- due to a bug in the currrent XML output, we may have typeDecl
	  // -- with existing uid. These should be treated as synthetic decls
	  if (declTable.containsKey (uid)) uid = -1;
	}
      
      
      return efac.op (new CTypeDeclOp (uid, globalContext, 
				       attrs.getValue ("name")));
    }

    else if (eName.equals ("typeRef"))
    {
      // -- set waiting to true since typeRef does not expect any children
      waiting = true;

      Integer tuid = Integer.valueOf (attrs.getValue ("tuid"));
      // -- if the type is already in the type table, just return the
      // -- expression representing it, otherwise it is a simple type
      // -- that needs to be created
      if (typeTable.containsKey (tuid))
      {
	return typeTable.get (tuid);
      }
      else
      {
	// -- a typeRef to an undefined type indicates a simple type
	// -- that must be resolved by name
	Expr type = efac.op 
	  (new CSimpleTypeOp (tuid.intValue (), 
			      attrs.getValue ("name")));
	typeTable.put (tuid, type);
	return type;
      }
    }

    else if (eName.equals ("struct"))
      {
	recordContext = efac.op (new CStructOp (curTypeUID, 
						attrs.getValue ("tag")));
	return recordContext;
      }
    else if (eName.equals ("union"))
      {
	recordContext = efac.op (new CUnionOp (curTypeUID, 
					       attrs.getValue ("tag")));
	return recordContext;
      }

    else if (eName.equals ("fieldDecl"))
      {
	return efac.op (new CFieldDeclOp (attrs.getValue ("name"),
					  parseInt (attrs.getValue ("uid")),
					  recordContext));
      }
    else if (eName.equals ("arrayT"))
    {
      return efac.op (new CArrayTypeOp (curTypeUID));
    }

    else if (eName.equals ("domain"))
    {
      waiting = true;
      return efac.intExpr (parseInt (attrs.getValue ("size")));
    }
    else if (eName.equals ("ptrT"))
    {
      return efac.op (new CPtrTypeOp (curTypeUID));
    }
    else if (eName.equals ("refT"))
    {
      return efac.op (new CRefTypeOp (curTypeUID));
    }
    else if (eName.equals ("qualType"))
      {
	boolean constQ = parseBoolean (attrs.getValue ("const"));
	boolean volatileQ = parseBoolean (attrs.getValue ("volatile"));
	boolean restrictQ = parseBoolean (attrs.getValue ("restrict"));
	
	return efac.op (new CQualTypeOp 
			(curTypeUID, constQ, volatileQ, restrictQ));
	
      }
    
    else if (eName.equals ("globals"))
    {
      return globalContext;
    }
    else if (eName.equals ("varDecl"))
    {
      return efac.op (new CVarDeclOp (parseInt (attrs.getValue ("uid")),
				      curVarContext (),
				      attrs.getValue ("name")));
    }
    else if (eName.equals ("intcst"))
    {
      waiting = true;
      return efac.intExpr (parseInt (attrs.getValue ("value")));
    }
    else if (eName.equals ("stringcst"))
    {
      waiting = true;
      return efac.strExpr (attrs.getValue ("value"));
    }

    else if (eName.equals ("functionDecl"))
    {
      // -- create CFG for current function
      cfg = new CFGraph ();
      funcContext = buildFuncDecl (attrs.getValue ("name"),
				   parseInt (attrs.getValue ("uid")),
				   globalContext);
      return funcContext;
    }

    else if (eName.equals ("resultDecl"))
    {
      return 
	efac.op (new CResultDeclOp (parseInt (attrs.getValue ("uid")),
				    funcContext));
    }

    else if (eName.equals ("argList"))
    {
      return efac.op (TupleOp.TUPLE);
    }

    else if (eName.equals ("parmDecl"))
    {
      return efac.op (new CParmDeclOp (parseInt (attrs.getValue ("uid")),
				       funcContext,
				       attrs.getValue ("name")));
    }

    else if (eName.equals ("locals"))
    {
      return efac.op (TupleOp.TUPLE);
    }

    else if (eName.equals ("bb"))
    {
      int id = parseInt (attrs.getValue ("id"));
      BasicBlock newBB = cfg.getBBSafe (id);

      // note: since CFGraph constructor links 0 <-> 1, isLinked must be
      // checked on 1, otherwise 0.linkAfter (1) will point 1's next and prev
      // to itself
      if (bb != null && !newBB.isLinked ())
	bb.linkAfter (newBB);

      bb = newBB;

      return efac.op (new CBasicBlockOp (id));
    }
    else if (eName.equals ("stmtList"))
    {
      return efac.op (TupleOp.TUPLE);
    }
    else if (eName.equals ("succList"))
    {
      return efac.op (TupleOp.TUPLE);
    }
    
    else if (eName.equals ("edge"))
    {
      int dst = parseInt (attrs.getValue ("dst"));
      cfg.addEdge (bb, cfg.getBBSafe (dst),
	  Boolean.parseBoolean (attrs.getValue ("trueValue")),
	  Boolean.parseBoolean (attrs.getValue ("falseValue")));
      // -- edges are not added to the parse stack, so null is returned
      return null;
    }

    else if (eName.equals ("var") ||
	     eName.equals ("result") ||
	     eName.equals ("parm"))
    {
      waiting = true;
      Expr decl = declTable.get (Integer.valueOf (attrs.getValue ("uid")));
      assert decl != null : "Cannot find " + eName + 
	" with uid " + attrs.getValue ("uid");
      return decl;
    }
    else if (eName.equals ("function"))
      {
	waiting = true;
	return buildFuncDecl (attrs.getValue ("name"),
			      parseInt (attrs.getValue ("uid")),
			      globalContext);
      }

    else if (eName.equals ("labelDecl"))
    {
      return efac.op 
	(new CLabelDeclOp (attrs.getValue ("name"),
			   parseInt (attrs.getValue ("uid")),
			   funcContext,
			   parseInt (attrs.getValue ("labelUID"))));
    }
    else if (eName.equals ("args"))
    {
      return efac.op (new TupleOp ("args"));
    }
    else if (eName.equals ("defaultCase"))
    {
      waiting = true;
      return efac.op (CSwitchExprOp.DEFAULT_CASE);
    }
    else if (ops.containsKey (eName))
    {
      return efac.op (ops.get (eName));
    }

    return null;
  }

  /**
   * Putting stuff in tables and patching happens here.
   */
  void postProcess (Expr e)
  {
    // -- add declaration to the delcaration table. 
    if (CDeclOp.isDecl (e))
      {
	// -- don't add any synthetic decls -- these ones have
	// -- a negative uid
	int uid = CDeclOp.declUID (e);
	if (uid >= 0 && !declTable.containsKey (uid))
	  declTable.put (uid, e);

	// fall-through to handle CFunctionDecl
      }
    
    
    if (CFunctionDeclOp.isCFunctionDecl (e))
      {
	// -- log current CFG into the table
	cfgTable.put (CDeclOp.declUID (e), cfg);
	cfg = null;
	bb = null;
	funcContext = null;
      }
    else if (CRecordOp.isCRecord (e))
      {
	recordContext = null;
      }
    else if (isCTypeDecl (e))
    {
      int uid = CDeclOp.declUID (e);
      
      // -- if uid is positive this is a real typedecl, otherwise it
      // -- is a synthetic one
      if (uid >= 0 && !typeTable.containsKey (curTypeUID))
	{

	  Expr type = 
	    efac.op (new CDeclaredTypeOp (curTypeUID)).unaryApply (e);
	  typeTable.put (CTypeOp.typeUID (type), type);
	}
      else
	{
	  Expr type = CTypedDeclOp.declType (e);

	  // -- log type into the type table
	  typeTable.put (curTypeUID, type);
	}
    }

    else if (isCBasicBlock (e))
    {
      bb.setStmts (CBasicBlockOp.getStmtList (e).args ());
      // don't add bb to CFunctionDecl
      parseStack.removeCurrent ();
    }
    else if (isLabelExpr (e))
    {
      Expr labelDecl = CStmtOp.labelExprDecl (e);
      cfg.addLabel (labelDecl, bb);
      // -- Absolutelly add labels to the list of expressions. This
      // makes it possible to get the list of all the labels of a
      // basic block.  

      // parseStack.removeCurrent ();
    }
    
    // -- replace unary minus by the appropriate operator
    else if (e.op () == MINUS && e.arity () == 1)
    {
      parseStack.removeCurrent ();
      parseStack.top ().add (un_minus (e.arg (0)));
    }
  }


  /**
   * Returns the {@code CProgram} result of the parse
   *
   * @return a <code>CProgram</code> value
   */
  public CProgram getCProgram ()
  {
    return cProgram;
  }


  /**
   * Initialize the table of constant operators
   *
   */
  private void initOps ()
  {
    ops.put ("abs",	NumericOp.ABS);
    ops.put ("minus",	NumericOp.MINUS);
    ops.put ("plus",	NumericOp.PLUS);
    ops.put ("mul",	NumericOp.MULT);
    ops.put ("mod",	NumericOp.MOD);
    ops.put ("div",	NumericOp.DIV);

    ops.put ("addrOf",		CExprOp.ADDR_OF);
    ops.put ("arrayRef",	CExprOp.ARRAY_REF);
    ops.put ("compRef",		CExprOp.COMPONENT_REF);
    ops.put ("ref",		CExprOp.REF);

    ops.put ("callExpr",	CCallExprOp.CALL_EXPR);

    ops.put ("modifyExpr",		CStmtOp.MODIFY_EXPR);
    ops.put ("condExpr",		CStmtOp.COND_EXPR);
    ops.put ("goto",		CStmtOp.GOTO);
    ops.put ("labelExpr",	CStmtOp.LABEL_EXPR);
    ops.put ("return",		CStmtOp.RETURN);

    ops.put ("eq",	ComparisonOp.EQ);
    ops.put ("uneq",	ComparisonOp.NEQ);
    ops.put ("ne",	ComparisonOp.NEQ);
    ops.put ("lt",	ComparisonOp.LT);
    ops.put ("le",	ComparisonOp.LEQ);
    ops.put ("gt",	ComparisonOp.GT);
    ops.put ("ge",	ComparisonOp.GEQ);

    ops.put ("lor",	BoolOp.OR);
    ops.put ("land",	BoolOp.AND);
    ops.put ("lxor",	BoolOp.XOR);
    ops.put ("lnot",	BoolOp.NOT);

    ops.put ("bor",	BitOp.OR);
    ops.put ("band",	BitOp.AND);
    ops.put ("bxor",	BitOp.XOR);
    ops.put ("bnot",	BitOp.NOT);
    ops.put ("lshift",	BitOp.LSHIFT);
    ops.put ("rshift",	BitOp.RSHIFT);

    ops.put ("switchExpr",	CSwitchExprOp.SWITCH_EXPR);
    ops.put ("switchCond",	CSwitchExprOp.SWITCH_COND);
    ops.put ("switchBody",	CSwitchExprOp.SWITCH_BODY);
    ops.put ("caseLabelExpr",	CSwitchExprOp.CASE_LABEL_EXPR);

    ops.put ("memRef",	CMemRefOp.MEM_REF);
    ops.put ("symbol",	CMemRefOp.SYMBOL);
    ops.put ("base",	CMemRefOp.BASE);
    ops.put ("index",	CMemRefOp.INDEX);
    ops.put ("step",	CMemRefOp.STEP);
    ops.put ("offset",	CMemRefOp.OFFSET);
    ops.put ("original",CMemRefOp.ORIGINAL);

    ops.put ("uneq",	CExprOp.UNEQ);
    ops.put ("ult",	CExprOp.UNLT);
    ops.put ("unle",	CExprOp.UNLEQ);
    ops.put ("ungt",	CExprOp.UNGT);
    ops.put ("unge",	CExprOp.UNGEQ);

    ops.put ("ltgt",	CExprOp.LTGT);

    ops.put ("predec",	CExprOp.PREDEC);
    ops.put ("preinc",	CExprOp.PREINCR);
    ops.put ("postdec",	CExprOp.POSTDEC);
    ops.put ("postinc",	CExprOp.POSTINC);

    ops.put ("min",	CExprOp.MIN);
    ops.put ("max",	CExprOp.MAX);

    ops.put ("vaarg",	CExprOp.VA_ARG);

    ops.put ("asm",	CExprOp.ASM);
  }

  private void debugDump ()
  {
    cProgram.print 
      (new PrintWriter (new OutputStreamWriter (System.out), true));

    //System.out.println ("parseStack size: " + parseStack.size ());
    //System.out.println ("Top of parse stack is: " + parseStack.top ());
    //System.out.println ();

    System.out.println ("Tags ignored: " + skippedTags);
    System.out.println ();
  }

  /**
   * Parses a C-XML file.
   *
   * @param file a <code>File</code> value
   * @exception SAXException if an error occurs
   * @exception IOException if an error occurs
   * @exception ParserConfigurationException if an error occurs
   */
  public void parse (File file)
    throws SAXException, IOException, ParserConfigurationException
  {
    SAXParserFactory pfac = SAXParserFactory.newInstance ();
    
    SAXParser saxParser = pfac.newSAXParser ();
    saxParser.parse (file, this);
  }

  /**
   * Parses a C-XML resource specified by a URI.
   *
   * @param uri a <code>String</code> value
   * @exception SAXException if an error occurs
   * @exception IOException if an error occurs
   * @exception ParserConfigurationException if an error occurs
   */
  public void parse (String uri)
    throws SAXException, IOException, ParserConfigurationException
  {
    parse (new File (uri));
  }



  public static void main (String[] args)
  {
    CXMLParser chandler = new CXMLParser (new ExprFactoryImpl ());
    
    try
      {
	chandler.parse (args [0]);
	chandler.debugDump ();
      }
    catch (Exception ex)
      {
	ex.printStackTrace ();
      }
  }

  /**
   * A Stack in which each element is a List
   */
  private static class StackOfList<T> extends LinkedList<LinkedList<T>>
  {

    public StackOfList ()
    {
      super ();
    }

    public void pushEmpty ()
    {
      add (new LinkedList<T> ());
    }

    public void push (LinkedList<T> v)
    {
      add (v);
    }

    public LinkedList<T> top ()
    {
      return getLast ();
    }

    public LinkedList<T> pop ()
    {
      return removeLast ();
    }

    public void removeCurrent ()
    {
      top ().removeLast ();
    }
  }

}
