package edu.toronto.cs.expr;

/**
 * Array select (SEL) and update (UPD) operators.  
 *<p> To create a
 * selector {@code A[e]} use {@code fac.op (ArrayOp.ARRAY_SEL).binApply (A,
 * e);}</p>
 *
 * <p> To create an update {@code A[e] = v} use 
 * {@code fac.op (ArrayOp.ARRAY_UPD).naryApply (new Expr[]{A, e, v})}
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class ArrayOp extends NamedOp
{

  /**
   * Array selector operator: {@code A[e]}
   *
   */
  public static final ArrayOp ARRAY_SEL = new ArrayOp ("ARRAY_SEL", 2);
  
  
  /**
   * Array update operator: {@code A WITH [e] := v}
   *
   */
  public static final ArrayOp ARRAY_UPD = new ArrayOp ("ARRAY_UPD", 3);
    
  private ArrayOp (String name, int arity)
  {
    super (name, arity);
  }


  /**
   * Returns true if {@code e} is an array selector expression
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isArraySel (Expr e)
  {
    return e.op () == ARRAY_SEL;
  }
  
  /**
   * Assuming that {@code e.op} is an {@code ArrayOp} returns the
   * array name
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr getArray (Expr e)
  {
    return e.arg (0);
  }

  /**
   * Assuming that {@code e.op} is an {@code ArrayOp} returns the
   * array index
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr getIndex (Expr e)
  {
    return e.arg (1);
  }
  

  /**
   * Returns true if {@code e} is an array update expression
   *
   * @param e an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  public static final boolean isArrayUpd (Expr e)
  {
    return e.op () == ARRAY_UPD;
  }
  
  /**
   * Assuming that {@code e.op} is an {@code ArrayOp.ARRAY_UPD} returns the
   * value.
   *
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr getUpdExpr (Expr e)
  {
    return e.arg (2);
  }
  
  
  // -- creation methods

  /**
   * Creates a selector expression {@code a[e]}
   *
   * @param a an <code>Expr</code> value
   * @param e an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr sel (Expr a, Expr e)
  {
    return e.getFactory ().op (ARRAY_SEL).binApply (a, e);
  }

  /**
   * Creates a selector expression {@code a[n]}
   *
   * @param a an <code>Expr</code> value
   * @param n an integer value
   * @return an <code>Expr</code> value
   */
  public static final Expr sel (Expr a, int n)
  {
    return a.getFactory ().op (ARRAY_SEL).binApply 
	 (a, a.getFactory ().intExpr (n));
  }

  /**
   * Creates an update expression {@code a WITH [e] := v}
   *
   * @param a an <code>Expr</code> value
   * @param e an <code>Expr</code> value
   * @param v an <code>Expr</code> value
   * @return an <code>Expr</code> value
   */
  public static final Expr upd (Expr a, Expr e, Expr v)
  {
    return e.getFactory ().op (ARRAY_UPD).naryApply (new Expr[]{a, e, v});
  }
  
  
  
}
