package edu.toronto.cs.expr;

import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * An interface representing an expression.
 *
 * @author <a href="mailto:arie@cs.toronto.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public interface Expr
{

  
  /**
   * An empty array of expressions. A convinience constant that should
   * be used to avoid excessive construction of empty arrays.
   *
   */
  public static final Expr[] EMPTY_EXPR_ARRAY = new Expr [0];

  /**
   * The top-level operator of this expression.
   *
   * @return top-level operator
   */
  Operator op ();

  Expr nullaryApply ();
  
  /**
   * Applies this expression to a single argument.
   *
   * @param expr an argument to apply this expression to
   * @return an expression with a sole argument {@code expr}
   */
  Expr unaryApply (Expr expr);

  /**
   * Binary application. See {@link unaryApply}
   *
   * @param expr1 first argument
   * @param expr2 second argument
   * @return an expression with two arguments
   */
  Expr binApply (Expr expr1, Expr expr2);
  

  /**
   * n-ary application. See {@link #unaryApply}
   *
   * @param expr an array of arguments
   * @return an expression with arguments
   */
  Expr naryApply (Expr[] expr);
  
  /**
   * n-ary application over a list of expressions. See {@link
   * #naryApply(Expr[])}
   *
   * @param expr a list of arguments
   * @return an expression with arguments
   */
  Expr naryApply (List expr);
  
  /**
   * The arity of this expression, i.e. the number of children, or
   * top-level sub-expressions it has.
   *
   * @return number of children of this expression
   */
  int arity ();
  
  /**
   * The ith argument of this expression. 
   * 
   *
   * @param i an index of an argument to be returned. Must be between
   * 0 (inclusive) and {@code arity ()} (exclusive)
   * @return the ith argument of this expression
   * @throws java.lang.IndexOutOfBoundsException when {@code i} is
   * less than 0 or greater or equal to the arity of this expression.
   */
  Expr arg (int i);
  
  /**
   * The list of all arguments of this expression.
   *
   * @return a list of arguments.
   */
  List<Expr> args ();

  /**
   * Returns true if this expression is well-formed, false otherwise.
   *
   * @return true if this expression is well-formed, false otherwise.
   */
  boolean isWellFormed ();

  /**
   * Returns an expression obtained from this expression by replacing
   * every sub-expression that is a key in {@code subMap} with the
   * key's value.
   *
   * @param subMap a substitution map between expressions
   * @return an expression obtained by substitution
   */
  Expr subst (Map subMap);

  /**
   * A guarded substitution. This method behaves similar to {@link
   * #subst}, but only performs substitution that are allowed by a guard.
   *
   * @param subMap a substitution map between expressions
   * @param guard a substitution guard
   * @return an expression obtained by substitution
   */
  Expr guardedSubst (Map subMap, SubstGuard guard);

  /**
   * Returns an expression obtained from this expression by replacing
   * every opeartor that has a key in a substituion map with the key's
   * value.
   *
   * @param subMap a substition map between operators and expressions
   * @return an expression obtained by substition
   */
  Expr substOp (Map subMap);

  /**
   * Returns the factory that manages this expression.
   *
   * @return an expression factory
   */
  ExprFactory getFactory ();

  /** 
   * Checkes whether this expression contains a given sub-expression. 
   * 
   * @param e an expression whose containment is thought.
   * @return true if {@code e} is a sub-expression of this expression,
   * false otherwise.
   */
  boolean contains (Expr e);

  
  
  /**
   * Fills outSet with the set of variables in the expression
   *
   * @param outSet a <code>Set</code> value
   * @return a <code>Set</code> value
   */
  Set varSet (Set outSet);

  

/////// common infix operators
//   Expr and (Expr expr);
//   Expr or (Expr expr);
//   Expr not ();
//   Expr impl (Expr expr);
//   Expr iff (Expr expr);
  
//   Expr plus (Expr expr);
//   Expr minus (Expr expr);
//   Expr div (Expr expr);
//   Expr mult (Expr expr);
  
//   Expr bitAnd (Expr expr);
//   Expr bitOr (Expr expr);
//   Expr bitNot ();
  
//   Expr leq (Expr expr);
//   Expr geq (Expr expr);
//   Expr eq (Expr expr);
}
