package edu.toronto.cs.expr;


/**
 * An interface representing an expression factory. An expression
 * factory is responsible for creating and managing expressions. 
 *
 * <p>Expression are constructed using the builder design pattern
 * where a non-well-formed expression is first constructed from an
 * operator using {@link #op} method, which then made well-formed via
 * {@code XXXapply} methods of {@code Expr} interface.</p>
 *
 * <p>The following example illustrates how to construct a boolean
 * conjunction expression</p>
 *
 * <p><code>
    ExprFactory fac;
    Expr e1,e2;
    Expr e = fac.op (BoolOp.AND).binApply (e1, e2);
 *</code></p>
 *
 * @author <a href="mailto:arie@cs.toronto.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public interface ExprFactory
{

  /**
   * Constructs an expression representing an integer {@code n}
   *
   * @param n an integer whose expression is thought
   * @return an expression representing {@code n}
   */
  Expr intExpr (int n);
  
  /**
   * Constructs an expression representing a rational {@literal n/d}.
   *
   * @param n the numerator
   * @param d the denumerator
   * @return an expression representing {@literal n/d}
   */
  Expr ratExpr (int n, int d);
  
  /**
   * Constructs an expression representing an identifier {@code name}.
   *
   * @param name a String name of an identifier
   * @return an expression representing an identifier {@code name}
   */
  Expr var (String name);

  /**
   * returns a special non-deterministic value
   *
   * @return an <code>Expr</code> value
   */
  Expr nondetExpr ();

  /**
   * Returns a string expression representing {@code s}.
   *
   * @param s a <code>String</code> for the expression
   * @return an <code>Expr</code> representing {@code s}
   */
  Expr strExpr (String s);
  
  
  /**
   * Constructs an expression with top-level operator {@code op}. The
   * newly constructed expression has no arguments and is not
   * necessarily well-formed.
   *
   * @param op the top-level operator
   * @return an expression whose top-level operator is {@code op}
   */
  Expr op (Operator op);

  /**
   * Returns the constant 'true'
   *
   * @return a an expression for 'true'
   */
  Expr trueExpr ();

  /**
   * Returns the constant 'false'
   *
   * @return an expression for 'false'
   */
  Expr falseExpr ();

  

}
