package edu.toronto.cs.expr;

import java.util.*;

import edu.toronto.cs.util.SoftSoftHashMap;
import edu.toronto.cs.util.Primes;


public class ExprFactoryImpl implements ExprFactory
{

  Map unique;

  public ExprFactoryImpl ()
  {
    unique = new SoftSoftHashMap ();

  }

  public Expr intExpr (int num)
  {
    //return op (new JavaObjectOp (new Integer (num)));
    return ratExpr (num, 1);
  }
  public Expr ratExpr (int n, int d)
  {
    // -- alternative, if creator method is defined in RationalOp:
    // return RationalOp.ratExpr (this, n, d);
    return op (new RationalOp (n, d));
  }

  public Expr trueExpr ()
  {
    return op (BoolOp.TRUE);
  }
  public Expr falseExpr ()
  {
    return op (BoolOp.FALSE);
  }
  public Expr var (String name)
  {
    return op (new VariableOp (name));
  }
  public Expr nondetExpr ()
  {
    return op (NondetOp.NONDET);
  }
  public Expr strExpr (String s)
  {
    return op (new StringOp (s));
  }
  
  

  public Expr op (Operator op)
  {
    Expr expr = new ExprImpl (op);
    // -- Nullary expressions are canonized
    if (op.arity () == 0) 
      expr = canonize (expr);
    return expr;
  }


  public Expr canonize (Expr expr)
  {
    if (expr.op () instanceof NonCanonOp)
      return expr;

    Expr u = (Expr)unique.get (expr);
    if (u != null) return u;

    unique.put (expr, expr);
    return expr;
  }

  // empty Expr to help with printing open Exprs
  // ... probably not the right way to do this
  Expr EMPTY_EXPR = new ExprImpl (new VariableOp ("empty"));

  class ExprImpl extends AbstractExpr
  {
    Operator op;
    Expr[] args;


    public ExprImpl (Operator _op)
    {
      op = _op;
      args = EMPTY_EXPR_ARRAY;
    }


    public Operator op ()
    {
      return op;
    }

    public Expr naryApply (Expr[] expr)
    {
      args = expr;
      //return this;
      // -- applying an operator canonizes it
      return canonize (this);
    }

    public int arity ()
    {
      if (args == null) return 0;
      return args.length;
    }

    public Expr arg (int i)
    {
      /* this is probably bad
      if (i >= args.length)
	return EMPTY_EXPR;
	*/
      try
      {
	return args [i];
      }
      catch (ArrayIndexOutOfBoundsException e)
      {
	throw new ArrayIndexOutOfBoundsException
	  (op ().name () + ": " + e.getMessage ());
      }
    }

    public List<Expr> args ()
    {
      return Arrays.asList (args);
    }

    public boolean isWellFormed ()
    {

      if (op.isWellFormed (args))
      {
        for (int i = 0; i < arity (); i++)
          if (!arg (i).isWellFormed ()) return false;
        return true;
      }
      return false;
    }

    public Expr subst (Map subMap)
    {
      // -- if current expression is in the substitution map, 
      // -- return the map entry
      if (subMap.containsKey (this))
        return (Expr) subMap.get (this);

      if (arity () == 0) return this;

      // -- otherwise apply substitution to children


      Expr[] kids = new Expr [args.length];
      for (int i = 0; i < kids.length; i++)
        kids [i] = args [i].subst (subMap);

      return ExprFactoryImpl.this.op (op ()).naryApply (kids);
    }

    public Expr guardedSubst (Map subMap, SubstGuard guard)
    {
      if (!guard.okSubst (this))
        return this;

      // -- if current expression is in the substitution map, 
      // -- return the map entry
      if (subMap.containsKey (this))
        return (Expr) subMap.get (this);

      if (arity () == 0) return this;

      // -- otherwise apply substitution to children

      Expr[] kids = new Expr [args.length];
      for (int i = 0; i < kids.length; i++)
        kids [i] = args [i].guardedSubst (subMap, guard);

      return ExprFactoryImpl.this.op (op ()).naryApply (kids);
    }

    public Expr substOp (Map subMap)
    {

      // -- get the operator from the map, can be null if 
      // -- this operator should not be replaced
      Expr newOp = (Expr) subMap.get (op ());

      // -- bail out quickly on nullary expressions
      if (arity () == 0)
        return newOp == null ? this : newOp;

      // -- if newOp is null, duplicate the current operator
      if (newOp == null)
        newOp = ExprFactoryImpl.this.op (op ());      

      // -- recursively apply substOp to arguments
      Expr[] kids = new Expr [args.length];
      for (int i = 0; i < kids.length; i++)
        kids [i] = args [i].substOp (subMap);

      // -- apply the new operator to the arguments and return
      return newOp.naryApply (kids);
    }


    /* 
     * returns the set of all variables in the expression
     */
    public Set varSet (Set outSet)
    {
      // -- if hit a variable, add it to the out set
      if (VariableOp.isVariable (this))
	{
	  outSet.add (this);
	  return outSet;
	}
      
      
      // -- at the leaf, backout
      if (arity () == 0)
	return outSet;
      
      // -- process children
      for (int i = 0; i < arity (); i++)
	args [i].varSet (outSet);
    
      return outSet;
    }



    public boolean equals (Object v)
    {
      if (this == v) return true;
      if (v == null) return false;
      if (v instanceof ExprImpl)
        return equals ((ExprImpl) v, true);

      return false;
    }


    public boolean equals (ExprImpl expr, boolean identity)
    {
      if (op.equals (expr.op ()) && arity () == expr.arity ())
      {
        for (int i = 0; i < arity (); i++)
        {
          if (!identity && !args [i].equals (expr.arg (i)))
            return false;
          else if (identity && args [i] != expr.arg (i))
            return false;
        }
        return true;
      }
      return false;
    }


    public ExprFactory getFactory ()
    {
      return ExprFactoryImpl.this;
    }


    private int hashBound = 3;
    public int hashCode ()
    {
      // -- Compute the hash code for this expression. 
      // -- this assumes that expressions are immutable so hashCode of this
      // -- expression depeneds on its local information and identity hashcode
      // -- of its children
      long hashCode = op.hashCode () * Primes.primes [0];

      for (int i = 0; i < arity () && i < hashBound; i++)
        hashCode += 
          System.identityHashCode (args [i]) * Primes.getPrime (i + 1);

      return (int) (hashCode >>> 32);
    }


    public String toString ()
    {
      String s = "";

      // Special operator cases:

      // Tuple, no brackets
      if (op == TupleOp.PAIR_NB)
        return s + arg (0) + ", " + arg (1);

      // array selects and updates
      if (op == ArrayOp.ARRAY_UPD)
        return s + arg(0) + "[" + arg(1) + " |-> " + arg(2) + "]";
      if (op  == ArrayOp.ARRAY_SEL)
        return s + arg(0) + "[" + arg(1) + "]";

      // function call
      if (FunctionApplOp.isAppl (this))
        return FunctionApplOp.function (this) + "(" + FunctionApplOp.arg (this)
          + ")";

      // function type
      if (CompoundTypeOp.isFunT (this))
        return CompoundTypeOp.dom (this) + " -> " + CompoundTypeOp.rng (this);

      // CILListOp
      if (op == CILListOp.LIST)
        return "(list " + args () + ")";

      // choice
      if (ChoiceOp.isChoice (this))
        return "ch (" + arg (0) + ", " + arg (1) + ")";

      if (CompoundTypeOp.isArrayT (this))
        return op.name () + " " + arg (0) + "->" + arg (1);

      if (TupleOp.isTuple (this))
	return args ().toString ();

      // CRecordOp
      if (CRecordOp.isCRecord (this))
	return op.name ()
	  + ", tuid: " + CTypeOp.typeUID (this)
	  + ", tag: " + CRecordOp.tag (this)
	  + ", fields: " + args ();

      // CTypeOp
      if (CTypeOp.isCType (this))
	return op.name ()
	  + (CSimpleTypeOp.isCSimpleType (this) ?
	      ", typeName: "
	     + CSimpleTypeOp.simpleTypeName (this) : "")
	  + ", tuid: " + CTypeOp.typeUID (this)
	  + (CDeclaredTypeOp.isCDeclaredType (this) ?
	      ", TYPE_NAME: ("
	     + CDeclaredTypeOp.typeName (this) + ")" : "")
	  + (CArrayTypeOp.isCArrayType (this) ?
	     ", elmT: " + CArrayTypeOp.arrayRange (this)
	      + ", size: " + CArrayTypeOp.arraySize (this) : "")
	  + (CPtrTypeOp.isCPtrType (this) ?
	      ", subtuid: " + CTypeOp.typeUID (CPtrTypeOp.pointedType (this)) : "");

      // CNamedDecl
      // avoid loop via context
      // for scopedName, really we want the context name, not the op name, but
      // CFunctionDecl isn't quite a CNamedDecl
      /*
      if (CNamedDeclOp.isCNamedDecl (this))
	return
	  op.name ()
	  + ", scopedName: "
	  + (CRecordOp.isCRecord (CDeclOp.declContext (this)) ?
	     CRecordOp.tag (CDeclOp.declContext (this)) : "")
	  + (CFunctionDeclOp.isCFunctionDecl (CDeclOp.declContext (this)) ?
	     CNamedDeclOp.declName (CDeclOp.declContext (this)) : "")
	  + (CGlobalOp.isCGlobal (CDeclOp.declContext (this)) ?
	     CDeclOp.declContext (this).op ().name () : "")
	  + "::" + CNamedDeclOp.declName (this)
	  + ", uid: " + CDeclOp.declUID (this)
	  + (CTypedDeclOp.isCTypedDecl (this) ?
	      ", typetuid: " + CTypeOp.typeUID (CTypedDeclOp.declType (this)) :
	      "")
	  + (CInitTypedDeclOp.isCInitTypedDecl (this) ?
	      ", init: " + CInitTypedDeclOp.initExpr (this) : "");
      */
      if (CNamedDeclOp.isCNamedDecl (this))
	return
	  (CRecordOp.isCRecord (CDeclOp.declContext (this)) ?
	     CRecordOp.tag (CDeclOp.declContext (this)) : "")
	  + (CFunctionDeclOp.isCFunctionDecl (CDeclOp.declContext (this)) ?
	     CNamedDeclOp.declName (CDeclOp.declContext (this)) : "")
	  + (CGlobalOp.isCGlobal (CDeclOp.declContext (this)) ?
	     CDeclOp.declContext (this).op ().name () : "")
	  + "::"
	  + (CNamedDeclOp.declName (this) == null ?
	    CDeclOp.declUID (this) : CNamedDeclOp.declName (this));
	  

      if (CFunctionDeclOp.isCFunctionDecl (this))
	return
	  op.name ()
	  + ", name: (" + CFunctionDeclOp.declName (this) + ")"
	  + ", resultDecl: (" + CFunctionDeclOp.resultDecl (this) + ")"
	  + ", argList: (" + CFunctionDeclOp.args (this) + ")"
	  + ", locals: (" + CFunctionDeclOp.locals (this) + ")";


      if (op () instanceof PrefixOp)
	return op.name () + (arity () > 0 ? args () : "");

      // generic case
      return toIndentedString (0);
    }

    private String toIndentedString (int depth)
    {
      String s = "";

      // -- indentation
      for (int i = 0; i < depth; i++)
        s += "  ";

      if (op.arity () == 2 ||
          (op == BoolOp.AND && arity () == 2) ||
          (op == BoolOp.OR && arity () == 2) ||
          (op == BoolOp.ORN && arity () == 2) ||
          (op == NumericOp.PLUS && arity () == 2))
      {
        /*
           assert (arg (0) != null);
           assert (arg (0).toString () != null);
           assert (arg (1) != null);
           assert (arg (1).toString () != null);
           assert (op != null);
           assert (op.name () != null);
           */
        return s + "(" +
          (arg (0) == null ? "null" : arg (0).toString ()) +
          " " + op.name () + " " +
          (arg (1) == null ? "null" : arg (1).toString ()) +
          ")";
      }

      if (op.arity () == 1)
        return s + op.name () + " " + arg(0);

      if (op.arity () == 0)
        return s + op.name ();

      // -- root label
      s += op.name () + "\n";

      // -- add children with additional indentation
      for (Iterator it = args ().iterator (); it.hasNext ();)
      {
        ExprImpl ei = (ExprImpl) it.next ();
        if (ei != null)
          s += ei.toIndentedString (depth + 1) + "\n";
        else // ei == null
        {
          for (int i = 0; i < depth + 1; i++)
            s += "  ";
          s += "null\n";
        }
      }

      return s;  
    }
  }

  private static final SubstGuard STOP_AT_PLUS =
      new SubstGuard ()
      {
        public boolean okSubst (Expr e)
        {
          return e.op () != NumericOp.PLUS;
        }
      };

  /**
   * test code
   *
   * @param args a <code>String[]</code> value
   */
  public static void main (String[] args)
  {
    ExprFactory fac = new ExprFactoryImpl ();

    Expr x = fac.var ("x");
    Expr y = fac.var ("y");
    Expr z = fac.var ("z");

    System.out.println ("x is: " + x);
    System.out.println ("x is x: " + (x == fac.var ("x")));

    Expr exp = fac.op (ComparisonOp.EQ).binApply (x, y);
    //exp = fac.op (ComparisonOp.EQ).binApply (z, exp);

    System.out.println ("exp contains itself " + exp.contains (exp));
    System.out.println ("exp contains x " + exp.contains (x));
    System.out.println ("exp contains z " + exp.contains (z));

    System.out.println ("x = y: " + exp);

    Map subMap = new java.util.HashMap ();
    subMap.put (x, z);
    System.out.println ("z = y: " + exp.subst (subMap));

    System.out.println ("Variables of " + exp + " are " + 
			exp.varSet (new HashSet ()));
    exp = fac.op (ComparisonOp.EQ).binApply (exp, z);
    System.out.println ("Variables of " + exp + " are " + 
			exp.varSet (new HashSet ()));
    exp = fac.op (ComparisonOp.EQ).binApply (exp, fac.intExpr (10));
    System.out.println ("Variables of " + exp + " are " + 
			exp.varSet (new HashSet ()));
    System.out.println (exp + " hasNondet: " + NondetOp.hasNondet (exp));


    exp = fac.op (ComparisonOp.EQ).binApply (x, fac.nondetExpr ());
    System.out.println ("Variables of: " + exp + " are " + 
			exp.varSet (new HashSet ()));
    System.out.println (exp + " hasNondet: " + NondetOp.hasNondet (exp));
    
    

    // x - (x + z)
    Expr e = fac.op (NumericOp.MINUS).binApply
      (x, fac.op (NumericOp.PLUS).binApply (x, z));

    System.out.println ("x - (x + z): " + e);

    // should become z - (x + z), since we're guarding the + subtree
    System.out.println
      ("z - (x + z): " + e.guardedSubst (subMap, STOP_AT_PLUS));
  }
}

