package edu.toronto.cs.tp;


import java.util.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.tp.*;

import static edu.toronto.cs.expr.CVarDeclOp.*;

/**
 * A class for mapping operations in YASM's Expr syntax to strings
 * that can be fed to CVCL.
 *
 * When can this be used? Assignments crash it.
 *
 * @author <a href="mailto:anton@cs.toronto.edu">Anton Selyshchev</a>
 * @version 1.0
 */

public class CVCLExprPrinter
{
  Map<Operator,String> cvclOps;

  public CVCLExprPrinter()
  {
    cvclOps = new IdentityHashMap<Operator,String> ();
    initializeCvclOps();
  }

  void initializeCvclOps ()
  {
    cvclOps.put (SimpleTypeOp.INT, "INT");
    cvclOps.put (SimpleTypeOp.REAL, "REAL");
    cvclOps.put (BoolOp.TRUE, "TRUE");
    cvclOps.put (BoolOp.FALSE, "FALSE");
    cvclOps.put (BoolOp.NOT, "NOT");
    cvclOps.put (BoolOp.AND, "AND");
    cvclOps.put (BoolOp.OR, "OR");
    cvclOps.put (BoolOp.IMPL, "=>");
    cvclOps.put (ComparisonOp.GT, ">");
    cvclOps.put (ComparisonOp.GEQ, ">=");
    cvclOps.put (ComparisonOp.LT, "<");
    cvclOps.put (ComparisonOp.LEQ, "<=");
    cvclOps.put (ComparisonOp.EQ, "=");
    cvclOps.put (ComparisonOp.NEQ, "/=");
    cvclOps.put (NumericOp.PLUS, "+");
    cvclOps.put (NumericOp.MINUS, "-");
    cvclOps.put (NumericOp.UN_MINUS, "-");
    cvclOps.put (NumericOp.MULT, "*");
    cvclOps.put (NumericOp.DIV, "/");
    cvclOps.put (NumericOp.ABS, "ABS");
    cvclOps.put (CILIndirectionOp.DOT, ".");
    // un_minus
  }

  private String opToString (Operator op)
  {
    String result = cvclOps.get (op);
    if (result == null)
      throw new IllegalArgumentException ("Unknown operator: " + op + 
					  "of type " + op.getClass ());
    return result;
  }

  // it seems like variable names in CVCL can only use chars in
  // [a-zA-Z0-9_], so this cleans up a variable name
  private String exprVarNameToCvclVarName (String varname)
  {
    // XXX for now, we just replace ':' with '_'
    return varname.replace (':', '_').replace('@', '_');
  }

  // take an expression and return a String representation of the expression
  public String print (Expr e)
  {
    String result = "";

    if (e.op () instanceof TupleOp)
    {
	 if (e.op () != TupleOp.PAIR_NB)
	   result += "(";
	 for (int i = 0; i <= e.arity () - 2; i++)
	   result += print (e.arg (i)) + ",";
	 result += print (e.arg (e.arity () - 1));
	 if (e.op () != TupleOp.PAIR_NB)
	   result += ")";
	 return result;
    }

    // Special case for arrays.
    if (e.op () instanceof ArrayOp)
    {
      if (ArrayOp.isArraySel (e))
        return "(" + print(ArrayOp.getArray (e)) + "[" + 
          print(ArrayOp.getIndex (e)) + "])";

      if (ArrayOp.isArrayUpd (e))
        return "(" + print(ArrayOp.getArray (e)) + " WITH [" + 
          print(ArrayOp.getIndex (e)) + "] := " +
          print(ArrayOp.getUpdExpr (e)) + ")";

      throw new IllegalArgumentException ("unknown array op in " + e);
    }

    if (e.op () instanceof VariableOp)
    {
      return exprVarNameToCvclVarName(VariableOp.varName (e));
    }

    if (e.op () instanceof RationalOp) {
      if (RationalOp.isInt (e))
        return String.valueOf (RationalOp.getN (e));
      else
        return RationalOp.getN (e) + "/" + RationalOp.getD (e);
    }

    if (e.op () instanceof CompoundTypeOp) {
	 if (CompoundTypeOp.isArrayT (e))
	   return "ARRAY " + print (e.arg (0)) + " OF " + print (e.arg (1));
	 if (CompoundTypeOp.isFunT (e))
	   return print (e.arg (0)) + " -> " + print (e.arg (1));
    }

    try
    {
      if (e.arity () ==  0)
        return opToString (e.op ());

      if (e.arity () == 1)
        return opToString (e.op ()) + "(" + print (e.arg (0)) + ")";

      if (e.arity () >= 2)
      {

        // special case for function application
        if (FunctionApplOp.isAppl (e))
          return print (FunctionApplOp.function (e)) + "("
            + print (FunctionApplOp.arg (e)) + ")" ;

        result += "(";
        for (Iterator it = e.args ().iterator (); it.hasNext ();)
        {
          result += print ((Expr) it.next ());
          if (it.hasNext ())
            result += " " + opToString (e.op ()) + " ";
        }
        return result + ")";
      }
    }
    catch (IllegalArgumentException ex)
    {
      throw new RuntimeException (ex);
    }

    // EEK! We've failed, throw an error. Show the partial result to
    // help us see where we barfed.
    throw new IllegalArgumentException ("Cannot convert " 
        + e + " to cvcl, partial result: "
        + result);
  }  


  public static void main(String[] args)
  {
    ExprFactory fac = new ExprFactoryImpl ();
    //    CVCLTheoremProver tp = new CVCLTheoremProver (fac);
    CVCLExprPrinter printer = new CVCLExprPrinter();
    printer.initializeCvclOps();

    // -- create variables
    Expr x = fac.var ("x");
    Expr y = fac.var ("y");
    Expr intT = fac.op(SimpleTypeOp.INT);

    System.out.println (printer.print (x));

    //tp.declare (y, fac.op (SimpleTypeOp.INT));
    //    Expr intX = fac.op (SimpleTypeOp.INT).binApply(x, fac.intExpr(0));
    //Expr intY = fac.op (SimpleTypeOp.INT).binApply(y, fac.intExpr (0));


    Expr gt = fac.op (ComparisonOp.GT).binApply (x, fac.intExpr (3));
    Expr gt2 = fac.op (ComparisonOp.LT).binApply (x, fac.intExpr (2));
    Expr add = fac.op (NumericOp.PLUS).binApply (x, fac.intExpr (0));

    System.out.println (printer.print (gt));
    System.out.println (printer.print (add));

    // ------ Try an example with array operations.

    Expr sel = ArrayOp.sel(fac.var("A"), add);
    System.out.println (printer.print (sel));

    Expr ass = fac.op(CILAssignOp.ASSIGN).binApply(sel, fac.intExpr(5));
    Expr phi = fac.op(ComparisonOp.EQ).
      binApply(ArrayOp.sel(fac.var("A"), sel), fac.intExpr(10));

    Expr A = fac.var("A");
    Expr aupd = ArrayOp.upd(A, x, fac.intExpr(5));

    System.out.println (printer.print (aupd));
    System.out.println (printer.print (ArrayOp.sel(aupd, 
            ArrayOp.sel(aupd, x))));

    System.exit (0);


    
    CVCLProcess r = new CVCLProcess();
    
    System.out.println(r.declare(printer.print(x), printer.print(intT)).toString());
    System.out.println(r.where().toString());
    System.out.println(r.assertFormula(printer.print(gt)).toString());
    System.out.println("the command was " + printer.print(gt2));
    System.out.println(r.query(printer.print(gt2)).toString());

    // -- declare them to be INT
    //tp.declare (x, fac.op (SimpleTypeOp.INT));
    //tp.declare (y, fac.op (SimpleTypeOp.INT));

    //System.out.println("#########################");
    //System.out.println(printer.print(gt));
    //System.out.println(printer.print(add));

    // CVCLRunner r = new CVCLRunner();
    //     System.out.println(r.runWHERE().toString());
    //     System.out.println("WHERE MESSAGE COMPLETED");
    //     System.out.println(r.runDECLARE("x:INT").toString());
    //     System.out.println(r.runDECLARE("z:INT").toString());
    //     System.out.println(r.runASSERT("x > 0").toString());
    //     System.out.println(r.runASSERT("z > 5").toString());
    //     System.out.println(r.runQUERY("z > 3").toString());
    //     System.out.println(r.runWHERE().toString());
    //     System.out.println(r.runQUERY("z > 10").toString());
    //     System.out.println(r.runWHERE().toString());



  }



  // tp.push ();
  //     // -- assert x > 0;
  //     tp.assertFormula (fac.op (ComparisonOp.GT).binApply (x, fac.intExpr (0)));
  //     tp.push ();
  //     // -- assert y = 1
  //     tp.assertFormula (fac.op (ComparisonOp.EQ).binApply (y, fac.intExpr (1)));

  //     tp.push ();
  //     // -- query x > 0
  //     System.out.println (tp.query (fac.op (ComparisonOp.GT).binApply (x, fac.intExpr (0))));
  //     System.out.println ("Assumptions are: " + tp.getAssumptionsUsed ());
  //     tp.pop ();

  //     System.out.println ("stackLevel is " + tp.stackLevel ());

}



