package edu.toronto.cs.tp;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.util.*;

import edu.toronto.cs.yasm.YasmStatistics;

import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * An interface to the command line CVCL theorem prover.
 *
 * Provides methods to issue several commands and encapsulates command output
 * in a CVCLMessage.
 *
 * Currently does not fully handle errors; e.g., does not check for output from
 * cvcl on stderr, which indicates an invalid command.
 */
public class CVCLProcess
{
  YasmStatistics stats = new YasmStatistics ();
  Process cvclProcess;

  // indicates whether we're in a valid QUERY state, i.e., where
  // DUMP_ASSUMPTIONS output something meaningful and WHERE only includes our
  // assertions and not counterexample formulas
  boolean validQueryState;

  // XXX
  // out is OUR output TO the Process stdin
  // in is OUR input FROM the Process' stdout
  // err is OUT input FROM the Process' stderr

  Writer stdOut;
  TeeBufferedReader stdIn;

  // turns logging on or off.
  boolean logging;

  // writes a history of commands to a log file
  Writer log;

  // true when the last query was invalid and the stack level >= stack level at
  // which the query was made, false otherwise
  boolean invalidQueryState;

  // records the stack level at which an invalid query was made; should be -1
  // if no such query was made
  int invalidQueryStackLevel;

  int stackLevel;

  boolean debug;

  //
  // CVCL constants
  //

  // relative path to binary from CVCL_HOME
  private static final String cvclPath = "cvcl";

  // prompt, indicates previous command is finished and cvcl is ready for input
  private static final String PROMPT = "CVC> "; 

  // cvcl communication log filepath
  // XXX just put it where the jvm is run from, for now
  private static final String CVCL_LOG_PATH = "cvcl.log";

  // XXX Java regexes must match the entire string, unlike, e.g., Perl!

  // regex: captures stack level from WHERE output; scope is not captured
  private static final Pattern whereStackPattern =
    Pattern.compile ("Current stack level is (\\d+).*");

  // regex: captures formula from WHERE and DUMP_ASSUMPTIONS output
  private static final Pattern assertPattern =
    Pattern.compile ("ASSERT\\s*([^;]+).*"); 
  
  private static final String COMMENT_STRING = "%";
  
  public CVCLProcess ()
  {
    this(false);
  }
  
  public CVCLProcess (boolean logging)
  {
    this.logging = logging;
    System.out.println("LOGGING = " + logging);
    
    String cvclCmd;
    String[] cvclCmdWithArgs;

    // setup cvcl command line arguments
    // if it wasn't specified when java was invoked, 
    // we assume it's in the user's path
    cvclCmd = System.getProperty("cvcl.cmd");
    if (cvclCmd == null)
      cvclCmd = "cvcl";

    cvclCmdWithArgs = new String [] {cvclCmd, "+int"};

    try
    {
      ProcessBuilder cvclProcessBuilder = new ProcessBuilder (cvclCmdWithArgs);
      cvclProcessBuilder.redirectErrorStream (true);
      cvclProcess = cvclProcessBuilder.start ();
    }
    catch (Exception ex)
    {
      throw new RuntimeException (
          "Failed to start " + Arrays.toString (cvclCmdWithArgs) +
          ": " + ex.getMessage ());
    }

    // XXX this should be passed in
    debug = false;

    // initially there is no quey yet.
    invalidQueryState = false;

    // initially there's no query yet
    validQueryState = false;

    // initialize to an invalid stack level
    invalidQueryStackLevel = -1;

    openStreams ();

    stackLevel = where ().getStackLevel ();
  }

  public void setStats (YasmStatistics stats)
  {
    this.stats = stats;
  }

  /*
   * This method takes a String and removes all spaces from it
   */
  private static String removeSpaces (String s)
  {
    // iterate through the characters of the string. if a character is not
    // ' ' add it to a new string

    StringBuffer buf = new StringBuffer (s.length ());
    for (int i = 0; i < s.length (); i++)
    {
      char c = s.charAt (i);
      if (c != ' ')
        buf.append (c);
    }
    return buf.toString ();

  }

  private void openStreams ()
  {
    stdOut = new OutputStreamWriter (cvclProcess.getOutputStream ());

    try
    {
      if(logging)
      {
        //	log = new FileWriter (CVCL_LOG_PATH);
        log = SystemOutWrapper.writer;
        log.write (COMMENT_STRING + " *** Starting the Process\n");

      }
      else
      { 
        log = null;
      }

      stdIn = new TeeBufferedReader (
          new BufferedReader (
            new InputStreamReader (cvclProcess.getInputStream ()), 1024),
          log,
          COMMENT_STRING);

      // ensure we've got a prompt before we run any commands
      stdIn.prefixOn ();
      readUntilPrompt (stdIn, PROMPT);
    }
    catch (Exception e)
    {
      // stdIn.close();
      throw new CVCLCommunicationException (e);
    }

    // setup shutdown hook to close (and flush) the log
    Runtime.getRuntime ().addShutdownHook (
        new Thread ()
        {
          public void run ()
          {
            try
            {
              if (logging)
              {
                log.write (COMMENT_STRING + " *** Closing the Process\n");
                log.close ();
              }
            }
            catch (IOException ex)
            {
              // might as well give up, since this is a failure at shutdown
              System.err.println ("Failed to close log: " + ex.getMessage ());
            }
          }
        });
  }

  /**
   * Writes a String to stdOut and to the logFileWriter.
   *
   * @param cmd a <code>String</code> to be written
   */

  void writeCmd (String cmd)
  {
    // send a command to cvcl
    try
    {
      stdOut.write (cmd + ";\n");
      stdOut.flush ();
      if (logging) log.write ("\n" + cmd + ";\n");
      // send a hint to the TeeBufferedReader
      stdIn.prefixOn ();

      if (debug)
        System.err.println ("Wrote Command " + cmd + ";");
    }
    catch (Exception e)
    {
      // stdIn.close();
      throw new RuntimeException
        ("Error writing " + cmd + ": " + e);
    }
    // if anything shows up on stderr, assume we sent an illegal
    // command, eat up the prompt, and throw an exception
    // XXX this might be unnecessary since the Message classes check for
    // expected output

    /* XXX somehow check stderr here
       if (errIn.read () >= 0);
       {
       stdOut.write (";\n");
       stdOut.flush ();
    // readUntilPrompt (stdIn, "CVC> ");
    throw new CVCLCommunicationException ("Illegal command: " + cmd);
    }
    */
  }

  /** 
   * Reads everything from in until prompt is matched.
   *
   * Returns everything read, including prompt.
   *
   * Does not terminate if prompt is not matched.
   */
  private static String readUntilPrompt (Reader in, String prompt)
  {
    StringBuffer sb = new StringBuffer ();

    int charIndex;

    try
    {
      while (true)
      {
        // -- wait for the first character
        charIndex = in.read ();
        if (charIndex < 0)
          throw new IOException
            ("Unexpected end of file, last output was\n" + sb.toString ());
        sb.append ((char) charIndex);

        // -- read everything there is
        while (in.ready ())
        {
          charIndex = in.read ();

          if (charIndex < 0)
            throw new IOException
              ("Unexpected end of file, last output was\n" + sb.toString ());

          sb.append ((char) charIndex);
        }

        // return the buffer if it ends with prompt
        if (sb.toString ().endsWith (prompt))
          return sb.toString ();
      }
    }
    catch (IOException ex)
    {
      throw new CVCLCommunicationException (ex);
    }
  }

  public int getStackLevel ()
  {
    writeCmd ("WHERE");
    return WhereMessage.readCVCLOutput (stdIn).getStackLevel();
  }

  /** 
   * Makes a declaration. 
   * 
   * @param var name of the declared variable.
   * @param type type of the declared variable.
   * @return a <code>VoidMessage</code> containing the result of the command.
   */
  public VoidMessage declare (String var, String type)
  {
    validQueryState = false;

    writeCmd (var + ":" + type);

    return VoidMessage.readCVCLOutput (stdIn);
  }


  public AssertListMessage dump_assumptions ()
  {
    if(invalidQueryState)
    {
      // stdIn.close();
      throw new CVCLStateException("dumpAssumtions called in invQueryState");
    }

    if (!validQueryState)
    {
      // stdIn.close();
      throw new CVCLStateException ();
    }

    writeCmd ("DUMP_ASSUMPTIONS");
    AssertListMessage m = AssertListMessage.readCVCLOutput(stdIn);

    return m;

  }

  public WhereMessage where()
  {
    if(invalidQueryState)
    {
      // stdIn.close();
      throw new CVCLStateException("where was called in invalidQuertState");
    }

    writeCmd ("WHERE");
    return WhereMessage.readCVCLOutput (stdIn);
  }

  public CVCLMessage assertFormula(String command)
  {
    if(invalidQueryState)
    {
      // stdIn.close();
      throw new CVCLStateException("Assert was called in invalidQuertState");
    }

    writeCmd ("ASSERT " + command);
    return VoidMessage.readCVCLOutput (stdIn);
  }

  public CVCLMessage push()
  {

    if(invalidQueryState)
    {
      // stdIn.close();
      throw new CVCLStateException("Push was called in invalidQuertState");
    }

    writeCmd ("PUSH");
    stackLevel++;
    return VoidMessage.readCVCLOutput (stdIn);
  }

  public CVCLMessage pop()
  {
    writeCmd ("POP");
    VoidMessage m = VoidMessage.readCVCLOutput (stdIn);
    stackLevel--; // not checking if stackLevel == 0, process produces error
    if (invalidQueryState && stackLevel < invalidQueryStackLevel)
      invalidQueryState = false;

    return m;
  }

  /**
   * Runs POPTO command, assuming level is base-0.
   */
  public CVCLMessage popto (int level)
  {

    writeCmd ("POPTO " + level);
    VoidMessage m = VoidMessage.readCVCLOutput (stdIn);

    stackLevel = level;

    if (invalidQueryState && stackLevel < invalidQueryStackLevel)
      invalidQueryState = false;

    return m;
  }

  public ValidityMessage query(String command)
  {
    stats.startTPWriteQuery ();
    writeCmd ("QUERY " + command);
    stats.stopTPWriteQuery ();

    ValidityMessage m = ValidityMessage.readCVCLOutput (stdIn);

    if (m.isValid ())
      validQueryState = true;
    else
    {
      validQueryState = false;  
      invalidQueryState = true;
      invalidQueryStackLevel = stackLevel;
    }

    return m;  
  }

  /**
   * A marker interface for CVCL messages.
   */
  public static interface CVCLMessage
  {
  }

  /**
   * Encapsulates the empty response to commands such as ASSERT and PUSH;
   */
  public static class VoidMessage implements CVCLMessage
  {
    static VoidMessage VOID_MESSAGE = new VoidMessage ();

    public static VoidMessage readCVCLOutput (BufferedReader in)
    {
      String line = readUntilPrompt (in, PROMPT);

      if (line.equals (PROMPT))
      {
        // if the previous command only yielded a prompt, we're ok
        return VOID_MESSAGE;
      }
      else
      {
        // otherwise, the previous command output more than we expected, so
        // we treat this an error
        // stdIn.close();
        throw new CVCLParsingException (line);
      }
    } 

    public String toString()
    {
      return ("(no output)");
    }
  }

  /*
   * Encapsulates the response to a DUMP_ASSUMPTION command or part of the
   * response to a WHERE command, namely, the list of asserted formulas.
   */
  public static class AssertListMessage implements CVCLMessage
  {
    List<String> asserts;

    public AssertListMessage (List<String> asserts)
    {
      this.asserts = asserts;
    }

    /**
     * Reads the output from CVCL and parses it as a AssertListMessage.
     *
     * Assumes the previous command was WHERE or DUMP_ASSUMPTIONS and that
     * the remaining output looks like:
     *
     * % (something)
     * ASSERT ...;
     * ASSERT ...;
     * ...
     *
     * CVC>
     */
    public static AssertListMessage readCVCLOutput (BufferedReader in)
    {
      try
      {
        List<String> assertions = new ArrayList<String> ();

        // skip "%" line
        in.readLine ();

        // read everything up to and including the prompt
        String output = readUntilPrompt (in, PROMPT);
  
  // break output into lines, with newlines dropped
        StringTokenizer t = new StringTokenizer(output, "\n");

        // for each line ...
        while (t.hasMoreElements ())
        {
          String line = t.nextToken ();


          if (line.equals (PROMPT))
          {
            // reached PROMPT--done
            break;
          }

          // read from ASSERT to ";", dropping leading and trailing whitespace
          if (line.startsWith ("ASSERT"))
          {
            while (!line.contains (";"))
              line = line  + t.nextToken();
          }
          line = removeSpaces(line);

          Matcher m = assertPattern.matcher (line);

          // found an ASSERT line
          if (m.matches ())
          {
            // System.out.println("mumber of groups " + m.groupCount ());
            if (m.groupCount () == 1)
              assertions.add (m.group (1));
            else{
              // stdIn.close();
              throw new CVCLParsingException ("This one " +line);
            }

          }
        }

        return new AssertListMessage (assertions);

      }
      catch (Exception ex)
      {
        // stdIn.close();
        throw new CVCLCommunicationException (ex.getMessage ());
      }
    }

    /*
     * This method returns a list of assertions that compose this message
     */

    public List<String> getAsserts()
    {
      return asserts;
    }


    public String toString ()
    {
      return asserts.toString ();
    }
  }



  /**
   * Encapsulates the response to a WHERE command, namely, the stack level and
   * the assumption list.
   */
  public static class WhereMessage implements CVCLMessage
  {
    private int stackLevel;
    private AssertListMessage assertMsg;


    public WhereMessage (int stackLevel, AssertListMessage assertMsg)
    {
      this.stackLevel = stackLevel;
      this.assertMsg = assertMsg;
    }


    public AssertListMessage getAssertMsg()
    {
      return assertMsg;
    }

    public List<String> getAsserts ()
    {
      return assertMsg.getAsserts ();
    }

    public int getStackLevel ()
    {
      return stackLevel;
    }

    /**
     * Parses response to WHERE command and encapsulates response in a
     * WhereMessage.
     *
     * Assumes output looks like:
     *
     * Current stack level is 1 (scope 1).
     * % Active assumptions:
     * ASSERT (x > 0);
     * ...
     *
     * CVC>
     *
     * OR
     *
     * Current stack level is 1 (scope 1).
     * % No active assumptions 
     *
     * CVC>
     */
    public static WhereMessage readCVCLOutput (BufferedReader in)
    {
      String line;

      try
      {
        line = in.readLine (); 
      }
      catch(IOException e)
      {
        
        throw new CVCLCommunicationException (e.getMessage ());
      }

      Matcher matcher = whereStackPattern.matcher (line);

      // make sure we matched exactly the stack level group
      if (!matcher.matches () || matcher.groupCount () != 1)
      {
        
        throw new CVCLParsingException (line);
      }

      int stackLevel = Integer.parseInt (matcher.group (1));

      return new WhereMessage
        (stackLevel, AssertListMessage.readCVCLOutput (in));
    }

    public String toString()
    {
      return "Stack level: " + (new Integer (stackLevel)) +
        ", Active assumptions: " + assertMsg;
    }
  }

  /**
   * Encapsulates the response to a QUERY command, namely, the validity of the
   * query.
   */
  public static class ValidityMessage implements CVCLMessage
  {
    boolean valid;

    public ValidityMessage (boolean valid)
    {
      this.valid = valid;
    }

    public String toString()
    {
      return valid ? "Valid" : "Invalid";
    }

    public boolean isValid ()
    {
      return valid;
    }

    /**
     * Parses the result of a QUERY command and returns a corresponding
     * ValidityMessage.
     *
     * Assumes the pending output on in looks like:
     *
     * InValid.
     * CVC>
     *
     * OR
     *
     * Valid.
     * CVC>
     *
     * (In CVCL 20060527, "Invalid" is capitalized as shown.)
     */
    public static ValidityMessage readCVCLOutput (BufferedReader in)
    {
      String line;
      String errorLine;
      String s;

      try
      {
        line = readUntilPrompt (in, PROMPT);
      }
      catch (Exception e)
      {
        throw new CVCLCommunicationException (e);
      }

      /* We use endsWith() instead of equals() because when we give
       * long inputs to CVCL, it sometimes prints a hyphen, which
       * prompts the user for more input. We then read these back in
       * when we read CVCL's output. Hence, we use endsWith() to
       * ignore them.
       */
      if (line.toUpperCase ().endsWith ("INVALID.\n" + PROMPT))
      {
        return new ValidityMessage (false);
      }
      else if (line.toUpperCase ().endsWith ("VALID.\n" + PROMPT))
      {
        return new ValidityMessage (true);
      }
      else
      {
        // stdIn.close();
        throw new CVCLCommunicationException
          ("Unexpected output '" + line + "'");
      }
    }
  }

  /**
   * Indicates a failure to send/receive to/from CVCL.
   */
  public static class CVCLCommunicationException extends RuntimeException
  {
    public CVCLCommunicationException (String message)
    {
      super (message);
    }

    public CVCLCommunicationException (Exception ex)
    {
      super (ex);
    }
  }

  /**
   * Indicates a failure to interpret a command response from CVCL.
   */
  public static class CVCLParsingException extends RuntimeException
  {
    public CVCLParsingException (String message)
    {
      super ("Unable to parse: '" + message + "'");
    }
  }

  /**
   * Indicates an attempt to issue a command in an invalid state, e.g.,
   * DUMP_ASSUMPTIONS after an invalid QUERY.
   */
  public static class CVCLStateException extends RuntimeException
  {
    public CVCLStateException(String s)
    {
      super(s);
    }

    public CVCLStateException()
    {
      super();
    }
  }
}

