package edu.toronto.cs.tp;

import edu.toronto.cs.expr.*;
import java.util.List;


/**
 * A generic theorem prover interface.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public interface ITheoremProver
{

  /**
   * Returns an {@link edu.toronto.cs.expr.ExprFactory} used by this
   * theorem prover.
   *
   * @return an <code>ExprFactory</code> 
   */
  ExprFactory exprFactory ();


  /**
   * Declares that the expression {@code name} is of type given by the
   * expression {@code type}. It is assumed that {@code name} is an
   * identifier expression and {@code type} is a valid type
   * expression.
   *
   * @param name an identifier being declared
   * @param type the type to be associated with {@name} 
   */
  void declare (Expr name, Expr type);

  /**
   * Returns true if {@code expr} is valid, false otherwise. Note that
   * this may change the context of the theorem prover.
   *
   * @param expr an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  boolean query (Expr expr);
  
  /**
   * Returns true if the {@expr} is valid, false if it is not. Ensures
   * that the state of the theorem prover is unchanged
   *
   * @param expr an expression to check for validity
   * @return true if {@code expr} is valid, false otherwise.
   */
  boolean quickQuery (Expr expr);

  /**
   * Returns {@code Boolean.TRUE} if {@code tExpr} is valid, {@code
   * Boolean.FALSE} if {@code fExpr} is valid, and {@code null}
   * otherwise. The state of the theorem prover depends on the last
   * query made.
   *
   * @param tExpr an <code>Expr</code> value
   * @param fExpr an <code>Expr</code> value
   * @return a <code>Boolean</code> value
   */
  Boolean checkTruth (Expr tExpr, Expr fExpr);

  /**
   * Same as {@link edu.toronto.cs.tp.ITheoremProver#checkTruth} but
   * does not change the state of the theorem prover.
   *
   * @param tExpr an <code>Expr</code> value
   * @param fExpr an <code>Expr</code> value
   * @return a <code>Boolean</code> value
   */
  Boolean quickCheckTruth (Expr tExpr, Expr fExpr);
  
  /**
   * Same as {@code checkTruth (expr, expr.not ())}.
   *
   * @param expr an <code>Expr</code> value
   * @return a <code>Boolean</code> value
   */
  Boolean checkTruth (Expr expr);
  
  /**
   * Same as {@link
   * edu.toronto.cs.tp.ITheoremProver#checkTruth(Expr)}, but leaves
   * the theorem prover state unchanged.
   *
   * @param expr an <code>Expr</code> value
   * @return a <code>Boolean</code> value
   */
  Boolean quickCheckTruth (Expr expr);
  
  
  /**
   * Adds {@code expr} to the list of currect assertions.
   *
   * @param expr an <code>Expr</code> value
   */
  void assertFormula (Expr expr);

  /**
   * Adds all expressions in @{code List<Expr> l} to the list of
   * currect assertions.
   *
   * @param l a <code>List<Expr></code> value
   */
  void assertFormula (List<Expr> l);
  
  
  /**
   * Quick check of validity of {@code e1 => e2}. Theorem prover state
   * is unchanged.
   *
   * @param e1 an <code>Expr</code> value
   * @param e2 an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  boolean implies (Expr e1, Expr e2);
  
  /**
   * Returns the list of assumptions used in the most recent proof.
   *
   * @return a <code>List<Expr></code> value
   */
  List<Expr> getAssumptionsUsed ();
  

  
  /**
   * Returns the current stack level of contexts.
   *
   * @return an <code>int</code> value
   */
  int stackLevel ();

  /**
   * Pushes an empty context in the context stack.
   *
   */
  void push ();

  
  /**
   * Removes the top most context in the context stack.
   *
   */
  void pop ();


  /**
   * Removes contexts until @{code stackLevel == i}.
   *
   * @param i an <code>int</code> value
   */
  void popto (int i);
  
  
}
