package edu.toronto.cs.tp;

import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.tp.cvcl.ValidityChecker;
import edu.toronto.cs.tp.cvcl.CVectorExpr;
import edu.toronto.cs.tp.cvcl.Type;
import edu.toronto.cs.tp.cvcl.CVCLUtil;


/**
 * Implementation of {@link edu.toronto.cs.tp.ITheoremProver} based on
 * a binary interface to CVCLite.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class JCVCLTheoremProver extends AbstractTheoremProver
{
  
  // -- an instance of validity checker
  ValidityChecker vc;
  
  // -- an instance of a converter between CVCL expressions and
  // -- edu.toronto.cs.expr.Expr
  CVCLExprConverter exprConverter;
  
  // -- a list of cvcl expressions that keeps track of all declared
  // -- identifiers to avoid loosing them to a garbage collector.
  private List<edu.toronto.cs.tp.cvcl.Expr> decls = 
    new LinkedList<edu.toronto.cs.tp.cvcl.Expr> ();
  
  /**
   * Constructs a {@code JCVCLTheoremProver} given an expression
   * factory, a validity checker itself, and a converter to/from CVCL
   * expressions
   *
   * @param _exprFactory an <code>ExprFactory</code> value
   * @param _vc a <code>ValidityChecker</code> value
   * @param _exprConverter a <code>CVCLExprConver</code> value
   */
  public JCVCLTheoremProver (ExprFactory _exprFactory, 
			     ValidityChecker _vc,
			     CVCLExprConverter _exprConverter)
  {
    super (_exprFactory);
    vc = _vc;
    exprConverter = _exprConverter;
  }

  /**
   * A convinience constructor that does not require an external
   * {@code CVCLExprConverter}.
   *
   * @param _exprFactory an <code>ExprFactory</code> value
   * @param _vc a <code>ValidityChecker</code> value
   */
  public JCVCLTheoremProver (ExprFactory _exprFactory,
			     ValidityChecker _vc)
  {
    this (_exprFactory, _vc, new CVCLExprConverter (_exprFactory, _vc, null));
  }

  /**
   * A convinience constructor that only requires an {@code
   * ExprFactory} instance.
   *
   * @param _exprFactory an <code>ExprFactory</code> value
   */
  public JCVCLTheoremProver (ExprFactory _exprFactory)
  {
    this (_exprFactory, CVCLUtil.newValidityChecker ());
  }
  
  

  /**
   * Returns the instance of the {@link
   * edu.toronto.cs.tp.cvcl.ValidityChecker} used internally. This
   * should only be used to provide backward compatability to classes
   * that want to use {@code ValidityChecker} directly.
   *
   * @return an instance of <code>ValidityChecker</code> 
   */
  public ValidityChecker getValidityChecker ()
  {
    return vc;
  }

  /**
   * Returns the instance of {@link
   * edu.toronto.cs.expr.CVCLExprConverter} that is used
   * internally. This should only be used to provide backward
   * compatibility to classes that want to use {@code
   * CVCLExprConverter} directly.
   *
   * @return an instances of <code>CVCLExprConverter</code> 
   */
  public CVCLExprConverter getExprConverter ()
  {
    return exprConverter;
  }
  
  
  public void declare (Expr name, Expr type)
  {
    // -- for now, we only allow declaration of variables
    if (!(name.op () instanceof VariableOp))
      throw new 
	IllegalArgumentException ("First argument to declare must be a " + 
				  "VariableOp expression");
    
    // -- to avoid loosing declarations to garbage collector, we 
    // -- store this declaration 
    decls.add (vc.varExpr (VariableOp.varName (name), cvclType (vc, type)));

  }

  // XXX At the moment we ignore types and assume that they are always
  // XXX integer, but will need to work on this.
  private static Type cvclType (ValidityChecker vc, Expr type)
  {
    Operator op = type.op ();
    if (op == SimpleTypeOp.INT)
      return vc.intType ();
    else if (op == SimpleTypeOp.REAL)
      return vc.realType ();
    else if (op == SimpleTypeOp.VOID)
      return vc.createType ("VOID");
    else if (op == SimpleTypeOp.DUMMY)
      return vc.createType ("DUMMY");
    else
      throw new IllegalArgumentException ("Expression " + type + 
					  " is not a valid type expression");
    
  }
  
  
  public boolean query (Expr expr)
  {
    return vc.query (exprConverter.toCVCL (expr), false);
  }
  
  public List<Expr> getAssumptionsUsed ()
  {
    CVectorExpr vector = new CVectorExpr (0);
    vc.getAssumptions (vector);
    
    return cvcVectorToList (vector);
  }

  public void assertFormula (Expr expr)
  {
    vc.assertFormula (exprConverter.toCVCL (expr));
  }

  public int stackLevel ()
  {
    return vc.scopeLevel ();
  }
  public void push ()
  {
    vc.push ();
  }
  public void pop ()
  {
    vc.pop ();
  }
  public void popto (int l)
  {
    vc.popto (l);
  }
  

  /**
   * Utility method to convert a vector of CVCL expressions to a list
   * of {@link edu.toronto.cs.expr.Expr}.
   *
   * @param v a <code>CVectorExpr</code> value
   * @return a <code>List<Expr></code> value
   */
  protected List<Expr> cvcVectorToList (CVectorExpr v)
  {
    // -- apparently a CVector can be larger than ArrayList, but we
    // -- ignore this at the moment
    int vSize = (int) v.size ();

    List<Expr> out = new ArrayList<Expr> (vSize);
    
    for (int i = 0; i < vSize; i++)
      out.add (exprConverter.fromCVCL (v.get (i)));

    return out;
  }
  
  

  /**
   * Simple test of the theorem prover
   *
   */
  public static void test1 ()
  {

    ExprFactory fac = new ExprFactoryImpl ();

    JCVCLTheoremProver tp = new JCVCLTheoremProver (fac);

    // -- create variables
    Expr x = fac.var ("x");
    Expr y = fac.var ("y");
    
    // -- declare them to be INT
    tp.declare (x, fac.op (SimpleTypeOp.INT));
    tp.declare (y, fac.op (SimpleTypeOp.INT));
    
    
    tp.push ();
    // -- assert x > 0;
    tp.assertFormula (fac.op (ComparisonOp.GT).binApply (x, fac.intExpr (0)));
    tp.push ();
    // -- assert y = 1
    tp.assertFormula (fac.op (ComparisonOp.EQ).binApply (y, fac.intExpr (1)));
    
    tp.push ();
    // -- query x > 0
    System.out.println (tp.query (fac.op (ComparisonOp.GT).binApply (x, fac.intExpr (0))));
    System.out.println ("Assumptions are: " + tp.getAssumptionsUsed ());
    tp.pop ();

    System.out.println ("stackLevel is " + tp.stackLevel ());
    
  }


  public static void main (String[] args)
  {
    test1 ();
  }
  
  
}
