package edu.toronto.cs.yasm;

import java.util.*;
import java.io.*;

import java.util.prefs.*;

import edu.toronto.cs.util.*;
import edu.toronto.cs.util.Clapi.*;

import edu.toronto.cs.yasm.pcfa.*;
import edu.toronto.cs.yasm.pgenerator.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.ctl.antlr.*;
import edu.toronto.cs.ctl.antlr.CTLNodeParser.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.modelchecker.*;
import edu.toronto.cs.cfa.CFAMvSetFactory.*;
import edu.toronto.cs.boolpg.BoolProgramCompiler;
import edu.toronto.cs.proof2.*;
import edu.toronto.cs.proof2.CTLProver.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.yasm.pprogram.*;
import edu.toronto.cs.yasm.abstractor.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.util.*;
import edu.toronto.cs.yasm.wp.WPComputer;
import edu.toronto.cs.yasm.reachability.*;
import edu.toronto.cs.yasm.nontermination.*;

import edu.toronto.cs.cparser.*;
import edu.toronto.cs.tp.*;


/**
 * YasrApp.java
 *
 * Yet Another Software Model Checker
 *
 * Created: Tue Jun 22 13:38:03 2004
 *
 * @author <a href="mailto:maxin@epoch.cs">Xin Ma</a>
 * @version 1.0
 */
public class YasrApp 
{
  public static boolean DEBUG = true;  

  /**
   * Standard Output
   *
   */
  public static PrintWriter out = 
    new PrintWriter (new OutputStreamWriter (System.out), true);

  /**
   * Standad Error
   *
   */
  public static PrintWriter err = 
    new PrintWriter (new OutputStreamWriter (System.err), true);

  /**
   * Preferences of the application
   *
   */
  public static final Preferences PREFS 
    = Preferences.userRoot ().node ("edu/toronto/cs/yasm/YasrApp");


  PProgram pProgram = null;

  PProgramReachabilityCompiler compiler = null;  

  Refiner refiner = null;


  File cFile;

  File initPredFile;
  
  /**
   * CTL Property being checked
   *
   */
  CTLNode prop;

  /**
   * Model used for reachability analysis
   *
   */
  ReachabilityStructure model;
  

  /**
   * Our model-checker
   *
   */
  ReachabilityChecker yasr;

  NonTerminationChecker yasnt;
  
  /**
   * If true then a proof is produced,  false by default
   *
   */
  boolean doProof = true;

  /**
   * true if functions should be inlined
   *
   */
  boolean inline = false;
  
  /**
   * our algebra
   *
   */
  BelnapAlgebra algebra;


  String refinerType = "new"; // <-- override this
  String pGeneratorType = "bend:wpd:wps:wpa";
  int selectorType = -1;
  boolean stmtBlocking = false;


  boolean found = false;
  int newPredFound = 0;
  static int MAX_PREDICATES_PER_REFINEMENT = 1;

  ExprFactory fac;
  
  boolean useHyperEdges;
  boolean unknownInit;

  
  StopWatch overall = new StopWatch ();


  // -- for EG
  List pathForEG;
  Map  statesOnPathForEG;
  
  
  public YasrApp () 
  {
    fac = new ExprFactoryImpl ();
    // setting up some defaults
    // redundant, but safe
    doProof = true;

    useHyperEdges = true;
    unknownInit = true;
    pGeneratorType = "bend:wpd:wps:wpa";
    refinerType = "cbj";
    
    
  }

  public void setCFile (File f)
  {
    cFile = f;
  }
  

  public void setInitPredFile (File f)
  {
    initPredFile = f;
  }

  public void setProp (CTLNode v)
  {
    prop = v;
  }
  
  public CTLNode getProp ()
  {
    return prop;
  }

  public void setRefinerType (String v)
  {
    refinerType = v;
  }

  public void setSelectorType (String s)
  {
    if (s.equals ("dd"))
      selectorType = PProgram.DDSelector;
    else if (s.equals ("bin"))
      selectorType = PProgram.BinSelector;
    else if (s.equals ("lin"))
      selectorType = PProgram.LinSelector;
    else
      throw new RuntimeException ("Unknown selector type: " + s);
  }

  public void setStmtBlocking (boolean v)
  {
    stmtBlocking = v;
  }

  public void setMaxPredsPerRefinement (int i)
  {
    MAX_PREDICATES_PER_REFINEMENT = i;
  }

  public void setPGeneratorType (String v)
  {
    pGeneratorType = v;
  }
  
  public void setHyper (boolean v)
  {
    useHyperEdges = v;
  }
  public void setUnknownInit (boolean v)
  {
    unknownInit = v;
  }

  


  /**
   * doProof setter
   *
   * @param v a <code>boolean</code> value
   */
  public void setDoProof (boolean v)
  {
    doProof = v;
  }

  public boolean isDoProof ()
  {
    return doProof;
  }
  

  public void setInline (boolean v)
  {
    inline = v;
  }
  
  public void setModel (ReachabilityStructure v)
  {
    model = v;
    algebra = model.getAlgebra ();
  }
  
 
  public void setReachabilityChecker (ReachabilityChecker v)
  {
    yasr = v;
  }


  


  static Refiner getRefiner (String name, ExprFactory fac, PProgram p)
  {
    if (name.equals ("local-cbj-vs"))
      return new LocalCBJPredicateRefinerVisitor (fac, p, true, true);

    
    throw new RuntimeException ("Unknown refiner: " + name);
  }


  ReachabilityStructure compileCProgram ()
    throws antlr.ANTLRException, FileNotFoundException
  {

    GnuCParser gnuCParser = null;

    if (pProgram == null)
      {
	//stats.startParse ();
	gnuCParser = GnuCParserUtil.getParser
	  (new FileReader (cFile));

	Expr programTree =
	  CILTransformerUtil.getProgram (fac, gnuCParser, stmtBlocking);
	//stats.stopParse ();
	  
	
	pProgram = (new NullExprAbstractor (fac)).
	  doProgramAbstract (programTree, selectorType);	  

        PProgramPatcher.patch (pProgram);

	// initialize symbol table
	SymbolTable st =
	  SymbolTableInitializer.init (pProgram, new SymbolTable (), true);

	pProgram.setSymbolTable (st);
	
      }
      

    if (refiner == null)
      {
        refiner = getRefiner (refinerType, fac, pProgram);

	// XXX Old way of communicating the symbol table to 
	// XXX the theorem prover
	refiner.doDeclsRefine (pProgram);


        if (gnuCParser != null && initPredFile != null)
	  {
	    // parse initial predicates using the declarations of
	    // the target C program

	    gnuCParser =
	      GnuCParserUtil.getParser
	      (new FileReader (initPredFile), gnuCParser.symbolTable);

	    Expr predTree =
	      CILTransformerUtil.getProgram (fac, gnuCParser);

	    refiner.addPredicates
	      (PredicateExtractor.extract (predTree));
	    
	  }

	
      }

    //stats.startRefine ();
    pProgram = refiner.doProgramRefine (pProgram);
    //    stats.stopRefine ();


    System.err.println ();
    System.err.println ("Refined Boolean Program");
    System.err.println ();
    //       pProgram.print (YasrApp.err);
    System.err.println ();
    System.err.println ("Refined Boolean Program END");
    System.err.println ();

    //Ou Wei - test 
    System.out.println ("Test Printing Begin ....");

    //       PrintWriter myout = new PrintWriter(System.out);
    //       pProgram.print(myout);
    //       myout.flush ();

    pProgram.accept 
      (new PrintingVisitor (new IndentPrintWriter (System.out)), null);
    System.out.println ("Test Printing End ....");


    if (compiler == null)
      {
	if (inline)
	  compiler = new InliningReachabilityCompiler ();
	else
	  compiler = new SummarizingReachabilityCompiler ();
      }
      

    compiler.setPProgram (pProgram);
    // XXX just for now
    //compiler.setMaxDDVars (8);
    compiler.setMaxDDVars (100);
    //compiler.setMaxDDVars (20);
      

    setModel (compiler.compile ());

    return model;
  }

  /**
   * Prepare our property to be checked
   *
   */
  CTLNode prepareProperty ()
  {
    if (! (prop instanceof CTLEFNode))
      {
	YasrApp.out.println ("Property is not supported");
	System.exit (1);
      }
    
    CTLNode right = prop.getRight ();
    
    if (! (right instanceof CTLEqualsNode))
      {
	YasrApp.out.println ("Only 'EF pc = XXX' is supported");
	System.exit (1);
      }
    
    String pcValue = right.getRight ().toString ();
    
    YasrApp.err.println ("Looking for reachability of " + pcValue);
    PStmt stmt = pProgram.getStmt (pcValue);
    if (stmt == null) 
      {
	YasrApp.err.println ("No such statement!");
	System.exit (1);
      }
    
    NodeSet nodes = model.getNodeSet ().createNext ();
    nodes.add (new Node (new Node.NodeId (stmt.getId ()),
			 model.getMvSetFactory ().top ()));
    model.setNodeSet (nodes);
    return null;
  }



  /**
   * Prepare our property to be checked
   *
   */
  CTLNode prepareEGProperty ()
  {

    // -- set "END" node with "FALSE", other nodes willl be set "TRUE"
    // -- during the "initialize ()" of in NonTerminationChecker.java

    String pcValue = "END";
    
    YasrApp.err.println ("Check EG true");
    PStmt stmt = pProgram.getStmt (pcValue);
    if (stmt == null) 
      {
	YasrApp.err.println ("No such statement!");
	System.exit (1);
      }
    
    NodeSet nodes = model.getNodeSet ().createNext ();
    nodes.add (new Node (new Node.NodeId (stmt.getId ()),
			 model.getMvSetFactory ().bot ()));
    model.setNodeSet (nodes);
    return null;
  }



  AlgebraValue extractValue ()
  {
    int initId = pProgram.getStmt ("init").getId ();
    Node initNode = model.getNodeSet ().get (new Node.NodeId (initId));
    
    if (initNode == null) return model.getAlgebra ().bot ();

    
//     YasrApp.err.println ("init node is labeled by: ");
//     YasrApp.err.println (initNode);
    
//     YasrApp.err.println ("looking at another node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(12)));

//     YasrApp.err.println ("looking at next to another node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(13)));

//     YasrApp.err.println ("looking at the before error node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(14)));

//     YasrApp.err.println ("looking at the before error node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(15)));

//     YasrApp.err.println ("looking at the before error node");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId(16)));
    

//     YasrApp.err.println ("looking at init node before extractValue");
//     YasrApp.err.println (initNode);
    
//     YasrApp.err.println ("fooEntry");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId (3)));
    
//     YasrApp.err.println ("five is");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId (5)));


//     YasrApp.err.println ("and 4 is");
//     YasrApp.err.println (model.getNodeSet ().get (new Node.NodeId (4)));
      


    MvSet initValue = initNode.getValue ();
    
    initValue = initValue.
      existAbstract (model.getPreCube ());
    
    
    

    //     YasrApp.err.println ("PreCube is: ");
    //     YasrApp.err.println (new Node (new Node.NodeId (0),
    // 				   model.getPreCube ()));
    //     YasrApp.err.println ("Extracted: ");
    //     YasrApp.err.println ("Dumping");
    //     YasrApp.err.println (new Node (new Node.NodeId (0), initValue));


    return initValue.getValue ();
  }
  



  
  public AlgebraValue run () throws CTLNodeParserException, 
				    antlr.ANTLRException,
				    FileNotFoundException
  {
  
  
    AlgebraValue value = null;
    NodeSet result = null;
  
    do
      {
	YasrApp.out.println ("Compiling C program: " + cFile);
	if (refiner != null)
	  {
	    YasrApp.out.println ("Old predicates: " + 
				 refiner.getPredicates ());
	    YasrApp.out.println ("New predicates: " + 
				 refiner.getNewPredicates ());
	  }
	
	compileCProgram ();
      

	// -- force garbage collection 
	System.gc ();

	// -- convert the property to checkable form
	if (prop instanceof CTLEGNode)
	  {
	    prepareEGProperty ();  

	    // -- construct the model-checker
	    yasnt = new NonTerminationChecker (model);
	  }
	else
	  {
	    prepareProperty ();	      

	    // -- construct the model-checker
	    yasr = new ReachabilityChecker (model);
	  }
    

	

	// -- do the actual model-checking
	YasrApp.out.println ();
	YasrApp.out.println ();
	YasrApp.out.println ("***** Model-Checking ***** ");

	StopWatch sw = new StopWatch ();



	// -- convert the property to checkable form
	if (prop instanceof CTLEGNode)
	  {
	    yasnt.doNonTermination ();
	  }
	else
	  {
	    yasr.doReachablity ();
	  }
    



	YasrApp.out.println ("Done in: " + sw);


	// -- finally, print what we think is the result in the 
	// -- initial state
	
	// -- Print  " \forall {s \in States} INIT(s) -> result(s) "
	// --
	value = extractValue ();	

	// -- old print-out of result 

	// 	YasrApp.out.println ();
	// 	YasrApp.out.println ("******* SEE RESULT BELOW ***********");
	// 	YasrApp.out.println ("result is: " + value);
	// 	YasrApp.out.println ("********** LOOK UP  ^ **********");
	
	// 	//YasrApp.err.println (model.getNodeSet ().dump ());
	
	// 	YasrApp.out.println ("Predicates used: ");
	// 	YasrApp.out.println (refiner.getPredicates ());

	// -- new print-out of result

	YasrApp.out.println ();
	YasrApp.out.println ("MODEL-CHECKING");
	YasrApp.out.println ("\tPROPERTY: " + prop);
	YasrApp.out.println ("\tRESULT: " + value);

	YasrApp.out.println ("Predicates used: ");
		
	YasrApp.out.println (refiner.getPredicates ());
	YasrApp.out.println ();


	
	// -- stop if asked to run only once
	if (!isDoProof ())
	  break;
	  
	found = false;
	if (value.equals (algebra.infoBot ()))
	  {
	    YasrApp.out.println ("Unknown result, starting refinement");
	    if (prop instanceof CTLEGNode)
	      doNTRefine ();
	    else
	      doRefine ();
	  }

      } while (found);

    return value;
  
  } // end of run


  // -- non-termination refine
  void doNTRefine ()
  {
    ReachabilityProver prover = initNTProver (algebra, model.getEdgeSet (),
					      model.getPreCube ());
    int initId = pProgram.getStmt ("init").getId ();
    Node initNode = model.getNodeSet ().get (new Node.NodeId (initId));
    ProofGoal rootGoal = new EGPathGoal (null, 
					 model.getNodeSet (),
					 algebra.infoBot (),
					 initNode);
    
    pathForEG = new ArrayList ();
    statesOnPathForEG = new HashMap ();
    
    findCycle (prover, rootGoal);
    
  }
  

  void doRefine ()
  {
    ReachabilityProver prover = initProver (algebra, 
					    model.getEdgeSet (),
					    model.getPreCube ());
    
    int initId = pProgram.getStmt ("init").getId ();
    Node initNode = model.getNodeSet ().get (new Node.NodeId (initId));

    ProofGoal rootGoal = new PathProofGoal (null, 
					    model.getNodeSet (),
					    algebra.infoBot (),
					    initNode);

    findPath (prover, rootGoal);
					    
  }
  

  
  void findCycle (ReachabilityProver prover, ProofGoal goal)
  {
    // -- the bound can be removed, currently set it a very 
    // -- big number
    findCycle (prover, goal, 10000);
  }
  
  void findCycle (ReachabilityProver prover, ProofGoal goal, 
		  int bound)
  {

    // -- save information for EG path
    if(goal instanceof StateProofGoal)
      {	
	pathForEG.add (goal);


	// -- check if the same state has been visited before
	// -- if so, we have found a cycle
	// -- otherwise, add this states to the state set

	Node.NodeId id =  ( (StateProofGoal)goal).getState ().getId ();
	MvSet state = ( (StateProofGoal)goal).getState ().getValue ();

	if (statesOnPathForEG.containsKey (id))
	  {
	    Set stateSet = (HashSet) statesOnPathForEG.get (id);

	    if (!stateSet.add (state))
	      {
		System.err.println ("\n >>>>> A Cycle is detected !!!\n");
		
		printEGPath ();
		return;
	      }
	    else
	      {
		statesOnPathForEG.put (id,stateSet);
	      }
	  }
	else
	  {
	    Set stateSet = new HashSet ();
	    stateSet.add (state);
	    statesOnPathForEG.put (id,stateSet);
	  }


	// -- check if we have found a node whose EG value is true
	if (! ((StateProofGoal)goal).getValue ().equals (algebra.infoBot ())) 
	  {

	    System.err.println ("\n >>>>> Found a EG(true)=true node !!!\n");

	    printEGPath ();
	    return;
	  }
	else
	  {
	    if (bound == 0) 
	      {

		System.err.println 
		  ("\n >>>>> Reach the bound for detecting EG path !!!\n");

		printEGPath ();
		return;
	      }
	  }


      }

    YasrApp.out.println ("\n*************************************" +
			 "**********************");

    if (goal instanceof StateProofGoal)
      {
	YasrApp.out.println ("Looking at StateProofGaol " + goal);
	YasrApp.out.println ("\t" + 
			     pProgram.getStmt (((StateProofGoal)goal).
					       getState ().getId ().
					       intValue ()));
      }
    else
      {
	if (goal instanceof EdgeProofGoal)
	  YasrApp.out.println ("Looking at EdgeProofGoal " + goal);
	else
	  {
	    throw new RuntimeException ("Wrong type of proof goal" + goal);
	  }
      }
    
    
    
    prover.expand (goal);
    ProofGoal[] subGoals = goal.getSubGoals ();
    if (subGoals == null) 
      {
	// -- hit a point when no proof rules are applicable
	// -- see if we can refine
	if (goal instanceof EdgeProofGoal)
	  {
	    if (((EdgeProofGoal) goal).getValue ().equals (algebra.infoBot ()))
	      {
		YasrApp.out.println ("Edge to refine: " + goal);
		refineEdge ((EdgeProofGoal) goal);
		return;
	      }

	    // -- else : true edge; no refinement is needed 
	  }
      }
    else
      {
	for (int i = 0; i < subGoals.length && !found; i++)
	  findCycle (prover, subGoals [i], bound - 1);
      }


  }
  
  // -- ugly way to print EG path 
  void printEGPath ()
  {
    
    if (pathForEG.size () == 0)
      return;

    System.err.println ("\n EG path <<<<<<<<<<<<<<<<<<<");
    
    for (Iterator it = pathForEG.iterator (); it.hasNext ();)
      {
	StateProofGoal goal = (StateProofGoal) it.next ();
	
	System.err.println ("\nNode  : " + goal.getState ().getId () +
			    "    (EG true) Value : " + goal.getValue ());
	

	CTLNode[] nodeState = 
	  model.getStatePresenter ().toCTL (goal.getState ().getValue ());

	System.err.println ("Predicate Assignment  : " + 
			    Arrays.asList (ctlToExpr (nodeState)));

	ProofGoal[] subGoals = goal.getSubGoals ();
	if (subGoals != null) 
	  {
	    // -- print the edge information, assume the EdgeProofGoal
	    // -- is always the second subgoal
	    if (subGoals[1] instanceof EdgeProofGoal)
	      {
		System.err.println ("\nby edge : " + 
				    (EdgeProofGoal) subGoals[1]);
	      }
	    else
	      {		
		throw new RuntimeException ("Woring subgoal type!");
	      }
	  }    
      }

    System.err.println ("\n >>>>>>>>>>>>>>>>>>>");
                          
    pathForEG = new ArrayList ();
    statesOnPathForEG = new HashMap ();
    return;
  }
  
  
  String[]  ctlToExpr (CTLNode [] ctl)
  {
    String [] result = new String [ctl.length];
    
    for (int i=0; i < ctl.length; i++)
      result[i] = ctlToExpr (ctl[i]);
    
    return result;
  }

  String ctlToExpr (CTLNode ctl)
  {
    String ctlName = ctl.getLeft ().toString ();
    String value = ctl.getRight ().toString ();

    String result = 
      pProgram.getPredTable ().
      getByCtlName (ctlName).getName ().toString ();
    
    if (value.equals ("false"))
      result = "!" + result;
    if (value.equals ("maybe"))
      result = "?" + result + "?";
    return result;
  }
  
  
  void findPath (ReachabilityProver prover, ProofGoal goal)
  {
    if (goal instanceof StateProofGoal &&
	(! ((StateProofGoal)goal).getValue ().equals (algebra.infoBot ())))
      {
	return;
      }
//     else if (goal instanceof EdgeProofGoal)
//       {
// 	YasrApp.out.println ("Edge to refine: " + goal);
// 	if (((EdgeProofGoal)goal).getValue ().equals (algebra.infoBot ()))
// 	  refineEdge ((EdgeProofGoal) goal);
// 	//found = true;
// 	return;
//       }
    
    YasrApp.out.println ("Looking at " + goal);
    if (goal instanceof StateProofGoal)
      {
	YasrApp.out.println ("\t" + 
			     pProgram.getStmt (((StateProofGoal)goal).
					       getState ().getId ().
					       intValue ()));
      }
      
    prover.expand (goal);
    ProofGoal[] subGoals = goal.getSubGoals ();
    if (subGoals == null) 
      {
	// -- hit a point when no proof rules are applicable
	// -- see if we can refine
	if (goal instanceof EdgeProofGoal &&
	    ((EdgeProofGoal) goal).getValue ().equals (algebra.infoBot ()))
	  {
	    YasrApp.out.println ("Edge to refine: " + goal);
	    refineEdge ((EdgeProofGoal) goal);
	    return;
	  }
      }
    else
      {
	for (int i = 0; i < subGoals.length && !found; i++)
	  findPath (prover, subGoals [i]);
      }
    
  }
  

  void refineEdge (EdgeProofGoal edgeGoal)
  {
    // -- get a statement corresponding to this edge
    PStmt stmt = pProgram.getStmt (edgeGoal.getSrc ().getId ().intValue ());
    if (stmt instanceof IfPStmt)
      refineIfStmt ((IfPStmt) stmt);
    else if (stmt instanceof PrllAsmtPStmt)
      refinePrllAsmt ((PrllAsmtPStmt) stmt, edgeGoal.getSrc (), 
		      edgeGoal.getDst ());
    else if (edgeGoal.getEdge () instanceof SummaryEdge)
      {
	YasrApp.out.println ("Refining summary edge: " + edgeGoal);
	
	SummaryEdge edge = (SummaryEdge) edgeGoal.getEdge ();
	//MvSet sumValue = edge.getRelation ().toMvSet ();
	MvSet sumValue = edgeGoal.getSrc ().getValue ().and 
	  (edgeGoal.getDst ().getValue ().
	   renameArgs (model.getPreToAuxMap ()));
	
	//sumValue = sumValue.renameArgs (model.getPostToAuxMap ());
	
	ProofGoal rootGoal = 
	  new PathProofGoal (edgeGoal, model.getSumNodeSet (),
			     edgeGoal.getValue (),
			     new Node (edgeGoal.getSrc ().getId (),
				       sumValue));
	ReachabilityProver prover = initProver (algebra, 
						model.getEdgeSet (),
						model.getPreCube ().and 
						(model.getAuxCube ()));
	
	findPath (prover, rootGoal);	
      }
    else
      {
	System.err.println ("Not IfPStmt/PrllAsmtPStmt/SummaryEdge:" +
			    "Do nothing");
	
      }
    
    
  }

  void refineIfStmt (IfPStmt ifStmt)
  {
    Expr realCond = ifStmt.getCond ().getOrigCond ();
    ifStmt.getCond ().setCond (realCond);
    
    while (realCond.op () == BoolOp.NOT)
      realCond = realCond.arg (0);
    
    YasrApp.out.println ("Trying to add IF predicate");
    YasrApp.out.println (realCond);
    if (!refiner.getPredicates ().contains (realCond))
      {
	found = true;
	refiner.addPredicate (realCond);
      }
    else
      YasrApp.out.println ("Ooops, have it already");
    
  }
  
  void refinePrllAsmt (PrllAsmtPStmt asmt, Node src, Node dst)
  {
    
    CTLNode[] srcState = model.getStatePresenter ().toCTL (src.getValue ());
    CTLNode[] dstState = model.getStatePresenter ().toCTL (dst.getValue ());
    
    
    System.out.println ("refining between states:");
    System.out.println ("src state: " + Arrays.asList (srcState));
    System.out.println ("dst state: " + Arrays.asList (dstState));
      
    PredicateGenerator pg = createPredicateGenerator (asmt, srcState, 
						      dstState);
    
    List newPred = null;
    if (pg.find  ())
      {
	newPred = pg.getNewPreds ();
	found = true;
      }
    if (newPred != null && !newPred.isEmpty ())
      {
	for (Iterator it = newPred.iterator (); it.hasNext ();)
	  refiner.addPredicate ((Expr) it.next ());
      }
  }
  
  

  public ReachabilityProver initNTProver (BelnapAlgebra algebra, 
					  EdgeSet edges, 
					  MvSet stateCube)
  {
    ReachabilityProver prover = new ReachabilityProver ();
    prover.addRule (new EGStepRule (algebra, edges, stateCube));
    prover.addRule (new FunctionSummaryRule (algebra, stateCube));
    return prover;
    
  }
  

  public ReachabilityProver initProver (BelnapAlgebra algebra, 
					EdgeSet edges,
					MvSet stateCube)
  {
    ReachabilityProver prover = new ReachabilityProver ();
    prover.addRule (new PathRule (algebra, stateCube));
    prover.addRule (new SimpleMinPathRule (algebra));
    prover.addRule (new StepRule (algebra, edges, stateCube));
    prover.addRule (new FunctionSummaryRule (algebra, stateCube));
    return prover;
  }


  PredicateGenerator createPredicateGenerator (Object stmt, 
					       CTLNode[] parentState, 
					       CTLNode[] childState)
  {
    PredicateTable pTable = pProgram.getPredTable ();

    Map<String,Expr> predMap = new HashMap<String,Expr> ();
    for (Iterator it = pTable.getVariables ().iterator (); it.hasNext ();)
      {
	PredicateTable.Predicate pred = (PredicateTable.Predicate) it.next ();
	predMap.put (pred.getCtlName (), pred.getName ());
      }
    // Map symbolicMap = refiner.getMemoryModel ((PrllAsmtPStmt) stmt);    
    // WPComputer wp = refiner.getWPComputer ((PrllAsmtPStmt) stmt);
    WPComputer wp = ((PrllAsmtPStmt) stmt).getWPComputer ();

    List<Expr> preds = refiner.getPredicates ();
    
    List<Expr> pState = getStatePred (predMap, parentState);
    List<Expr> cState = getStatePred (predMap, childState); 
    
    return findPredicateGenerator (pGeneratorType,  pProgram, 
				   (PrllAsmtPStmt) stmt,
				   refiner, wp, pState, cState);
    
  }

  static PredicateGenerator findPredicateGenerator (String name,
						    PProgram pProgram,
						    PrllAsmtPStmt stmt,
						    Refiner refiner,
						    WPComputer wp,
						    List<Expr> pState,
						    List<Expr> cState)
  {
    ChainPredicateGenerator pg = new ChainPredicateGenerator ();
    
    String[] result = name.split (":");
    for (int i = 0; i < result.length; i++ )
      {
	String generatorName = result[i];
	if (generatorName.equals ("wpd"))
	  {
	    pg.add (new WPPredicateGenerator.WPDiff 
		    (pProgram,
		     (PrllAsmtPStmt)stmt,
		     refiner, wp, pState, cState));
	  }
	else if (generatorName.equals ("wps"))
	  {
	    pg.add (new WPPredicateGenerator.WPSame 
		    (pProgram, 
		     (PrllAsmtPStmt)stmt,
		     refiner, wp, pState, cState));
	  }
	else if (generatorName.equals ("wpa"))
	  {
	    pg.add (new WPPredicateGenerator.WPAll
		    (pProgram, 
		     (PrllAsmtPStmt)stmt,
		     refiner, wp, pState, cState));
	  }
	else if (generatorName.equals ("bend"))
	  {
	    pg.add (new BackEndGenerator 
		    (pProgram, 
		     (PrllAsmtPStmt) stmt,
		     refiner, wp, pState, cState));
	  }
	else 
	  throw new NoSuchElementException ("Unknow Predicate Generator in "
					    + name + " of " + generatorName);
      }
    return pg;
  }

  private List<Expr> getStatePred (Map<String,Expr> predMap, CTLNode[] state)
  {
    List<Expr> vExpr = new ArrayList<Expr>();
    for (int i = 0; i < state.length; i++)
      {
	Expr leftChild = predMap.get (state [i].getLeft ().toString ());

	if (state [i].getRight ().toString ().equals ("false"))
	  vExpr.add 
	    (refiner.exprFactory ().op (BoolOp.NOT).unaryApply (leftChild));
	else
	  vExpr.add (leftChild);
      }
    return vExpr;
  }
  

  public void parseCmdLine (String[] args)
    throws CTLNodeParserException
  {
  OptParser optParser = new OptParser ().startClass (YasrApp.class);

  BooleanOpt runOnceOpt = optParser.opt ().longName ("once").
    description ("does not go into the refinement loop").argumentArity (0).
    asBoolean ().defaultValue (false);
  StringOpt ctlPropOpt = 
    optParser.opt ().longName ("property").shortName ('p').
    description ("property to be checked").required ().
    asString ();

//   BooleanOpt cFile = optParser.opt ().longName ("pfile").
//     shortName ('c').
//     description ("input file is assumed to be a predicate program").
//     asBoolean ().defaultValue (true);

  StringOpt refinerOpt = optParser.opt ().longName ("refiner").
    description ("Refiner to be used. Choices are:\n" + 
		 //"\told -- complete slow search\n" + 
		 "new -- incremental full search\n" + 
		 "csp1 -- first try at CSP based search\n" + 
		 "cbj -- conflict-driven back jumping with prunning\n" + 
		 "local-cbj-vs -- LocalCBJPredicateRefinerVisitor\n" +
		 "cbj-i -- as cbj but don't check inconsistencies\n" + 
		 "cbj-s -- as cbj but don't abstract statements\n" + 
		 "cbj-s-i -- as cbj but only use cheap heuristics\n" + 
		 "cffc -- forward check with back jumping with prunning\n"+ 
		 "cffc-i -- as cffc but don't check inconsistencies\n" + 
		 "cffc-s -- as cffc but don't abstract statements\n" + 
		 "cffc-s-i -- as cffc but only use cheap heuristics\n" + 
		 "cffc-d -- as cffc but no use all predicates\n" +
		 "cffc-s-i-d -- take a guess\n" +
		   "\tDEFAULT IS local-cbj-vs\n")
    .asString ().defaultValue ("local-cbj-vs");

  StringOpt pGeneratorOpt = optParser.opt ().longName ("pg").
    description ("The order in which predicate generators are used\n " + 
		   "  separated by colon (':'). The choices are:\n" + 
		   "wpd -- weakest precondition based on difference\n" + 
		   "wps -- weakest precondition based on similarities\n" + 
		   "bend -- backend generator that solves\n" + 
		   "\tincompleteness of the refiner\n" + 
		   "\tDEFAULT IS bend:wpd:wps\n").asString ().
    defaultValue ("bend:wpd:wps:wpa");

  BooleanOpt hyperOpt = optParser.opt ().longName ("nohyper").
    description ("disables hyper edges").argumentArity (0).
    asBoolean ().defaultValue (false);

  BooleanOpt nonDetInitOpt = optParser.opt ().longName ("ndinit").
    description ("if present initial values of predicates are " + 
		 "assumed to be non-deterministic").
    argumentArity (0).asBoolean ().defaultValue (false);

  BooleanOpt inlineOpt = optParser.opt ().longName ("inline").
    description ("Inline function calls").
    argumentArity (0).asBoolean ().defaultValue (false);
    
  FileNameOpt initPredOpt = optParser.opt ().longName ("init-pred").
    description ("A file containing intial predicates").asFileName ();
  FileNameOpt inputFileOpt = 
    optParser.opt ().nameless ().required ().
    description ("boolean program").asFileName ();

  StringOpt selectorOpt = optParser.opt ().longName ("sel").
    description ("Return selector type. The choices are:\n" +
      "dd -- Decision diagram selector\n" +
      "bin -- Binary search PProgram selector\n" +
      "lin -- Linear search PProgram selector\n" +
      "\tDEFAULT IS dd\n").
    asString ().defaultValue ("dd");  

  BooleanOpt stmtBlockingOpt = optParser.opt ().longName ("blocking").
    description ("if present, consecutive simple assignments are merged " +
                 "into a basic block\n" +
                 "\tDEFAULT IS false\n").
    argumentArity (0).asBoolean ().defaultValue (false);

  IntOpt maxPredsPerRefinementOpt = optParser.opt ().longName ("maxpreds").
    description ("Max. number of predicates to discover per " +
                 "predicate-generation pass\n" +
                 "\tDEFAULT is 1\n").
    asInt ().defaultValue (1);             
  
  // -- actually parse the command line
  OptResult result = optParser.run (args);

  // -- initialize our class


  setCFile (new File (inputFileOpt.getString (result)));
  setProp (CTLNodeParser.parse (ctlPropOpt.getString (result)));
  setDoProof (!runOnceOpt.getBoolean (result));
  setInline (inlineOpt.getBoolean (result));
  setHyper (!hyperOpt.getBoolean (result));
  setUnknownInit (!nonDetInitOpt.getBoolean (result));


  setRefinerType (refinerOpt.getString (result));
  setPGeneratorType (pGeneratorOpt.getString (result));

  setSelectorType (selectorOpt.getString (result));
  setStmtBlocking (stmtBlockingOpt.getBoolean (result));

  setMaxPredsPerRefinement (maxPredsPerRefinementOpt.getInt (result));
  
  String initPredFileName = initPredOpt.getString (result);
  if (initPredFileName != null)
    setInitPredFile (new File (initPredOpt.getString (result)));
  else
    setInitPredFile (null);
} // end of parseCmdLine

 
  public static void main (String[] args)
{
  try
  {

    // -- create our application
    YasrApp yasmApp = new YasrApp ();

    // -- parse command line option
    yasmApp.parseCmdLine (args);


    System.err.println ("max preds: " + MAX_PREDICATES_PER_REFINEMENT);

    AlgebraValue value = yasmApp.run ();

    YasrApp.out.println ("Done in " + yasmApp.overall);

//     YasrApp.out.println ("QUERY: " + yasmApp.refiner.getVC ().sw);


    YasrApp.out.println ();
    YasrApp.out.println ();
    YasrApp.out.println ();
    YasrApp.out.println ("FINAL RESULT");
    YasrApp.out.println ("\tPROPERTY: " + yasmApp.getProp ());
    YasrApp.out.println ("\tRESULT: " + value);
    

    YasrApp.out.println ("Predicates found: ");
    YasrApp.out.println (yasmApp.refiner.getPredicates ());



  }
  catch (CTLNodeParserException ex)
  {
    YasrApp.out.println ("Could not parse CTL property: " + ex);
    ex.printStackTrace (YasrApp.err);
  }
  catch (antlr.ANTLRException ex)
  {
    YasrApp.out.println ("Could not parse C program: " + ex);
    ex.printStackTrace (YasrApp.err);
  }
  catch (FileNotFoundException ex)
  {
    YasrApp.out.println ("Could not find C file: " + ex);
    ex.printStackTrace (YasrApp.err);
  }
  catch (Throwable ex)
  {
    YasrApp.out.println ("Unhandled Exception: " + ex);
    ex.printStackTrace ();
  }

  // -- flush all our streams

  YasrApp.out.flush ();
  YasrApp.err.flush ();
} // end of main

  
  
} // YasrApp
