package edu.toronto.cs.yasm.abstractor;

import edu.toronto.cs.expr.*;

import java.util.*;

/** 
 * Extracts the predicates defined in an initial-predicates Expr tree, which is
 * like a tree for a C program, but is missing variable declarations, only
 * requires function names (not their signature) to be declared, only contains
 * boolean expressions as statements, and has no global (unscoped) statements.
 * <p>
 * An initial predicates spec looks like:
 * <code>
 * main
 * {
 *   x &gt; 0;
 *   y == 2;
 *   z;
 *   v &amp;&amp; w;
 * }
 * </code>
 * where the identifiers @{code v, w, x, y, z} refer to variables declared (in
 * the target program) at the top scope in main.
 *
 * To specify predicates over global variables, just declare a scope with a
 * name that doesn't match any function name in the target C program, e.g.:
 * <code>
 * __GLOBAL__
 * {
 *   global_lock == 0;
 * }
 * </code>
 *
 * XXX This all works out because we've kludged GnuCParser to prepend all
 * identifiers with their scope name.
 *
 * XXX Doesn't do type checking or declaration checking.
 * 
 * @author Kelvin Ku (kelvin@cs.toronto.edu)
 * @version 
 */
public class PredicateExtractor
{
  public static List<Expr> extract (Expr program)
  {
    List<Expr> predicates = new LinkedList<Expr> ();

    for (Iterator it = program.args ().iterator (); it.hasNext ();)
    {
      Expr e = (Expr) it.next ();
      if (e.op () == CILFunctionDefOp.FUNCTION_DEF)
      {
        predicates.addAll
          (CILScopeOp.getBody (CILFunctionDefOp.getBody (e)).args ());
      }
    }

    return predicates;
  }
}

