package edu.toronto.cs.yasm.abstractor;

import edu.toronto.cs.cparser.block.*;
import edu.toronto.cs.cparser.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.*;
import edu.toronto.cs.yasm.pprogram.*;

import edu.toronto.cs.tp.*;
import edu.toronto.cs.tp.cvcl.ValidityChecker;

import edu.toronto.cs.boolpg.abstraction.StmtAbstraction;

import edu.toronto.cs.yasm.wp.*;

import antlr.*;

import java.util.*;
import java.io.*;

/**
 * Describe class RefiningVisitor here.
 *
 *
 * Created: Tue Jul 19 15:52:12 2005
 *
 * @author <a href="mailto:owei@epoch.cs">Ou Wei</a>
 * @version 1.0
 */
public class PredicateRefinerVisitor extends BasePProgramVisitor implements Refiner
{

  public edu.toronto.cs.yasm.YasmStatistics stats;

  List predicates; // e.t.c.expr.Expr's
  List newPredicates; // e.t.c.expr.Expr's

  ExprFactory fac;
  // -- our theorem prover
  JCVCLTheoremProver tp;

  // DeclsParser declsParser;
  // SymbolicExecutor symbolicExecutor;

  // -- marks whether a PProgram component has been refined already
  Map isRefined;

  Expr unknownCond;

  public PredicateRefinerVisitor (ExprFactory _fac)
  {
    this (_fac, new ArrayList ());
  }

  public PredicateRefinerVisitor (ExprFactory _fac, List _newPredicates)
  {
    fac = _fac;

    predicates = new ArrayList ();
    newPredicates = new ArrayList ();
    addToNewPredicates (_newPredicates);

    tp = new JCVCLTheoremProver (fac);

    isRefined = new IdentityHashMap ();
  }


  // -- ***** methods for implementing Refiner interface *****


  /**
   * Refines a predicate program with new predicates
   *
   * @param p an input predicate program
   * @return a refined version of <code>p</code>
   */
  public PProgram doProgramRefine (PProgram p)
  {
    return (PProgram) p.accept (this, null);
  }
  

  public void doDeclsRefine (PProgram p)
  {
    // -- process global declarations
    for (Iterator it = p.getGlobalDecls ().values ().iterator ();
        it.hasNext ();)
      doDeclRefine ((PDecl) it.next ());

    // -- process local declarations
    for (Iterator it = p.getFunctionDefs ().values ().iterator ();
        it.hasNext ();)
      doFunctionDeclsRefine ((PFunctionDef) it.next ());

    // -- create an unknown variable not used anywhere else
    Expr unknownVar = fac.var ("__UNKNOWN__");

    // -- tell our theorem prover that this is an integer variable
    tp.declare (unknownVar, fac.op (SimpleTypeOp.INT));

    // -- make unknown if-condition -- a condition that can never
    // -- be shown to be neither true nor false
    unknownCond = fac.op (ComparisonOp.EQ).binApply
      (unknownVar, fac.intExpr (0));

    p.setDeclsRefined (true);    
  }

  public void doFunctionDeclsRefine (PFunctionDef f)
  {
    // -- process local declarations
    for (Iterator it = f.getLocalDecls ().values ().iterator ();
        it.hasNext ();)
      doDeclRefine ((PDecl) it.next ());

    // -- declare return index
    tp.declare 
      (f.getReturnIndexVar (), tp.exprFactory ().op (SimpleTypeOp.INT));

    Expr returnType;


    // -- create return value Expr
    try
    {
      returnType = f.getReturnType (tp);
    }
    catch (RecognitionException ex)
    {
      ex.printStackTrace ();
      throw new RuntimeException
        ("Error parsing return type for function " + f.getFunctionName ());
    }

    // -- declare return value
    tp.declare (f.getReturnValueVar (),
	tp.exprFactory ().op (SimpleTypeOp.INT));
  }

  /**
   * Declares the variable corresponding to this declaration with the
   * theorem prover.
   *
   * @param pDecl a <code>PDecl</code> value
   */
  public void doDeclRefine (PDecl pDecl)
  {
    pDecl.declare (tp);
  }

  /**
   * Sets a stats object to collect statistics
   *
   * @param stats an <code>YasmStatistics</code> value
   */
  public void setStats (YasmStatistics _stats)
  {    
    stats = _stats;    
  }


  /**
   * return predicates used for refinement
   *
   * @return a <code>List</code> value
   */
  public List getPredicates ()
  {
    return predicates;
  }


  /**
   * retunre new predicates used for refinment
   *
   * @return a <code>List</code> value
   */
  public List getNewPredicates ()
  {
    return newPredicates;
  }


  public ExprFactory exprFactory ()
  {
    return tp.exprFactory ();
  }
  
  /**
   * Returns a copy of {@code ITheoremProver} used by this refiner
   *
   * @return an instance of <code>ITheoremProver</code> 
   */
  public ITheoremProver getTP ()
  {
    return tp;
  }
  
  public ValidityChecker getVC ()
  {
    return tp.getValidityChecker ();
  }

  public CVCLExprConverter getCVCLConverter ()
  {
    return tp.getExprConverter ();
  }

  /**
   * add new predicates to predicate list
   *
   * @param l a <code>List</code> value
   */
  public void addToNewPredicates (List l)
  {
    isRefined = new IdentityHashMap (); // all nodes to be refined again

    for (Iterator it = l.iterator (); it.hasNext ();)
      addToNewPredicates ((Expr) it.next ());
  }

  /**
   * add new predicates to predicate list
   *
   * @param p an <code>Expr</code> value
   */
  public void addToNewPredicates (Expr p)
  {
    isRefined = new IdentityHashMap (); // all nodes to be refined again

    // -- only add predicates which are not already in newPredicates
    if (!newPredicates.contains (p))
      newPredicates.add (p);
    else
      System.err.println ("Refiner: addToNewPredicates: Not adding " +
                          "existing predicate: " + p);
  }

  /** 
   * @deprecated Use addToNewPredicates
   */
  public void addPredicates (List l)
  {
    addToNewPredicates (l);
  }

  /** 
   * @deprecated Use addToNewPredicates
   */
  public void addPredicate (Expr p)
  {
    addToNewPredicates (p);
  }


  /**
   * check if there are new predicates for refinement
   *
   * @return a <code>boolean</code> value
   */
  public boolean hasNewPredicates ()
  {
    return !newPredicates.isEmpty ();
  }



  /** 
   * Parses stream read from Reader; each predicate parsed is added to
   * the list of predicates.
   * 
   * @param in Source of predicates.
   */
  public void readPredicates (Reader in) 
    throws RecognitionException, TokenStreamException 
    {
      CILLexer lexer = new CILLexer (in);
      lexer.setTokenObjectClass (CToken.class.getName ());

      PredicateParser parser = new PredicateParser (lexer);
      parser.setASTNodeClass (Block.class.getName ());
      Block.setTokenVocabulary (CILTokenTypes.class.getName ());

      parser.parse();

      Block cPred = (Block) parser.getAST ();
      while (cPred != null)
      {
        /*
           System.err.println ("attempting to parse predicate:");
           TNode.setTokenVocabulary (CILTokenTypes.class.getName ());
           TNode.printTree (cPred);
         */

        Expr pred = SymbolicExecutorUtil.newSymbolicExecutor (fac).pureExpr
                    (cPred, new MemoryModel ());
        System.err.println ("Adding predicate: " + pred);
        addToNewPredicates (pred);

        cPred = (Block) cPred.getNextSibling ();
      }
    }


  /**
   * return CVCL Memory Model
   *
   * @param asmt a <code>PrllAsmtPStmt</code> value
   * @return a <code>CVCLMemoryModel</code> value
   */
  public CVCLMemoryModel getMemoryModel (PrllAsmtPStmt asmt)
  {
    return (CVCLMemoryModel) asmt.getRefinerInfo ();
  }


  /**
   * retrun WP computer used for refinement
   *
   * @param asmt a <code>PrllAsmtPStmt</code> value
   * @return a <code>WPComputer</code> value
   */
  public WPComputer getWPComputer (PrllAsmtPStmt asmt)
  {
    return getPredRefinerInfo (asmt).wp;
  }

  private PredicateRefinerInfo getPredRefinerInfo (PrllAsmtPStmt asmt)
  {
    PredicateRefinerInfo info = (PredicateRefinerInfo) asmt.getRefinerInfo ();
    if (info == null)
    {
      info = new PredicateRefinerInfo ();
      asmt.setRefinerInfo (info);
    }
    return info;
  }

  /**
   * create Memory Model
   *
   * @param s a <code>PrllAsmtPStmt</code> value
   * @return a <code>MemoryModel</code> value
   */
  public MemoryModel computeRegularMemoryModel (PrllAsmtPStmt s)
  {
    assert false : "Use getWPComputer!";

    return null;

  }

  /**
   * create CVCL memory model
   *
   * @param s a <code>PrllAsmtPStmt</code> value
   * @return a <code>CVCLMemoryModel</code> value
   */
  public CVCLMemoryModel computeMemoryModel (PrllAsmtPStmt s)
  {
    assert false : "Use getWPComputer!";
    return null;
  }


  class PredicateRefinerInfo 
  {
    CVCLMemoryModel mmModel;
    WPComputer wp;
  }



  // -- *****  methods from BasePProgramVisitor *****



  /**
   * refine the block of predicate statements starting from 
   * thethe statement <code>s</code>.
   *
   * @param s a <code>PStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  private Object refinePStmtBlock(PStmt s, Object o)
  {

    PStmt head;
    
    head = s;
    

    while (s != null)
      {

	if (!isRefined.containsKey (s))
	  {
	    // -- have not refined this node on this refinement pass; remember 
	    // -- that we've visited this node
	    isRefined.put (s, s);
	    s.accept (this, o);
	  }
	else
	  {
	    // -- node was refined on this pass already
	    System.out.println ("Saw " + s + " twice");
	  }

	s = s.getNext ();
      }    

    return head;

  }
  
  /**
   * refine a predicate program <code>p</code>
   * <p>
   * the declarations of the program should have been 
   * "refined" by calling <code>doDeclsRefine</code>, 
   * i.e., all the variables have been registered in
   * the Theorem Prover (validity checker)
   * <p>
   * @param p a <code>PProgram</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitPProgram (PProgram p, Object o)
  {

    assert p.isDeclsRefined () == true
      : "Decls not refined; cannot refine PProgram";


    // -- refine each function
    for (Iterator it = p.getFunctionDefs ().values ().iterator (); 
	 it.hasNext (); )
      {
	((PFunctionDef) it.next ()).accept (this, o);
      }

    // -- add new predicates to PProgram
    for (Iterator it = newPredicates.iterator (); it.hasNext(); )
      p.getPredTable ().newPredicate ((Expr) it.next ());

    // -- add new predicates to Refiner (this)
    predicates.addAll (newPredicates);

    // -- reset new predicates
    newPredicates = new ArrayList ();

    return p;

  }
 
 

  /**
   * refine the function refered to by <code>fdef</code>
   *
   * @param fdef a <code>PFunctionDef</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitFunctionDef (PFunctionDef fdef, Object o)
  {

    PStmt s;
    
    s = fdef.getHead ();
    
    // -- refine the body of the function
    refinePStmtBlock(s, o);
 
    return fdef;
    
  }



  /**
   * this method should not be called if all predicate 
   * statments can be been properly handled by other 
   * "visit" methods including 
   * <code>visitIf</code>,  <code>visitPrllAsmt</code>,  
   * <code>visitGoto</code>, <code>visitReturnSelector</code>,
   * <code>visitNDGoto</code>,  <code>visitSkip</code>,  
   *
   * @param s a <code>PStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitPStmt (PStmt s, Object o)
  {
      throw new RuntimeException ("Can't refine: " + s.getClass ());
  }


  // -- returns true if every operator/operand in expr is known
  private boolean knownCond (Expr expr)
  {
    if (expr == null)
      return false;

    if (expr.op () instanceof VariableOp)
      return true;

    if (expr.op () instanceof RationalOp)
      return true;

    if (expr.op () instanceof JavaObjectOp)
      return true;

    if (expr.op () instanceof BoolOp ||  
        (expr.op () instanceof NumericOp &&
        expr.op () != NumericOp.MOD) ||
        expr.op () instanceof ComparisonOp ||
	expr.op () instanceof CILIndirectionOp)
    {
      for (int i = 0; i < expr.arity (); i++)
        if (!knownCond (expr.arg (i)))
          return false;
      return true;
    }
    
    return false;  
  }


  /**
   * refine a condition expression refered to by <code>e</code> and
   * return the new, refined condtion
   *
   * @param e a <code>PCond</code> value
   * @param o an <code>object</code> value
   * @return a <code>PCond</code> value
   */
  private PCond refinePCond(PCond e, Object o)
  {
    // System.err.println ("doCondRefine: " + e);

    // -- check that condition is as refined as can be already
    // -- XXX currently, each condition can be refined at most once
    // -- XXX at which point it's value changed from IBOT to something 
    // -- XXX else. 
    if (e.getCond ().op () != BiLatticeOp.IBOT)
      return e;

    // -- NOTE: YasmApp.preRefine now sets cond's abstracted condition to the
    // if-condition itself

    try
    {
      Expr sourceExpr = (Expr) e.getOrigCond ();

      // -- unable to convert to cvcl, e.g. contains an unknown operator
      if (!knownCond (e.getOrigCond ()))
	e.setOrigCond (unknownCond);
      

      sourceExpr = e.getOrigCond ();  

      // -- check if the condition is a constant
      Boolean result = tp.quickCheckTruth (sourceExpr);

      // -- result != null iff original condition is a constant
      if (result != null)
      {
        e.setCond 
          (result == Boolean.TRUE ? fac.trueExpr () : fac.falseExpr ());
        return e;
      }

      // -- check if condition is negated and strip the negation
      boolean negated = sourceExpr.op () == BoolOp.NOT;
      Expr condExpr = negated ? sourceExpr.arg (0) : sourceExpr;

      // -- see if the condition is equivalent to one of our 
      // -- new predicates
      for (Iterator it = newPredicates.iterator (); it.hasNext ();)
      {
        Expr pred = (Expr) it.next ();
        if (condExpr == pred /* || 
            CVCLUtil.quickQuery (vc, 
              vc.iffExpr (cvcl.toCVCL (condExpr),
                cvcl.toCVCL (pred))) */)
        {
          // -- got a new predicate equivalent to the condition
          e.setCond (negated ? 
              fac.op (BoolOp.NOT).unaryApply (pred) : pred);
          return e;
        }

      }
    }
    catch (Exception ex)
    {
      ex.printStackTrace ();
      throw new RuntimeException (ex);
    }
    return e;

  }


  /**
   * refine <code>IfPStmt</code>
   * 
   * @param s an <code>IfPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitIf (IfPStmt s, Object o)
  {
 
    // System.err.println (">>> doIfRefine " + s.getLabel ());
    
    // - refine the condtion of the IfPStmt
    s.setCond (refinePCond(s.getCond (), null));

    // -- refine the ThenStmt 
    refinePStmtBlock (s.getThenStmt (), null);

    // -- refine the ElseStmt 
    refinePStmtBlock (s.getElseStmt (), null);
    
    return s;
    
  }



  /**
   * refine <code>GotoPStmt</code>
   *
   * @param s a <code>GotoPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitGoto (GotoPStmt s, Object o)
  {

    // -- do nothing

    return s;
     
  }
  


  /**
   * refine <code>PrllAsmt</code>
   *
   * @param s a <code>PrllAsmtPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitPrllAsmt (PrllAsmtPStmt s, Object o)
  {
    assert false : "Shouldn't be using base refiner!";
    return null;    
  }
  


  /**
   * refine <code>SkipPStmt</code>
   *
   * @param s a <code>SkipPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitSkip (SkipPStmt s, Object o)
  {
   
    // -- do nothing

    return s;
     
  }

  /**
   * refine <code>BreakPStmt</code>
   * 
   * @param s a <code>BreakPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitBreak (BreakPStmt s, Object o)
  {

    // -- do nothing

    return s;
        
  }



  /**
   * refine <code>visitReturnSelector</code>
   *
   * @param r a <code>ReturnSelectorPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitReturnSelector (ReturnSelectorPStmt r, Object o)
  {
    // -- do nothing

    return r;
  }



  //*****************************************

  // the code bellow is commented out since it may not compile
//   public static void main (String[] args)
//   {
//     // -- create the refiner
//     PredicateRefinerVisitor refiner = new PredicateRefinerVisitor ();

//     // -- assume that we have some pProgram
//     PProgram pProgram;
    
//     // -- this recursivelly refines the whole program
//     pProgram = refiner.accept (pProgram);
    
//     // -- assume that we have some function that we are interested in
//     PFunctionDef fun;
//     // -- this refines a single function
//     fun = refiner.accept (fun);

//     /*
//      * With the above, we have the following interface
//      * Given a PredicateRefinerVisitor refiner and a program element 'foo'
//      * refiner.accept (foo)
//      * returns a refined version of 'foo'
//      *
//      * We should also move declaration part into a separate visitor since
//      * we only run it once at the very beginning of the whole process.
//      */
    
//   }
  
  
}


