package edu.toronto.cs.yasm.eqsolver;

import java.util.*;

/**
 * A representation of a fixpoint equational system. An equational
 * system is a sequence of equations. Each equation is of the form
 * <code> m X = f </code>, where {@code m} is either greatest or least
 * fixpoint modality, and {@code X = f} an equation.
`*
 *
 * Created: Fri Jun  1 09:42:37 2007
 *
 * @author <a href="mailto:arie@pcawf.sei.cmu.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public class EqSystem<V>
{

  /**
   * Fixpoint modalities.
   */
  public enum FixPtModality {
    /** Least Fixed Point */
    LFP, 
      /** Greatest Fixed Point */
      GFP, 
      /** An unspecified fixed point modality */
      ANY};



  /**********************************************************************/
  /* INTERNALLY USED FIELDS */
  /**********************************************************************/

  /**
   * A map from variables to information associated with them.
   */
  protected Map<Integer,VarInfo> varInfo;

  /**
   * The smallest id used so far
   */
  protected int leastId = 0;
  /**
   * The largest id used so far
   */
  protected int greatestId = 0;
  /**
   * A variable that does not yet appear in the system
   */
  protected int freshVar = 0;
  /**
   * A set of all free variables. A variable is free if it is used on
   * the rhs of some equation, but never appears on the lhs.
   */
  protected Set<Integer> freeVars;
  
  
  /**
   * Creates an empty equation system
   *
   */
  public EqSystem () 
  {
    varInfo = new TreeMap<Integer,VarInfo> ();
    freeVars = new TreeSet<Integer> ();
  }


  /**
   * Adds an equation to the beginning of the current system.
   * 
   * @param m fixpoint modality of the added equation
   * @param eq the equation to be added
   */
  public void addFirst (FixPtModality m, Equation<V> eq)
  {
    newVarInfo (--leastId, m, eq);
  }

  /**
   * Adds an equation to the end of the current system.
   * 
   * @param m fixpoint modality of the added equation
   * @param eq the equation to be added
   */  
  public void addLast (FixPtModality m, Equation<V> eq)
  {
    newVarInfo (++greatestId, m, eq);
  }
  
  /**
   * Returns an equation for a variable
   *
   * @param var a varaible for which an equation is thought
   * @return an equation for {@code var}, or null if there is no such
   * equation
   */
  public Equation<V> eqForVar (int var)
  {
    VarInfo vInfo = varInfo.get (var);
    return vInfo != null ? vInfo.getDefEq () : null;
  }

  public FixPtModality varModality (int var)
  {
    VarInfo vInfo = varInfo.get (var);
    return vInfo != null ? vInfo.getFpMod () : FixPtModality.ANY;
  }
  
  
  /**
   * Returns the set of all free variables of this system. A variable
   * is free if it is used on the rhs of some equation, but never
   * appears on the lhs.
   *
   * @return the set of all free variables of this system.
   */
  public Set<Integer> getFreeVars ()
  {
    return freeVars;
  }
  
  /**
   * Returns a fresh variable that is not yet used in the current system
   * 
   * @return a fresh variable
   */
  public int getFreshVar ()
  {
    return freshVar;
  }

  /**
   * Returns true if a variable is bound.
   *
   * @param v a variable to be checked
   * @return true if {@code v} is bound, false otherwise
   */
  public boolean isBound (int v)
  {
    return varInfo.containsKey (v);
  }

  /**
   * Returns true if a variable is free.
   *
   * @param v a variable to be checked
   * @return true if {@code v} is free, false otherwise
   */  
  public boolean isFree (int v)
  {
    return freeVars.contains (v);
  }

  /**
   * Returns the set of all bound variables. A variable is bound if it
   * appears on the rhs of some equation in the system.
   *
   * @return the set of all bound variables.
   */
  public Set<Integer> boundVars ()
  {
    return varInfo.keySet ();
  }
  
  
  /**
   * Returns true if the current system is closed, i.e., has no free
   * variables.
   *
   * @return true if the current system has no free variables, false
   * otherwise.
   */
  public boolean isClosed ()
  {
    return getFreeVars ().isEmpty ();
  }
  

  /**
   * Returns the number of equations in the system.
   *
   * @return number of equations in the system.
   */
  public int size ()
  {
    return varInfo.size ();
  }

  /**
   * Returns true if the current system is empty.
   *
   * @return true if the system is empty, false otherwise
   */
  public boolean isEmpty ()
  {
    return size () == 0;
  }
  

  
  public Comparator<Integer> varComparator ()
  {
    return new Comparator<Integer> ()
      {
	public int compare (Integer o1, Integer o2)
	  {
	    VarInfo vInfo1 = varInfo.get (o1);
	    VarInfo vInfo2 = varInfo.get (o2);
	    
	    if (vInfo1 != null && vInfo2 != null)
	      return vInfo1.getId () - vInfo2.getId ();
	    
	    return o1.intValue () - o2.intValue ();
	  }
      };
  }
  

  public Map<Integer,V> evaluate (Map<Integer,V> env)
  {
    Map<Integer,V> res = new TreeMap<Integer,V> ();
    
    for (Map.Entry<Integer,V> entry : env.entrySet ())
      res.put (entry.getKey (), entry.getValue ());
    
    for (VarInfo vInfo : varInfo.values ())
      {
	Equation<V> eq = vInfo.getDefEq ();
	res.put (eq.getDefVar (), eq.eval (env));
      }
    return res;
  }
  
  public boolean isFixpoint (Map<Integer,V> env)
  {
    for (VarInfo vInfo : varInfo.values ())
      {
	Equation<V> eq = vInfo.getDefEq ();
	int var = eq.getDefVar ();
	
	V oldV = env.get (var);
	V newV = eq.eval (env);
	
	if (! newV.equals (oldV))
	  return false;
      }
    return true;
  }
  
  
  
  protected VarInfo newVarInfo (int id, FixPtModality m, Equation<V> eq)
  {
    int var = eq.getDefVar ();
    
    if (varInfo.containsKey (var))
      throw new IllegalArgumentException ("Duplicate equation for variable: " 
					  + var);
    
    VarInfo vInfo = new VarInfo (id, m, eq);

    varInfo.put (var, vInfo);


    int maxV = var;

    // -- update the list of free variables
    // -- as a side effect, also comput the largest variable occuring in 
    // -- in the current equation

    // -- remove variable bound by current equation
    freeVars.remove (var);
    
    // -- add all variables introduced by the current equation and 
    // -- not bound already into the list of free variables
    for (int v : eq.getUseVars ())
      {
	if (!isBound (v)) freeVars.add (v);
	if (v > maxV) maxV = v;
      }
    
    
    // -- update fresh variable
    if (maxV >= freshVar)
      freshVar = maxV + 1;
    

    return vInfo;
  }
  

  protected class VarInfo 
  {
    protected int var;
    protected int id;
    
    protected FixPtModality fpMod;
    
    protected Equation<V> defEq;

    public VarInfo (int id, FixPtModality fpMod, Equation<V> defEq)
    {
      this.var = defEq.getDefVar ();
      this.id = id;
      this.fpMod = fpMod;
      this.defEq = defEq;
    }
    
    
    public int getVar ()
    {
      return var;
    }
    
    public int getId ()
    {
      return id;
    }
    
    public FixPtModality getFpMod ()
    {
      return fpMod;
    }
    
    public Equation<V> getDefEq ()
    {
      return defEq;
    }
    
    
  }
  
  
}
