package edu.toronto.cs.yasm.pgenerator;

import edu.toronto.cs.tp.ITheoremProver;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.pprogram.PrllAsmtPStmt;
import edu.toronto.cs.yasm.pprogram.PProgram;
import edu.toronto.cs.yasm.wp.WPComputer;
import edu.toronto.cs.yasm.abstractor.Refiner;


import java.util.*;

/**
 * Generates predicates based on the weakest pre-condition of 
 * existing predicates
 *
 * This class solely handles the part that generates new predicate(s) 
 * Call the constructor Function to initiate.  
 * Call the function newPred() to return a list of Predicates
 *
 * @author <a href="mailto:maxin@epoch.cs">Xin Ma</a>
 * @version 1.0
 */
public abstract class WPPredicateGenerator extends BasePredicateGenerator
{

  
  public WPPredicateGenerator (PProgram p, 
			       PrllAsmtPStmt _stmt,
			       Refiner _refiner,
			       WPComputer _wp,
			       List<Expr> _pStatePred, 
			       List<Expr> _cStatePred)
  {
    super (p, _stmt,_refiner, _wp, _pStatePred, _cStatePred);
  }
  

  
  public boolean find ()
  {
    newPreds = newPred ();
    return !newPreds.isEmpty ();
  }
  
  
  /**
   * This is the main function to call after construct and object
   * to obtain one or more new Predicates in the form of a list
   *
   * @return a <code>List</code> value
   */
  public List<Expr> newPred()
  {
    // -- if child state has no predicates, nothing to do
    // XXX as far as I can see, this should not happen
    if (cStatePred.size () == 0) return null;
    
    // -- find variables that we want to use to construct new
    // -- predicates
    List<Expr> l = findSuspects (pStatePred, cStatePred);
    if (l.size() != 0)
      return updatePred (l);
    else
      return l;

    // if the previous method fails, one of the reason is that 
    // none of the predicates change value in this maybe transition
    // Indicating the reason for maybe transition does not come
    // from the sofisfication of the current predicates (which means
    // substitution to find the new predicates can't help), but from
    // insufficient predicates due to prove or disprove logically the
    // maybe transition.

    // In this case, we need to do something more sofisticated.
//     AIntellegence Ai = 
//       new AIntellegence (cvclConverter, vc, existingPred, predMap, 
// 			 wp, pStatePred, cStatePred, prop);
//     return Ai.inducePred (0);
  }
  


  /**
   * Given parent state and child state return the list
   * of predicates that should be used for refining
   *
   * @param pStatePred a <code>CVectorExpr</code> value
   * @param cStatePred a <code>CVectorExpr</code> value
   * @return a <code>List</code> value
   */
  protected abstract List<Expr> findSuspects (List<Expr> pStatePred, 
					      List<Expr> cStatePred);

  

  
  /**
   * This method checks whether the change of value
   * of the same predicate before and after the maybe
   * transition is due to a change of
   * value (in a true transition), or it's during a 
   * maybe transition
   *
   * @param from an <code>Expr</code> value
   * @param to an <code>Expr</code> value
   * @return a <code>boolean</code> value
   */
  protected boolean maybeTransition (Expr from, Expr to)
  {
    // XXX variable 'from' is not used !

    // -- compute weakest pre-condition of 'to' with respect to 
    // -- current statement
    //     to = edu.toronto.cs.yasm.util.Util.preSubstitution (to);
    //     edu.toronto.cs.yasm.util.Util.substituteVars (to, symbolicMap);
    //     to = vc.importExpr (to);

    // -- ignore non-deterministic assignments for possible reason
    // -- of an unknown transition
    
    // XXX I THINK it expects the pos wp here
    to = computeWP (to);
    if (to == null) return false;
    to = ChoiceOp.lgetT (to);

    int scope = tp.stackLevel ();
    tp.push ();
    // -- assert  the constraint corresponding to the parent state
    tp.assertFormula (pStatePred);
    

    // -- check if parent state satisfies the pre-condition
    boolean result = tp.query (to);

    tp.popto (scope);
    return !result;
  }
  
  
  /**
   * Check if all the the predicates do not repeat the predicate
   * to be added
   *
   * @param l a <code>List</code> value
   * @return a <code>List</code> value
   */
  private List<Expr> updatePred (List<Expr> l)
  {
    List<Expr> newPreds = new ArrayList<Expr>();

    for (Expr pred : l)
      {

	Expr wp = computeWP (pred);
	if (wp == null)
	  {
	    System.err.println ("Need to know " + pred + 
				" that changes non-deterministically");
	    continue;
	  }
      
	// XXX I THINK it expects the pos wp here
	wp = ChoiceOp.lgetT (wp);

	for (Expr wpTerm : BoolOp.terms (wp, new HashSet<Expr> ()))
	  {
	    System.err.println ("AAA Found a predicate: " + wpTerm);
	    System.err.println ("AAA comparing to: " + existingPred);

	    // XXX confusing, 
	    // XXX but ah well: hasPredicate strips negation if wpTerm
	    // matches an existing pred
	    wpTerm = hasPredicate (wpTerm);

	    if (wpTerm == null)
	      System.err.println ("AAA But think I have it already");

	    else if (wpTerm != null)
	      {
		System.out.println ("\t**********ATTENTION**************");
		System.out.println ("\tNew Predicates is:");
		System.out.println ("\t" + wpTerm);
		newPreds.add (wpTerm);
	      }
	  }
      }
    return newPreds;
  }


  /**
   * Weakest precondition predicate generator that only looks at predicates
   * that have changed a value on a transition
   *
   */
  public static class WPDiff extends WPPredicateGenerator
  {
    
    public WPDiff (PProgram p, 
		   PrllAsmtPStmt _stmt,
		   Refiner _refiner,
		   WPComputer _wp,
		   List<Expr> _pStatePred, 
		   List<Expr> _cStatePred)
    {
      super (p, _stmt,_refiner, _wp,_pStatePred, _cStatePred);
    }
    


    /**
     * This returns a list of predicates that are different
     * before and after the maybe transition
     *
     * @param a a <code>CVectorExpr</code> value
     * @param b a <code>CVectorExpr</code> value
     * @return a <code>List</code> value
     */
    protected List<Expr> findSuspects (List<Expr> a, List<Expr> b)
    {
      List<Expr> l = new ArrayList<Expr> ();
      for (int i = 0; i < a.size (); i++)
	if (a.get (i) != b.get (i) &&
	    maybeTransition (a.get (i), b.get (i)))
	  l.add (a.get (i));

      return l;
    }

  }
  

  
  /**
   * Weakest precondition predicate generator that only looks at predicates 
   * that did not change a value on a transition
   *
   */
  public static class WPSame extends WPPredicateGenerator
  {


    public WPSame  (PProgram p, 
		    PrllAsmtPStmt _stmt,
		    Refiner _refiner,
		    WPComputer _wp,
		    List<Expr> _pStatePred, 
		    List<Expr> _cStatePred)
    {
      super (p, _stmt,_refiner, _wp,_pStatePred, _cStatePred);
    }
    

    /**
     * This returns a list of predicates that are different
     * before and after the maybe transition
     *
     * @param a a <code>CVectorExpr</code> value
     * @param b a <code>CVectorExpr</code> value
     * @return a <code>List</code> value
     */
    protected List<Expr> findSuspects (List<Expr> a, List<Expr> b)
    {
      List<Expr> l = new ArrayList<Expr> ();
      for (int i = 0; i < a.size (); i++)
	if (a.get (i) == b.get (i) &&
	    maybeTransition (a.get (i), b.get (i)))
	  l.add (a.get (i));

      return l;
    }

  }


  
  /**
   * Weakest precondition predicate generator that looks at all
   * predicates
   *
   */
  public static class WPAll extends WPPredicateGenerator
  {


    public WPAll  (PProgram p, 
		    PrllAsmtPStmt _stmt,
		    Refiner _refiner,
		    WPComputer _wp,
		    List<Expr> _pStatePred, 
		    List<Expr> _cStatePred)
    {
      super (p, _stmt,_refiner, _wp,_pStatePred, _cStatePred);
    }
    

    /**
     * This returns a list of predicates that are different
     * before and after the maybe transition
     *
     * @param a a <code>CVectorExpr</code> value
     * @param b a <code>CVectorExpr</code> value
     * @return a <code>List</code> value
     */
    protected List<Expr> findSuspects (List<Expr> a, List<Expr> b)
    {
      List<Expr> l = new ArrayList<Expr> ();
      for (int i = 0; i < b.size (); i++)
	// -- only add predicates that cause the transition to be maybe
	// XXX maybeTransition function does not use its first argument
	if (maybeTransition (null, b.get (i)))
	  l.add (b.get (i));
      return l;

    }

  }  
}
