package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.yasm.pprogram.PredicateTable.*;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.pcfa.*;
import edu.toronto.cs.yasm.pcfa.Node.*;
import edu.toronto.cs.yasm.reachability.*;

import edu.toronto.cs.mvset.*;
import edu.toronto.cs.algebra.BelnapAlgebra;
import edu.toronto.cs.algebra.AlgebraValue;
import edu.toronto.cs.modelchecker.StatePresenter;
import edu.toronto.cs.ctl.*;

import java.util.*;

/**
 * An implementaion {@code PProgramReachabilityCompiler} that inlines
 * function calls. 
 *
 *
 * @author <a href="mailto:arie@cs.toronto.edu">Arie Gurfinkel</a>
 * @version 1.0
 */
public class InliningReachabilityCompiler 
  extends BasePProgramVisitor
  implements PProgramReachabilityCompiler
{

  // -- true if initial values of the predicates should be treated as 
  // -- unknown
  boolean unknownInit;

  // -- algebra we are using
  BelnapAlgebra algebra;


  
  // -- constants for logical values
  MvSet mvT;
  MvSet mvF;
  MvSet mvD;
  MvSet mvM;

  // -- all pre-state variables cube
  MvSet preVarCube;
  // -- all post-state variables cube
  MvSet postVarCube;

  // -- a map from pre- to post-variables
  int[] preToPostMap;
  // -- a map from post- to pre-variables
  int[] postToPreMap;


  // -- compilation contexts for each function
  Map<PFunctionDef,CompilationContext> ctxts;
  // -- global compilation context
  CompilationContext globalCtx;
  // -- current compilation context
  CompilationContext ctx;

  // -- the predicate program we are compiling
  PProgram pProgram;
  // -- Predicate Manager from pProgram, for easy access
  PProgramPredicateManager pm;
  


  // -- transition relation (a set of edges) we are building
  EdgeSet edges;

  // -- our initial state
  MvSet init;


  // -- a way to create MDD based mv-sets
  MDDMvSetFactory mvSetFactory;
  //JCUDDMvSetFactory mvSetFactory;
  //JCUDDBelnapMvSetFactory mvSetFactory;
  
  int maxDDVars = 20;



  public InliningReachabilityCompiler ()
  {
    algebra = new BelnapAlgebra ();
    unknownInit = true;
  }
  

  // ----------------------------------------------------------------------
  // -- Implementation of the PProgramReachabilityCompiler interface
  // ----------------------------------------------------------------------

  public void setPProgram (PProgram v)
  {
    pProgram = v;
  }

  public void setMaxDDVars (int v)
  {
    maxDDVars = v;
  }

  public StatePresenter getStatePresenter ()
  {
    return new StatePresenter ()
      {
	public CTLNode[] toCTL (AlgebraValue[] state)
	{
	  // XXX Don't know how to handle this yet
	  throw new UnsupportedOperationException ();
	}
	// XXX Don't know if we ever want this at all!
	public CTLNode[][] toCTL (AlgebraValue[][] states)
	{
	  throw new UnsupportedOperationException ();
	}


	public CTLNode[] toCTL (MvSet cube)
	{
	  AlgebraValue[] values = 
	    (AlgebraValue[]) cube.cubeIterator ().next ();

	  //System.out.println ("toCTL: " + Arrays.asList (values));

	  List result = new LinkedList ();
	  for (Iterator it = 
		 pProgram.getPredTable ().getVariables ().iterator (); 
	       it.hasNext ();)
	    {
	      Predicate var = (Predicate) it.next ();
	      if (var.isShadow ()) continue;
	      CTLNode ctl = var.toCTL (values);
	      if (ctl != null) result.add (ctl);
	    }
	  return (CTLNode[])result.toArray (new CTLNode [result.size ()]);
	}
      };
  }

  public ReachabilityStructure compile ()
  {

    // -- seal everything, we are ready to rumble
    seal ();

    // -- create an empty edge set
    edges = new EdgeSet ();

    // -- clear contexts
    ctxts = new IdentityHashMap<PFunctionDef,CompilationContext> ();
    

    // -- use the visitor pattern to compile the pprogram 
    pProgram.accept (this, null);
    
    // -- show progress
    System.out.println ("done building edges?!");

    // -- bundle everything in a ReachabilityStructure and return
    PredicateTable pTable = pProgram.getPredTable ();
    
    // - create the model for reachabiltiy analysis
    return new ReachabilityStructure (algebra,
				      new NodeSet (),
				      new NodeSet (),
				      edges,
				      pTable.variableMap (0, 2),
				      pTable.variableMap (2, 1),
				      pTable.variableMap (1, 2),
				      mvSetFactory.buildCube 
				      (pTable.getVariableIds (2)),
				      mvSetFactory.buildCube 
				      (pTable.getVariableIds (0)),
				      getStatePresenter ()
				      );
    
  }



  // ----------------------------------------------------------------------
  // -- visitXXX methods 
  // ----------------------------------------------------------------------

  /**
   * Runs compiler on {@code p}
   *
   * @param p a <code>PProgram</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitPProgram (PProgram p, Object o)
  {


    // -- build global context
    globalCtx = createContext (null);



    // XXX We view a program as implicitly having an initialization
    // function in which the execution begins from the "init"
    // statement, then goes to the main function, and finally ends at
    // the "END" statement. It would be simpler to turn this into an
    // explicit function instead

    ctx = globalCtx;
    // -- compile the "init" and "END" statments
    compilePStmtBlock (p.getProgramStmts (), o);
    ctx = null;
    
    // -- compile each function
    for (Object def : p.getFunctionDefs ().values ())
      ((PFunctionDef) def).accept (this, o);

    // -- visitor pattern requires us to return something
    return o;
  }


  /**
   * Compiles a function defined by <code>fDef</code>. 
   *
   * @param fDef a <code>PFunctionDef</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitFunctionDef (PFunctionDef fDef, Object o)
  {


    // XXX this is a good place to deal with special statements XXX
    // like call-prologue and call-return that go between two
    // different functions

    // XXX this is a good place to do function level setups, like
    // invariants, skips, cubes and maps
    
    // XXX We will need to differentiate between global and local
    // XXX inconsistencies

    // -- SET WHAT THIS FUNCTION THINGS IS INCONSISTENT
    // -- convert syntactical representation of inconsistent states to
    // -- smenatic representation using three-valued predicate --
    // -- variables
    
    
    // -- set compilation context
    ctx = getContext (fDef);
    // -- compile the body of the function
    compilePStmtBlock (fDef.getHead (), getContext (fDef));
    // -- remove active compilation context
    ctx = null;
    
    return o;
    
  }

  /**
   * Compile if-then-else statement.
   *
   * @param ifStmt an <code>IfPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitIf (IfPStmt ifStmt, Object o)
  {
    // -- deal with the if condition
    MvRelation[] branches;
    boolean doAX;

    // -- check if this if is a non-deterministic choice
    if (NondetOp.isNondet (ifStmt.getCond ().getCond ()))
      {
	branches = new MvRelation[] {new SkipMvRelation (ctx.invar), 
				     new SkipMvRelation (ctx.invar)};
	doAX = false;
      }
    else
      // -- normal if-then-else
      {
	MvSet cond = exprToMvSet (ifStmt.getCond ().getCond ());
	// -- the then- and else-branches of the if
	branches = new MvRelation[] {doIteBranch (cond), 
				     doIteBranch (cond.not ())};
	doAX = true;
      }
    
    // -- build and add the edge
    edges.add 
      (new MvRelnHyperEdge 
       (toNodeId (ifStmt),
	new Node.NodeId[] { toNodeId (ifStmt.getThenGoto ().getDest ()),
			    toNodeId (ifStmt.getElseGoto ().getDest ()) },
	branches, doAX));


    // -- deal with the statments in branch
    compilePStmtBlock (ifStmt.getThenStmt (), o);
    compilePStmtBlock (ifStmt.getElseStmt (), o);

    return o;
  }


  /**
   * Compiles a parallel assignment statement.
   *
   * @param prllStmt a <code>PrllAsmtPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitPrllAsmt (PrllAsmtPStmt s, Object o)
  {

    edges.add (new MvRelnEdge (toNodeId (s),
			       toNodeId (s.getDest ()),
			       doPrllAssign (s, ctx)));
    return o;
    
  }

  /**
   * Compiles skip (noop) statement.
   * 
   * @param skip a <code>SkipPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitSkip (SkipPStmt skip, Object o)
  {
    // -- deal with "init" statement, the first statement of 
    //    a program, see visitPProgram ()
    if (skip.getLabel ().equals ("init"))
      return visitInit (skip, o);

    // -- deal with normal skip statment
    edges.add (new MvRelnEdge (toNodeId (skip),
			       toNodeId (skip.getDest ()),
			       new SkipMvRelation (ctx.invar)));
    return o;
  }
  
  protected Object visitInit (SkipPStmt init, Object o)
  {
    assert init.getLabel ().equals ("init") : init + " is not init!";
    edges.add (new MvRelnEdge (toNodeId (init),
			       toNodeId (init.getDest ()),
			       doInit ()));
    return o;
  }
  
  public Object visitGoto (GotoPStmt gotoStmt, Object o)
  {

    // -- contruct the relation for this goto Note that a goto may
    // -- correspond to a return selector of a function which we need
    // -- to handle differently. This case is detected by checking
    // -- that the parent function of the goto is different from the
    // -- parent function of its destination
    MvRelation reln;

    MvSet localsPreCube = null;
    MvSet invar = ctx.invar;
    
    if (gotoStmt.getParent () != gotoStmt.getDest ().getParent ())
      {
	// -- on the return edge, when we go fwd we abstract away all
	// -- locals, when go bwd there is nothing to do

	localsPreCube = mvSetFactory.buildCube 
	  (pProgram.getPredTable ().
	   getVariableIds 
	   (excludeScope (ctx.preds, gotoStmt.getDest ().getParent ()), 0));
	

	// -- get rid of invar since we don't handle it properly at the moment
	invar = null;	
      }
    
    edges.add (new MvRelnEdge (toNodeId (gotoStmt),
			       toNodeId (gotoStmt.getDest ()),
			       new SkipMvRelation (invar, localsPreCube, null)));

    return o;


  }
  
  /**
   * Compiles non-deterministic goto statement.
   *
   * @param ndGotoStmt a <code>NDGotoPStmt</code> value
   * @param o an <code>Object</code> value
   * @return an <code>Object</code> value
   */
  public Object visitNDGoto (NDGotoPStmt ndGotoStmt, Object o)
  {
    // -- compute a hyper-edge that captures all branches of the 
    // -- non-deterministic goto so that they can be handled in a single
    // -- backward image computation

    Node.NodeId[] dests;
    MvRelation[] relns;

    // -- allocate space to store destinations and relations of ndgoto
    dests = new Node.NodeId [ndGotoStmt.getDests ().size ()];
    relns = new MvRelation [ndGotoStmt.getDests ().size ()];

    int i = 0;
    for (Object d : ndGotoStmt.getDests ())
      {
	dests [i] = toNodeId ((PStmt) d);
	relns [i++] = new SkipMvRelation (ctx.invar);
      }

    edges.add (new MvRelnHyperEdge (toNodeId (ndGotoStmt), 
				    dests, relns, false));
    return o;
  }

  public Object visitReturnSelector (ReturnSelectorPStmt sel, Object o)
  {
    int i = 0;

    // -- create an edge for each destination of this return selector
    for (Object v : sel.getDests ())
      {
	PStmt dest = (PStmt) v;
	
	// -- find out which predicates are local in sel.getParent function
	// -- with respect to the destination
	MvSet localsPreCube = mvSetFactory.buildCube 
	  (pProgram.getPredTable ().
	   getVariableIds (excludeScope (ctx.preds, dest.getParent ()), 0));

	// -- a relation corresponding to kill L where L is the set of
	// -- all local variables. 
	MvRelation killLocals = 
	  new SkipMvRelation (null, localsPreCube, null);
	
	edges.add (new MvRelnEdge (toNodeId (sel),
				   toNodeId (dest),
				   new SeqMvRelation 
				   (new ReturnSelectorMvRelation
				    (null, sel.getReturnIndex ().eq (i++),
				     sel.getReturnIndex ().getBitCube ()),
				    killLocals)));
      }
    
    return o;
  }
  
  public Object visitFunctionCallPrologue (FunctionCallPrologue fPro,
					   Object o)
  {
    PFunctionDef caller = fPro.getParent ();
    PFunctionDef callee = fPro.getCall ().getFunctionDef ();

    // -- check that this is not a self-recursive call 
    if (caller == callee)
      throw new RuntimeException ("Function " + fPro.getParent () + 
				  " calls itself and cannot be inlined");


    // -- compute the relation corresponding to the call edge on the
    // -- fwd image fProReln gets rid of all global and callee
    // -- variables since they are reasigned by the arguments assignment
    // -- bwd image of fProReln gets rid of all inScope (callee) predictates
    // -- since their scope ends after the call
    MvRelation fProReln = doPrllAssign (fPro, getContext (callee),
					getContext (callee));
    
    // -- if this is a call to a function with multiple call sites
    // -- it must be wrapped in a FunctionCallMvRelation that keeps track 
    // -- of where this call is made from
    if (callee.getReturnSelector () instanceof ReturnSelectorPStmt)
      {
	ReturnSelectorPStmt ret = (ReturnSelectorPStmt) 
	  fPro.getCall ().getFunctionDef ().getReturnSelector ();
	
	fProReln = new FunctionCallMvRelation 
	  (null, ret.getReturnIndex ().eq (fPro.getCall ().getCallIndex ()), 
	   ret.getReturnIndex ().getBitCube (), fProReln);
      }
    
    edges.add (new MvRelnEdge (toNodeId (fPro),
			       toNodeId (fPro.getDest ()),
			       fProReln));
    return o;
  }

  // ----------------------------------------------------------------------
  // -- doXXX FUNCTIONS THAT ARE USED BY handleXXX LEVEL TO ACTUALLY 
  // -- MANIPULATE EXPRESSIONS AND MVSETS
  // ----------------------------------------------------------------------


  /**
   * Compiles a conditional branch.
   *
   * @param cond a <code>MvSet</code> value
   * @return a <code>MvRelation</code> value
   */
  protected MvRelation doIteBranch (MvSet cond)
  {
    return toMvRelation (currToNext (cond.eq (mvF).not ()).and 
			 (cond.and (doSkip (ctx.vars))), ctx);
  }


  protected MvRelation doPrllAssign (PrllAsmtPStmt assign,
				     CompilationContext ctx)
  {
    MvSet result = mvT;

    // -- compile each update in the assignment
    for (Object n : assign.getAsmts ())
      if (n instanceof NDAsmtPStmt)
	result = result.and (doNDUpdate ((NDAsmtPStmt) n));
      else
	result = result.and (doUpdate ((AsmtPStmt) n));

    // -- see if we have an invariant do deal with as well
    // -- XXX WHY DO WE HAVE INVAR POST AT ALL!
    MvSet invarPost = null;
    if (assign.getInconsistent () != null)
      invarPost = exprToMvSet (assign.getInconsistent ()).eq (mvF);
    
    return toMvRelation (result, invarPost, ctx);
  }
  
  protected MvRelation doPrllAssign (PrllAsmtPStmt assign, 
				     CompilationContext srcCtx,
				     CompilationContext dstCtx)
  {
    MvSet result = mvT;

    // -- compile each update in the assignment
    for (Object n : assign.getAsmts ())
      if (n instanceof NDAsmtPStmt)
	result = result.and (doNDUpdate ((NDAsmtPStmt) n));
      else
	result = result.and (doUpdate ((AsmtPStmt) n));

    
    return toMvRelation (result, srcCtx, dstCtx);
  }


  protected MvSet doUpdate (AsmtPStmt update)
  {

    return doAssign (exprToMvSet (update.getVar ()),
	      exprToMvSet (update.getTrueCond ()),
	      exprToMvSet (update.getFalseCond ()));
  }

  protected MvSet doNDUpdate (NDAsmtPStmt update)
  {
    MvSet result = mvF;
    
    for (Object o : update.getUpdates ())
      {
	AsmtPStmt u = (AsmtPStmt) o;
	result = result.or (doAssign (exprToMvSet (u.getVar ()),
				      exprToMvSet (u.getTrueCond ()),
				      exprToMvSet (u.getFalseCond ())));
      }
    return result;
  }

  

  /**
   * Compiles a single assignment: {@literal var := choose (tExp, fExp)}
   *
   * @param var a <code>MvSet</code> value
   * @param tExp a <code>MvSet</code> value
   * @param fExp a <code>MvSet</code> value
   * @return a <code>MvSet</code> value
   */
  protected MvSet doAssign (MvSet var, MvSet tExp, MvSet fExp)
  {
    MvSet nextVar = currToNext (var);

    MvSet leftBranch;

    leftBranch = nextVar.eq (mvT).and 
      (tExp.infoAnd (mvT).infoOr 
       (fExp.infoAnd (mvT).not ()));

    MvSet rightBranch;
    rightBranch = nextVar.eq (mvF).and 
      (fExp.infoAnd (mvT).infoOr
       (tExp.infoAnd (mvT).not ()));

    return leftBranch.or 
      (rightBranch.or 
       (nextVar.eq (mvM).and (mvD)));
    
  }


  /**
   * Returns a formula representing a skip relation on variables in
   * {@code vars}. 
   *
   * @return a <code>MvSet</code> value
   */
  protected MvSet doSkip (List vars)
  {
    
    MvSet skip = mvT;

    for (Object o : vars)
      {
	Predicate var = (Predicate) o;

	// -- just in case we have garbage in the list, ignore
	// -- it silently
	if (var.isShadow ()) continue;
	if (var instanceof IntVariable) continue;

	MvSet curr = var.getMvSet ();
	MvSet next = currToNext (curr);

	// mv-set for 
	// var := choice (var, !var)
	next = next.eq (mvT).and (curr).or
	  (next.eq (mvF).and (curr.not ())).or 
	  (next.eq (mvM).and (mvD));

	skip = skip.and (next);
      }

    return skip;
  }

  /**
   * Compiles initial statement.
   *
   * @return a <code>MvRelation</code> value
   */
  protected MvRelation doInit ()
  {

    // -- we only need initialize the predicates whose scope is in
    //    the "main" function, other variables are initialized
    //    at the corresponding call edges

    CompilationContext mainCtx = getContext (pProgram.getMainFunctionDef ());
    

    MvSet init = mvT;

    
    for (Iterator it = mainCtx.vars.iterator (); it.hasNext ();)
      {
	Predicate var = (Predicate) it.next ();
	if (var.isShadow ()) continue;
	if (var instanceof IntVariable) continue;

	if (unknownInit)
	  // -- initially all variables are unknown
	  init = init.and (doAssign (var.getMvSet (), 
				     mvF, 
				     mvF));
	else
	  // -- initially all variables are set non-deterministically
	  init = init.and (doAssign (var.getMvSet (), 
				     mvT, 
				     mvF).or
			   (doAssign (var.getMvSet (), 
				      mvF, 
				      mvT)));
      }

    return toMvRelation (init, mainCtx);
  }


  // ----------------------------------------------------------------------
  // -- PROTECTED METHODS
  // ----------------------------------------------------------------------
  /**
   * Initializes all data structures so that compilation can take
   * place. This is the first function called in {@link #compile}.
   *
   */
  protected void seal ()
  {

    // -- create an mv-set factory we want to use
    // XXX THIS IS VERY UGLY RIGHT NOW
    try 
      {
	if (mvSetFactory == null)
 	  //mvSetFactory = (JCUDDBelnapMvSetFactory)
	  //JCUDDBelnapMvSetFactory.newMvSetFactory (algebra, maxDDVars);
	  mvSetFactory = (MDDMvSetFactory)
	    MDDMvSetFactory.newMvSetFactory (algebra, maxDDVars);
	//      mvSetFactory = (JCUDDMvSetFactory)
	//        JCUDDMvSetFactory.newMvSetFactory (algebra, maxDDVars);

      }
    catch (Exception ex)
      {
	// -- don't expect this to actually happen.
	ex.printStackTrace ();
	throw new RuntimeException (ex);
      }


    // -- get predicate table, we'll use it a lot
    PredicateTable pTable = pProgram.getPredTable ();

    // -- tell the predicate table which factory we are using
    pTable.setMvSetFactory (mvSetFactory);

    // -- intialize constants for easy access
    mvT = mvSetFactory.top ();
    mvF = mvSetFactory.bot ();
    mvD = mvSetFactory.infoTop ();
    mvM = mvSetFactory.infoBot ();


    // -- build cubes and maps for pre- and post- variables
    preVarCube = mvSetFactory.buildCube (pTable.getVariableIds (0));
    postVarCube = mvSetFactory.buildCube (pTable.getVariableIds (1));
    preToPostMap = pTable.variableMap (0, 1);
    postToPreMap = pTable.variableMap (1, 0);


    // -- extract PredicateManager for easy access
    pm = pProgram.getPProgramPredicateManager ();
  }

  /**
   * Compiles the statements in a block starting from the
   * <code>blockHead</code>.
   *
   * @param blockHead a <code>PStmt</code> value : the first statement
   * of the block to be compiled
   * @param o an <code>Object</code> value 
   * @return an <code>Object</code> value
   */
  protected void compilePStmtBlock (PStmt blockHead, Object o)
  {
    for (PStmt s = blockHead; s != null; s = s.getNext ())
      s.accept (this, o);
  }


  // ----------------------------------------------------------------------
  // -- VARIOUS HELPER FUNCTIONS
  // ----------------------------------------------------------------------
  /**
   * Helper function that converts an expression {@code expr} into its
   * {@code MvSet} representation.
   *
   * @param expr an <code>Expr</code> value
   * @return a <code>MvSet</code> value
   */
  protected MvSet exprToMvSet (Expr expr)
  {
    // -- check if expr can be mapped directly to a DD node.
    Predicate p = pProgram.getPredTable ().getByName (expr);
    if (p != null) return p.getMvSet ();

    
    // -- deal with constants
    if (expr.op () == BiLatticeOp.TTOP || expr.op () == BoolOp.TRUE)
      return mvT;
    else if (expr.op () == BiLatticeOp.TBOT || expr.op () == BoolOp.FALSE)
      return mvF;
    else if (expr.op () == BiLatticeOp.IBOT)
      return mvM;
    else if (expr.op () == BiLatticeOp.ITOP)
      return mvD;
    else if (expr.op () == BoolOp.NOT)
      return exprToMvSet (expr.arg (0)).not ();
    // -- apply AND and OR at DD level
    else if (expr.op () == BoolOp.AND)
      {
	MvSet result = mvT;
	for (int i = 0; i < expr.arity (); i++)
	  result = result.and (exprToMvSet (expr.arg (i)));
	return result;
      }
    else if (expr.op () == BoolOp.OR)
      {
	MvSet result = mvF;
	for (int i = 0; i < expr.arity (); i++)
	  result = result.or (exprToMvSet (expr.arg (i)));
	return result;
      }
    else
      {
	// -- Should never end up here but if we do, print something
	// -- for debuggin and through an exception.
	pProgram.getPredTable ().dump ();
	System.err.println ("expr is: " + expr);
	System.err.println ("expr's hashcode is: " + expr.hashCode ());
	throw new RuntimeException ("Something is wrong: " + expr);
      }
  }  


  /**
   * Helper function to rename all pre-variables in {@code v} to
   * post-variables.
   *
   * @param v a <code>MvSet</code> value
   * @return a <code>MvSet</code> value
   */
  protected MvSet currToNext (MvSet v)
  {
    return v.renameArgs (preToPostMap);
  }

  /**
   * Returns a {@link NodeId} corresponding to a numeric id {@code v}.
   *
   * @param v an <code>int</code> value
   * @return a <code>NodeId</code> value
   */
  protected NodeId intToNodeId (int v)
  {
    return new NodeId (v);
  }
  
  
  /**
   * Returns a {@link NodeId} corresponding to a statement {@code p}.
   *
   * @param p a <code>PStmt</code> value
   * @return a <code>NodeId</code> value
   */
  protected NodeId toNodeId (PStmt p)
  {
    return new NodeId (p.getId ());
  }


  protected MvRelation toMvRelation (MvSet v, MvSet postInvar, 
				     CompilationContext ctx)
				     
  {
    // -- deal with skips first
    if (postInvar == null && v == ctx.skipMvSet) 
      return new SkipMvRelation (ctx.invar);

    return v.toMvRelation (ctx.invar, 
			   postInvar,
			   ctx.preCube, 
			   ctx.postCube, 
			   ctx.preToPost, 
			   ctx.postToPre);
    
    
  }
  

  protected MvRelation toMvRelation (MvSet v, CompilationContext ctx)
  {
    return toMvRelation (v, (MvSet) null, ctx);
  }

  /**
   * Constructs an {@link MvRelation} from an mv-set {@code v}. This
   * method will try and avoid creating relations for skip.
   */
  protected MvRelation toMvRelation (MvSet v, 
				     CompilationContext srcCtx,
				     CompilationContext dstCtx)
  {
    return v.toMvRelation (null, srcCtx.preCube,
			   dstCtx.postCube,
			   dstCtx.preToPost,
			   dstCtx.postToPre);
  }
  

  // ----------------------------------------------------------------------
  // -- Helpers to find predicates in various scopes.  This should
  // -- eventually migrate somewhere else. Maybe PredicateManager?
  // ----------------------------------------------------------------------
  
  /**
   * Returns a subset of {@code l} that is in the scope of {@code
   * fDef}.  If {@code fDef} is null, reduces {@code l} to global
   * predicates only.
   */
  protected List<Expr> reduceToScope (List<Expr> l, PFunctionDef fDef)
  {
    List<Expr> r = new LinkedList<Expr> ();
    for (Expr e : l)
      {
	List<String> scope = pm.getPredicateScope (e);
	// -- check for global scope (empty)
	if (scope.isEmpty () ||
	    (fDef != null && scope.contains (fDef.getFunctionName ())))
	  r.add (e);
      }
    return r;
  }

  /**
   * Returns a subset of {@code l} that is <b>not</b> in the scope of
   * {@code fDef}. If {@code fDef} is null, only global predicates are
   * excluded
   */
  protected List<Expr> excludeScope (List<Expr> l, PFunctionDef fDef)
  {
    List<Expr> r = new LinkedList<Expr> ();
    for (Expr e : l)
      {
	List<String> scope = pm.getPredicateScope (e);
	if (!scope.isEmpty () && 
	    (fDef == null || !scope.contains (fDef.getFunctionName ())))
	  r.add (e);
      }

    return r;
  }
  

  /**
   * Returns predicates in the scope of both {@code fDef1} and {@code
   * fDef2}.  If {@code fDef1} and {@code fDef2} are the same, returns
   * only global predicates since these are the ones available to both
   * functions.
   */
  protected List<Expr> inScope (PFunctionDef fDef1, PFunctionDef fDef2)
  {
    PFunctionDef scope = fDef1 == fDef2 ? null : fDef2;
    return reduceToScope (pm.getFuncPreds (fDef1.getFunctionName ()), scope);
  }
  
  /**
   * Returns all predicates in the scope of function {@code fDef}.
   * If {@code fDef} is null, returns predicates in global scope.
   */
  protected List<Expr> inScope (PFunctionDef fDef)
  {
    return fDef == null ?
      reduceToScope (pm.getFuncPreds ("main"), null) :
      pm.getFuncPreds (fDef.getFunctionName ());
  }
  
  protected List<Expr> locals (PFunctionDef fDef)
  {
    return excludeScope (inScope (fDef), null);
  }
  
  
  /**
   * Returns true if {@code fPro} corresponds to a self-recursive call.
   *
   * @param fPro a <code>FunctionCallPrologue</code> value
   * @return a <code>boolean</code> value
   */
  protected boolean isSelfCall (FunctionCallPrologue fPro)
  {
    return fPro.getParent () == fPro.getCall ().getFunctionDef ();
  }


  // ----------------------------------------------------------------------
  // -- Methods to manipulate CompilationContext
  // ----------------------------------------------------------------------

  
  /**
   * Returns a {@code CompilationContext} of the source of statement
   * {@code s}.
   *
   * @param s a {@code PStmt} for which compilation context is thought.
   * @return the compilation context of {@code s}
   */
  protected CompilationContext getSrcCtx (PStmt s)
  {
    return getContext (s.getParent ());
  }
  
  /**
   * Returns a {@code CompilationContext} of {@code s.getDest}.
   *
   * @param s a {@code PStmt} for whose destination compilation
   * context is thought
   * @return the compilation context of {@code s.getDest}
   */
  protected CompilationContext getDstCtx (PStmt s)
  {
    return getContext (s.getDest ().getParent ());
  }
  
  
  /**
   * Returns a compilation context corresponding to function defined
   * by {@code fDef}.
   *
   * @param fDef a function definition
   * @return a <code>CompilationContext</code> of {@code fDef}
   */
  protected CompilationContext getContext (PFunctionDef fDef)
  {
    if (fDef == null) return globalCtx;
    
    CompilationContext ctx = ctxts.get (fDef);
    if (ctx == null)
      {
	ctx = createContext (fDef);
	ctxts.put (fDef, ctx);
      }
    return ctx;
  }
  
  
  /**
   * Creates a compilation context for function defined by {@code fDef}.
   *
   * @param fDef a function definition
   * @return a newly created <code>CompilationContext</code> of {@code fDef}
   */
  protected CompilationContext createContext (PFunctionDef fDef)
  {
    CompilationContext ctx = new CompilationContext ();
    

    PredicateTable pt = pProgram.getPredTable ();
    
    // -- find all predicates in the scope of fDef and build cubes and maps
    ctx.preds= inScope (fDef);
    ctx.vars = pt.getVariables (ctx.preds);
    ctx.preCube = mvSetFactory.buildCube (pt.getVariableIds (ctx.preds, 0));
    ctx.postCube = mvSetFactory.buildCube (pt.getVariableIds (ctx.preds, 1));
    ctx.preToPost = pt.variableMap (ctx.preds, 0, 1);
    ctx.postToPre = pt.variableMap (ctx.preds, 1, 0);

    // -- build invariant based on what is inconsistent right now
    // XXX only works if fDef is non-null, i.e., invar in global scope
    if (fDef != null && fDef.getInconsistent () != null)
      ctx.invar = exprToMvSet (fDef.getInconsistent ()).eq (mvT).not ();
    
    // -- build a skip mv-set
    ctx.skipMvSet = doSkip (ctx.vars);
    
    return ctx;
  }
  
  /**
   * This class provides an encapsulation of a compilation context.
   *
   */
  protected static class CompilationContext
  {
    
    /**
     * A list of predicates in scope of this context.
     *
     */
    protected List<Expr> preds;
    /**
     * A list of DD variables in scope of this context.
     *
     */
    protected List vars;

    /**
     * A set of all pre-state variables as an {@code MvSet}
     *
     */    
    protected MvSet preCube;
    /**
     * A set of all post-state variables as an {@code MvSet}
     *
     */
    protected MvSet postCube;

    /**
     * A map that turns all pre-state variables into their post-state
     * versions
     *
     */    
    protected int[] preToPost;
    /**
     * A map that turns all post-state variables into their pre-state
     * versions
     *
     */
    protected int[] postToPre;
    
    /**
     * An invariant of this context.
     *
     */
    protected MvSet invar;

    /**
     * A skip relation as an {@code MvSet} for this context.
     *
     */
    protected MvSet skipMvSet;
    

  }
  



  // ----------------------------------------------------------------------
  // -- DEBUGGING AND TESTING HELPERS
  // ----------------------------------------------------------------------
  
  /**
   * <b>BAD</b> Returns a string representation of an {@link
   * MvSet}. This is only used for debugging and should be moved into
   * {@code edu.toronto.cs.mvset} package.
   *
   * @param mvSet a <code>MvSet</code> value
   * @return a <code>String</code> value
   */
  protected String dumpMvSet (MvSet mvSet)
  {
    StringBuffer sb = new StringBuffer ();
    for (Iterator it = mvSet.cubeIterator (); it.hasNext ();)
      {
	sb.append (Arrays.asList ((Object[])it.next ()).toString ());
	sb.append ("\n");
      }
    return sb.toString ();
  }

  
}
