package edu.toronto.cs.yasm.pprogram;

import java.util.*;
import edu.toronto.cs.cparser.block.*;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.abstractor.*;
import edu.toronto.cs.tp.ITheoremProver;
import edu.toronto.cs.yasm.cfg.CFGraph;

import static edu.toronto.cs.expr.CILFunctionDefOp.*;
import static edu.toronto.cs.expr.CFunctionDeclOp.*;
import static edu.toronto.cs.expr.CResultDeclOp.*;
import static edu.toronto.cs.expr.TupleOp.*;

/**
 * PFunction.java
 *
 *
 * Created: Fri Jun 25 14:14:01 2004
 *
 * @author <a href="mailto:kelvin@tallinn.cs">Kelvin Ku</a>
 * @version 1.0
 */
public class PFunctionDef
{
  // -- only one of the following source fields should be in use!

  // -- source Block
  Block sourceBlock;

  // -- source Expr
  Expr sourceExpr;

  /**
   * Source CFG.
   */
  CFGraph cfg;

  // -- head of a graph of program statements defining this function
  PStmt head;

  // -- entry point to this function; note that this differs from head after
  // return selector is added
  PStmt entry;

  // -- head of a graph of program statements defining the return selector
  // block of this function
  PStmt sel;

  // -- Map of local declarations, including function parameters:
  //    (String) name -> (PDecl) declaration
  Map localDecls;

  // -- PProgram this function belongs to
  PProgram parent;

  // -- the set of inconsistent states w.r.t to predicates of current
  // -- function.
  //
  // -- Note that this is a syntactical representation of inconsistent
  // -- states in boolean programs, i.e. expression defined on program
  // -- variables, which will be later converted to semantic
  // -- representation using three valued predicate variables during
  // -- compilation phase.

  Expr inconsistent;

  Expr returnIndexVar;
  Expr returnValueVar;

  public PFunctionDef (PProgram _parent, PStmt _head, Map _localDecls,
                       Expr _sourceExpr)
  {
    parent = _parent;
    head = _head;
    localDecls = _localDecls;
    sourceExpr = _sourceExpr;
  }

  public PFunctionDef (Expr _sourceExpr)
  {
    this (null, null, null, _sourceExpr);
    if (isCFunctionDecl (sourceExpr))
      returnValueVar = resultDecl (sourceExpr);
    else
      returnValueVar = 
	sourceExpr.getFactory ().var (getFunctionName () + "::@return_value");

    returnIndexVar =
      sourceExpr.getFactory ().var (getFunctionName () + "::@return_index");
  }

  public PFunctionDef (CFGraph cfg, Expr sourceExpr)
  {
    this (sourceExpr);
    this.cfg = cfg;
  }

  public CFGraph getCFG ()
  {
    return cfg;
  }

  public void setEntry (PStmt _entry)
  {
    entry = _entry;
  }
  
  public PStmt getEntry ()
  {
    return entry;
  }

  public void setReturnSelector (PStmt _sel)
  {
    sel = _sel;
  }

  public PStmt getReturnSelector ()
  {
    return sel;
  }

  /**
   * Adds {@code ep} to the return selector's list of dests; returns the
   * index where {@code ep} was added.
   */
  public int addReturnSite (FunctionCallEpilogue ep)
  {
    ((ReturnSelectorPStmt) sel).addDest (ep);
    return ((ReturnSelectorPStmt) sel).getDests ().size () - 1;
  }

  public Object accept (PProgramVisitor v, Object o)
  {
    return v.visitFunctionDef (this, o);
  }

  /** 
   * Returns a label for call site #{@code index}. 
   * 
   * @param index 
   * @return 
   */
  public String callSiteLabel (int index)
  {
    return getFunctionName () + "_CALL_" + index;
  }

  /** 
   * Returns a label for return site #{@code index}. 
   * 
   * @param index 
   * @return 
   */
  public String returnSiteLabel (int index)
  {
    return getFunctionName () + "_RETURN_" + index;
  }

  /** 
   * Returns a label for the return selector. 
   * 
   * @return 
   */
  public String returnSelectorLabel ()
  {
    return getFunctionName () + "_RETURN_SELECTOR_HEAD";
  }

  /** 
   * Returns a label for the entry statement. 
   * 
   * @return 
   */
  public String entryLabel ()
  {
    return getFunctionName () + "_ENTRY";
  }
  
  public void setHead (PStmt v)
  {
    head = v;
  }

  /**
   * Appends v to head.
   */
  public void addToHead (PStmt v)
  {
    head.setTail (v);
  }

  public PStmt getHead ()
  {
    return head;
  }

  public PStmt getEntryPoint ()
  {
    return entry;
  }

  public void setLocalDecls (Map v)
  {
    localDecls = v;
  }

  public Map getLocalDecls ()
  {
    return localDecls;
  }

  public int getNumCallSites ()
  {
    if (sourceBlock != null)
      return sourceBlock.getNumCallSites ();

    if (sourceExpr != null)
      return CILFunctionDefOp.getNumCallSites (sourceExpr);  

    return -1;  
  }

  public String getFunctionName ()
  {
    if (sourceBlock != null)
      return sourceBlock.getFunctionName ();
    
    if (sourceExpr != null)
      if (isCILFunctionDef (sourceExpr))
	return CILFunctionDefOp.getFunctionName (sourceExpr);
      else if (isCFunctionDecl (sourceExpr))
	return CFunctionDeclOp.declName (sourceExpr);

    return null;  
  }

  public Expr getReturnType (ITheoremProver tp)
  throws antlr.RecognitionException
  {
    if (sourceBlock != null)
      return DeclsParserUtil.parseBlockTypeSpecifiers
        (sourceBlock.getFunctionReturnType (), tp);

    if (sourceExpr != null)
      return DeclsParserUtil.parseExprTypeSpecifiers
        (CILFunctionDefOp.getReturnType (sourceExpr), tp);

    assert false : "Both sourceBlock and sourceExpr are null";
    return null;
  }

  public Set getFormalParams ()
  {
    Set paramSet = new HashSet ();
    
    ExprFactory fac = getParent ().getExprFactory ();
    List params = new LinkedList ();
    //    params.add (fac.var (getFunctionName () + "::@return_index"));
    if (CILFunctionDefOp.hasParameters (sourceExpr))
      params.addAll (CILFunctionDefOp.getParameterNames (sourceExpr).args ());

    for (Iterator it = params.iterator (); it.hasNext ();)
       {
	 paramSet.add (it.next ());
       }
    
    return paramSet;
  }

  public PProgram getParent ()
  {
    return parent;
  }

  public void setParent (PProgram v)
  {
    parent = v;
  }

  public Expr getSourceExpr ()
  {
    return sourceExpr;
  }
  
  public int getLineNum ()
  {
    Expr src = getSourceExpr ();
    if (src == null) return -1;
    if (src.arity () != 5)
      {
	System.err.println ("Strange function def without a line number");
	System.err.println ("Srouce expr: " + src);
	return -1;
      }
    
    return RationalOp.getN (src.arg (4));
  }

  public boolean isNonVoid ()
  {
    if (sourceBlock != null)
      return sourceBlock.isNonVoidFunction ();
    
    if (sourceExpr != null)
      return CILFunctionDefOp.isNonVoid (sourceExpr);

    assert false : "Both sourceBlock and sourceExpr are null";
    return false;
  }

  public void setInconsistent (Expr v)
  {
    inconsistent = v;
  }

  public Expr getInconsistent ()
  {
    return inconsistent;
  }

  public Expr getReturnIndexVar ()
  {
    return returnIndexVar;
  }

  public Expr getReturnValueVar ()
  {
    return returnValueVar;
  }

  public Expr getFullParametersExpr ()
  {
    List<Expr> params = new LinkedList ();
    if (CILFunctionDefOp.isCILFunctionDef (sourceExpr))
    {
      ExprFactory fac = getParent ().getExprFactory ();
      params.add (getReturnIndexVar ());
      if (CILFunctionDefOp.hasParameters (sourceExpr))
	params.addAll
	  (CILFunctionDefOp.getParameterNames (sourceExpr).args ());
      return fac.op (CILListOp.LIST).naryApply (params);
    }
    else if (isCFunctionDecl (sourceExpr))
    {
      params.add (getReturnIndexVar ());
      params.addAll (CFunctionDeclOp.args (sourceExpr).args ());
      return tuple (params);
    }
    throw new RuntimeException ("Unknown sourceExpr " + sourceExpr.op ());
  }

  public String toString ()
  {
    return getFunctionName ();
  }

  public void print ()
  {
    if (head != null)
      head.print ();
  }

  public ExprFactory getExprFactory ()
  {
    return parent.getExprFactory ();
  }
} 

