package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.yasm.abstractor.*;
import edu.toronto.cs.cparser.*;
import edu.toronto.cs.cparser.block.*;
import edu.toronto.cs.boolpg.abstraction.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.clang.*;

import edu.toronto.cs.tp.SymbolTable;

import java.util.*;
import java.io.*;

/**
 * PProgram.java
 *
 *
 * Created: Fri Jun 25 14:14:01 2004
 *
 * @author <a href="mailto:kelvin@tallinn.cs">Kelvin Ku</a>
 * @version 1.0
 */
public class PProgram 
{
  // -- (String) function-name -> (PFunctionDef) function-definition
  Map<String, PFunctionDef> functionDefs;

  // -- (String) variable-name -> (PDecl) variable-declaration
  Map globalDecls;

  // -- Map: (String) label-in-source -> (PStmt) corresponding statement
  Map labelledStatementMap;

  // -- direct map:
  //      (int) generated-label-index -> (PStmt) corresponding statement
  // statmentList contains all the statements of the program
  // including programStmts and statements in each function (start from
  // func.head)
  //
  // maps a unique id to every statement in the program
  List<PStmt> statementList;

  // -- table of predicates
  PredicateTable pTable;

  public static final String MAIN_FUNCTION_NAME = "main";

  // -- true if declarations (global and local) have been refined; false o.w.
  boolean declsRefined;

  ExprFactory fac;

  // -- the set of inconsistent predicates
  Expr inconsistent;

  // -- the "init" and "end" statements
  //    these two statements do not belong to any functions
  //    next relation: init -- next --> end -- next --> null
  //    dst  relation: init -- dst --> main.entry
  //                 main.exit (returnSelector) -- dst --> end -- dst --> end
  PStmt programStmts;


  // -- Predicate Manager w.r.t. scope
  PProgramPredicateManager pProgramPredicateManager;
  

  // -- return selector type
  int selectorType;

  // -- return selector types
  public static final int LinSelector = 0;
  public static final int BinSelector = 1;
  public static final int DDSelector = 2;

  /**
   * Source program data structures.
   *
   * Note: CProgram contains the symbol table for the source program. Also,
   * given a CProgram, globalDecls is not defined as it is redundant.
   */
  SymbolTable st;
  CProgram cProgram;

  public PProgram (ExprFactory fac, CProgram cProgram, Map functionDefs)
  {
    this (fac, functionDefs, null);
    // retaining functionDefs for PProgramPredicateManager
    this.cProgram = cProgram;
  }

  public PProgram (ExprFactory _fac, Map _functionDefs, Map _globalDecls)
  {
    this (_fac, _functionDefs, _globalDecls, DDSelector);
  }

  public PProgram (ExprFactory _fac, Map _functionDefs, Map _globalDecls,
                   int _selectorType)
  {
    fac = _fac;
    functionDefs = _functionDefs;
    globalDecls = _globalDecls;
    selectorType = _selectorType;

    labelledStatementMap = new HashMap ();
    statementList = new LinkedList ();
    pTable = new PredicateTable ();

    inconsistent = null;
    // globalDeclsProcessed = false;
    declsRefined = false;

    // set PProgramPredicateManager for this predicate program
    pProgramPredicateManager = new PProgramPredicateManager(this);
  }

  public CProgram getCProgram ()
  {
    return cProgram;
  }

  public void setSymbolTable (SymbolTable st)
  {
    this.st = st;
  }

  public SymbolTable getSymbolTable ()
  {
    return st;
  }

  /*
  public void addFunctionReturn ()
  {
    for (Iterator it = functionDefs.values ().iterator (); it.hasNext (); )
      {
	PFunctionDef def = (PFunctionDef) it.next ();
	if (!(def.getHead ().getLast () instanceof ReturnPStmt))
	  {
	    System.err.println (def.getFunctionName () +
				" has no terminating return, inserting one");
	    def.getHead ().setTail (new ReturnPStmt (def, (Block) null));
	  }
      }
  }
  */

  public Object accept (PProgramVisitor v, Object o)
  {
    return v.visitPProgram (this, o);
  }
  
  /** 
   * @deprecated Use getStatementList ()
   */
  public List getStmtList ()
  {
    return statementList;
  }

  public List getStatementList ()
  {
    return statementList;
  }

  public void setStatementList (List<PStmt> _statementList)
  {
    statementList = _statementList;
  }

  public PStmt getProgramStmts ()
  {
    return programStmts;
  }
  
  public void setProgramStmts (PStmt p)
  {
    programStmts = p;
  }


  public PProgramPredicateManager getPProgramPredicateManager ()
  {
    return pProgramPredicateManager;
  }
  


  /** 
   * @deprecated Use getLabelledStatementsMap ()
   */
  public Map getLabelledStmtsMap ()
  {
    return labelledStatementMap;
  }

  public Map getLabelledStatementMap ()
  {
    return labelledStatementMap;
  }

  public void setLabelMap (Map<String, PStmt> map)
  {
    labelledStatementMap = map;
  }

  /*
    public boolean isGlobalDeclsProcessed ()
    {
    return globalDeclsProcessed;
    }

    public void setGlobalDeclsProcessed (boolean _globalDeclsProcessed)
    {
    globalDeclsProcessed = _globalDeclsProcessed;
    }
  */

  public void setInconsistent (Expr v)
  {
    inconsistent = v;
  }

  public Expr getInconsistent ()
  {
    return inconsistent;
  }

  /** 
   * @deprecated Use getMainFunctionDef ()
   */
  public PStmt getHead ()
  {
    return getMainFunctionDef ().getHead ();
  }

  public Map<String, PFunctionDef> getFunctionDefs ()
  {
    return functionDefs;
  }

  public Map getGlobalDecls ()
  {
    return globalDecls;
  }

  public void setGlobalDecls (Map _globalDecls)
  {
    globalDecls = _globalDecls;
  }

  public PFunctionDef getMainFunctionDef ()
  {
    return (PFunctionDef) functionDefs.get (MAIN_FUNCTION_NAME);
  }

  public void printMainFunction (PrintWriter out)
  {
    PStmt mainFunctionHead = getMainFunctionDef ().getHead ();

    if (mainFunctionHead != null)
      mainFunctionHead.print (out);
  }

  /** 
   * Prints the PProgram to <code>out</code> in a nice format.
   * 
   * @param out - the <code>PrintWriter</code> to print to.
   */
  public void print (PrintWriter out)
  {
    for (Iterator it = statementList.iterator (); it.hasNext ();)
      {
	PStmt s = (PStmt) it.next ();

	// out.print (s.getId () + "/" + s.getLabel () + ": ");

	s.printMe (out);

	out.print ("\tnext: ");
	if (s.getNext () != null)
	  out.print ("" + s.getNext ().getId () + "/" +
		     s.getNext ().getLabel ());
	out.println ();


	out.print ("\tdest: ");
	if (s.getRealDest () != null)
	  out.print ("" + s.getDest ().getId () + "/" +
		       s.getDest ().getLabel ());
	out.println ();

	out.println ();
      }
  }

  public String toString ()
  {
    StringWriter out = new StringWriter ();
    print (new PrintWriter (out, true));
    return out.toString ();
  }

  /** 
   * Returns the PStmt referred to by <code>label</code>
   * 
   * @param label the statement label.
   * @return the PStmt referred to by <code>label</code>; <code>null</code>
   * if the label is unknown.
   */
  public PStmt getStmt (String label)
  {
    System.err.println(">>> trying to get stmt with label " + label);
    PStmt pstmt = (PStmt) labelledStatementMap.get (label);

    if (pstmt != null) return pstmt;

    try
      {
	if (label.startsWith ("l"))
	  label = label.substring (1);

	pstmt = getStmt (Integer.parseInt (label) /*- 1*/);
	return pstmt;
      }
    catch (NumberFormatException ex)
      {
	ex.printStackTrace ();
	throw new RuntimeException ("Unknown label");
      }
  }

  public PStmt getStmt (int id)
  {
    System.err.println(">>> trying to get stmt " + id);
    return (PStmt) statementList.get (id);
  }
  

  public PredicateTable getPredTable ()
  {
    return pTable;
  }

  public static PProgram parse (NullExprAbstractor abstractor, String fileName) 
    throws ParseException, FileNotFoundException
  {
    return parse (abstractor, new File (fileName));
  }

  public static PProgram parse (NullExprAbstractor abstractor, File file)
    throws ParseException, FileNotFoundException
  {
    return parse (abstractor, file, DDSelector);
  }

  public static PProgram parse (NullExprAbstractor abstractor, File file,
                                int selectorType) 
    throws ParseException, FileNotFoundException
  {
    return parse (abstractor, file, selectorType, false);
  }

  public static PProgram parse (NullExprAbstractor abstractor, File file,
                                int selectorType, boolean stmtBlocking)  
    throws ParseException, FileNotFoundException
  {
    return parse (abstractor, new FileReader (file), selectorType,
		  stmtBlocking);
  }

  public static PProgram parse (NullExprAbstractor abstractor, Reader in,
                                int selectorType, boolean stmtBlocking)
    throws ParseException
  {
    try
      {
	Expr programTree = CILTransformerUtil.getProgram
	  (abstractor.getFactory (), in, stmtBlocking);

	return abstractor.doProgramAbstract (programTree, selectorType);
      }
    catch (Exception ex)
      {
	throw new ParseException ("Error parsing", ex);
      }
    
  }


  public static class ParseException extends Exception
  {
    public ParseException (String msg, Exception ex)
    {
      super (msg, ex);
    }
  }

  public ExprFactory getExprFactory ()
  {
    return fac;
  }

  public boolean isDeclsRefined ()
  {
    return declsRefined;
  }

  public void setDeclsRefined (boolean v)
  {
    declsRefined = v;
  }

  public int getSelectorType ()
  {
    return selectorType;
  }

  /** 
   * Class tester method
   * 
   * @param args 
   */
  public static void main (String[] args)
  {
    try
      {
	// -- parse input
	PProgram pProgram =
	  parse (new NullExprAbstractor (new ExprFactoryImpl ()), args [0]);
      }
    catch (Exception ex)
      {
	System.out.println ("Exception: " + ex);
	ex.printStackTrace ();
      }
  }
} 

