package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.util.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.pprogram.*;

import java.io.*;
import java.util.*;


/**
 * A scope-aware predicate manager that stores predicates based on the
 * scope of the variables occuring in them.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class PProgramPredicateManager extends SimplePredicateManager
{

  // -- list of all predicates we know about
  List<Expr> allPredicates;

  // -- map of all predicate managers for each function
  Map<Object,SimplePredicateManager> funcPredManagers;
  
  // -- PProgram we belong to
  PProgram pProgram;
  
  public PProgramPredicateManager (PProgram _pProgram)
  {
    super ();
    allPredicates = new ArrayList<Expr> ();
    pProgram = _pProgram;
    
    funcPredManagers = new HashMap<Object,SimplePredicateManager> ();
    
    // -- initialize by adding a SimplePredicateManager for each function
    // -- the program knows about. Function name is used as a key
    for (Object o : pProgram.getFunctionDefs ().values ())
      funcPredManagers.put (((PFunctionDef)o).getFunctionName (),
			    new SimplePredicateManager ());
  }
  

  
  
  public boolean newPred (Expr predicate)
  {
    if (super.newPred (predicate))
      {
	allPredicates.add (predicate);
	newFuncPred (predicate);
	return true;
      }
    return false;
    
  }


  public void merge()
  {
    super.merge ();
    for (SimplePredicateManager pm : funcPredManagers.values ()) pm.merge ();
  }
  


  /**
   * Returns all old predicates for a function {@code func}.
   *
   * @param func an <code>Object</code> value
   * @return a <code>List</code> value
   */
  public List getFuncPreds (Object func)
  {
    return getFunctionManager (func).getPredicates ();
  }


  /**
   * Returns all new predicates for a function {@code func}.
   *
   * @param func an <code>Object</code> value
   * @return a <code>List</code> value
   */
  public List getFuncNewPreds (Object func)
  {
    return getFunctionManager (func).getNewPredicates ();
  }
 

  public List getNewPredicates ()
  {
    return newPredicateList;
  }
 

  public List getPredicates ()
  {
    return predicateList;
  }

  
  // -- allPreds
  public List getAllPredicates ()
  {
    return allPredicates;
  }
  
  

  /**
   * Adds a new predicate, but only for the function {@code func}
   *
   * @param predicate an <code>Expr</code> value
   */
  protected void newFuncPred (Expr predicate)
  {
    List<String> pScope = getPredicateScope (predicate);

    assert pScope.size () < 2
      : " The predicate " + predicate + " contains variables from "
      + pScope + "!!";
    

    // -- global scope
    if (pScope.isEmpty ())
      addToAll (predicate);
    else
      addToSome (pScope, predicate);
  }
  
  
  /**
   * Adds {@code predicate} to all functions.
   *
   * @param predicate an <code>Expr</code> value
   */
  protected void addToAll (Expr predicate)
  {
    for (SimplePredicateManager pm : funcPredManagers.values ())
      pm.newPred (predicate);
  }
  
  /**
   * Adds {@code predicate} to functions in {@code pScope} list.
   *
   * @param pScope a <code>List</code> value
   * @param predicate an <code>Expr</code> value
   */
  protected void addToSome (List<?> pScope, Expr predicate)
  {
    for (Object o : pScope)
      getFunctionManager (o).newPred (predicate);
  }
  

  /**
   * Get a predicate manager for a function
   *
   * @param func an <code>Object</code> value
   * @return a <code>SimplePredicateManager</code> value
   */
  public SimplePredicateManager getFunctionManager (Object func)
  {
    return funcPredManagers.get (func);
  }
  


  // -- finding a scope of the predicates
  // -- should be moved to some utility class

  /**
   * get the names of all the functions whose local variables are used
   * in the predicate <code>p</code> and put them into a
   * <code>List</code>, e.g.
   * <p>
   * for "foo::x + bar::y > g", return {foo, bar}; 
   * <p>
   * for "foo::@return_value > bar::y", return {bar}, because
   * we treate "@return_value" as global variable
   * <p>
   * @param p an <code>Expr</code> value : a predicate
   * @return a <code>List</code> value
   */
  public List<String> getPredicateScope (Expr p)
  {

    List<String> pScope;
    
    pScope = new LinkedList<String> ();
    
    getPredicateScopeHelper(p, pScope);
    
    return pScope;
  }



  /**
   * Describe <code>getPredicateScopeHelper</code> method here.
   *
   * @param p an <code>Expr</code> value
   * @param funcScope a <code>List</code> value
   */
  private void getPredicateScopeHelper (Expr p, List<String> pScope)
  {
    if (p == null) return;
  
    if (! (p.op () instanceof VariableOp))
      {
	for (int i = 0; i < p.arity (); i++)
	  getPredicateScopeHelper (p.arg (i), pScope);
      }
    else
      {
	String[] token = VariableOp.varName(p).split("::");
	
	if (token.length > 1)
	  {
	    // -- note that we treat *::@return_value as global variable
	    
	    // -- if "@return_value" is not the variable
	    //    then this is a local variable
	    // XXX this is broken now that we use resultDecl
	    if (token[1].indexOf ("@return_value") < 0)
	      {
		if (!pScope.contains (token[0]))
		  pScope.add (token[0]);
	      }
	  }
      }
  }
}

