package edu.toronto.cs.yasm.pprogram;

import edu.toronto.cs.util.*;
import edu.toronto.cs.expr.*;
import edu.toronto.cs.yasm.pprogram.*;

import java.io.*;
import java.util.*;

/**
 * 
 * A {@code PredicateManager} is a class that manages a set of
 * predicates during predicate abstraction. The predicates are
 * partitioned into <b>old</b> and <b>new</b>. New predicates are
 * predicates that have not yet been used during any refinement. Old
 * predicates are predicates that have been used in some refinement
 * already. The class allows to add predicates to the <b>new</b>
 * predicate list, and merge new and old sets.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class SimplePredicateManager
{

  // -- list of predicates we know about
  List<Expr> predicateList;
  // -- list of new predicates being added to us
  List<Expr> newPredicateList;
  
  /**
   * Creates a new <code>SimplePredicateManager</code> instance.
   *
   */
  public SimplePredicateManager ()
  {
    predicateList = new ArrayList<Expr> ();
    newPredicateList = new ArrayList<Expr> ();
  }
  

  /**
   * Adds a new predicate. Returns true if {@code predicate} has been
   * added, and false if it is already contained in the new set.
   *
   * @param predicate a predicate to be added
   */
  public boolean newPred (Expr predicate)
  {
    if (newPredicateList.contains (predicate))
      {
	System.err.println ("Already have predicate: " + predicate);
	return false;
      }

    newPredicateList.add (predicate);
    return true;
  }

  /**
   * Adds a list of new predicates using {@link #addNewPredicate(Expr)}.
   *
   * @param predList a list of predicates to be added
   */
  public void newPred (List<Expr> predList)
  {
    for (Expr e : predList) newPred (e);
  }

  /**
   * Indicates that no new predicates will be added in this round, and
   * that new predicates can be merged with old.
   *
   */
  public void merge ()
  {
    predicateList.addAll (newPredicateList);
    newPredicateList = new ArrayList<Expr> ();
  }

  /**
   * Returns the list of old predicates contained in this manager.
   *
   * @return a <code>List</code> value
   */
  public List<Expr> getPredicates ()
  {
    return predicateList;
  }
  
  /**
   * Returns the list of all new predicates contained in this manager.
   *
   * @return a <code>List</code> value
   */
  public List<Expr> getNewPredicates ()
  {
    return newPredicateList;
  }
    
}
