package edu.toronto.cs.yasm.reachability;

import edu.toronto.cs.mvset.MvSet;
import edu.toronto.cs.mvset.MvSetFactory;
import edu.toronto.cs.mvset.MvRelation;
import edu.toronto.cs.mvset.SeqMvRelation;
import edu.toronto.cs.algebra.BelnapAlgebra;
import edu.toronto.cs.algebra.AlgebraValue;
import edu.toronto.cs.yasm.pcfa.*;

/**
 * Describe class FunctionSummaryRule here.
 *
 *
 * Created: Tue Jun 14 10:22:49 2005
 *
 * @author <a href="mailto:arie@eon.cs">Arie Gurfinkel</a>
 * @version 1.0
 */
public class FunctionSummaryRule extends ReachabilityProofRule 
{

  MvSetFactory mvFactory;
  MvSet stateCube;
  
  /**
   * Creates a new <code>FunctionSummaryRule</code> instance.
   *
   */
  public FunctionSummaryRule (BelnapAlgebra _algebra, 
			      MvSet _stateCube) 
  {
    super (_algebra);
    setStateCube (_stateCube);
  }

  protected void setStateCube (MvSet v)
  {
    stateCube = v;
    mvFactory = stateCube.getFactory ();
  }
  

  public ProofGoal[] apply (ProofGoal _goal)
  {
    // ----------------------------------------------------------------------
    // -- CHECK IF THIS RULE IS APPLICABLE
    // ----------------------------------------------------------------------
    
    // -- applicable only to Edge goals whose edge is 
    // -- a function summary, whose relation is a sequential
    // -- composition of call->body->exit edges
    if (! (_goal instanceof EdgeProofGoal))
      return null;
    
    EdgeProofGoal goal = (EdgeProofGoal) _goal;
    
    if (! (goal.getEdge () instanceof FunctionSummaryEdge))
      return null;

    FunctionSummaryEdge funSummary = (FunctionSummaryEdge) goal.getEdge ();

    // ----------------------------------------------------------------------
    // -- REAL FUNCTION BODY BEGINS HERE
    // ----------------------------------------------------------------------
    
    
    // -- extract the edges
    CallEdge callEdge = funSummary.getCallEdge ();
    SummaryEdge bodyEdge = callEdge.getBodySummary ();
    ReturnEdge exitEdge = callEdge.getReturnEdge ();

    // -- extract the relations involved in function summary
    MvRelation callReln = callEdge.getRelation ();
    MvRelation bodyReln = bodyEdge.getRelation ();
    MvRelation exitReln = exitEdge.getRelation ();
    

    
    
    // -- do the bwd images

    // -- get rid of all caller local varialbes from goal.getDst ()
    MvSet preExit = exitReln.bwdImage (goal.getDst ().getValue ());
    // -- find states at fEntry that reach preExit after the body of
    // -- the function
    MvSet preBody = bodyReln.bwdImage (preExit);
    
    // -- do the fwd images
    
    // -- find all states at fEntry that are reachable from the call site
    // -- under goal.getSrc ().getValue () condition
    MvSet postCall = callReln.fwdImage (goal.getSrc ().getValue ());
    
    // -- now we need to extract our states
    // XXX postCall variables should be used here
    // -- find a *state* at the intersection of postCall and preBody
    MvSet postCallPoint = 
      (MvSet) postCall.and (preBody).
      mintermIterator (mvFactory.top (), goal.getValue ()).next ();
    
    // -- find the values
    // -- find the value of transition between call site and fEntry
    AlgebraValue postCallValue = 
      postCall.and (postCallPoint).existAbstract (stateCube).getValue ();
    
    // -- find the starting value at which we get to fEntry
    AlgebraValue preBodyValue = 
      preBody.and (postCallPoint).existAbstract (stateCube).getValue ();
    
    
    // ----------------------------------------------------------------------
    // -- BIG HACK STARTS HERE
    // ----------------------------------------------------------------------

    // -- if the preBodyValue is true, we will not need to refine the
    // -- body of the function. Note that return edge is never refined
    // -- either. Thus, we can bail out from here adding the call edge
    // -- as the next proof goal
    if (preBodyValue.isTop ())
      return oneGoal 
	(new EdgeProofGoal (goal, goal.getSrc (),
			    new Node (callEdge.getDst (), postCallPoint),
			    callEdge, postCallValue));
				       
    
    // ----------------------------------------------------------------------
    // -- BIG HACK ENDS HERE
    // ----------------------------------------------------------------------
    

    // -- find all postBody states find all states at fExit that are
    // -- reachable from fEntry under postCallPoint condition
    MvSet postBody = bodyReln.fwdImage (postCallPoint);
    
    // -- find postBody point
    // -- find a *state* at the intersection of postBody and preExit
    MvSet postBodyPoint = 
      (MvSet) postBody.and (preExit).
      mintermIterator (mvFactory.top (), preBodyValue).next ();
    
    // -- get values
    // -- find a value of the transition between fEntry and fExit
    AlgebraValue postBodyValue = 
      postBody.and (postBodyPoint).existAbstract (stateCube).getValue ();
    // -- fina a value fo the transition between fExit and call site
    AlgebraValue preExitValue = 
      preExit.and (postBodyPoint).existAbstract (stateCube).getValue ();
    

    
    // -- first the exitEdge, then bodyEdge, then callEdge
    return threeGoal
      (new EdgeProofGoal (goal, new Node (exitEdge.getSrc (), postBodyPoint),
			  goal.getDst (), exitEdge, preExitValue),
       new EdgeProofGoal (goal, new Node (bodyEdge.getSrc (), postCallPoint),
			  new Node (bodyEdge.getDst (), postBodyPoint),
			  bodyEdge, postBodyValue),
       new EdgeProofGoal (goal, goal.getSrc (),
			  new Node (callEdge.getDst (),
				    postCallPoint),
			  callEdge, postCallValue));
			  


					      
    
    
  }
  
  // ----------------------------------------------------------------------
  // -- OLD CODE
  // ----------------------------------------------------------------------

  /*
  public ProofGoal[] __apply (ProofGoal _goal)
  {
    // -- applicable only to Edge goals whose edge is 
    // -- a function summary, whose relation is a sequential
    // -- composition of call->body->exit edges
    if (! (_goal instanceof EdgeProofGoal))
      return null;
    
    EdgeProofGoal goal = (EdgeProofGoal) _goal;
    
    if (! (goal.getEdge () instanceof FunctionSummaryEdge))
      return null;

    FunctionSummaryEdge funSummary = (FunctionSummaryEdge) goal.getEdge ();
    
    if (! (funSummary.getRelation () instanceof SeqMvRelation))
      return null;
    
    SeqMvRelation funSumReln = (SeqMvRelation) funSummary.getRelation ();
    
    // -- now we break it into 3 relations
    MvRelation callReln = funSumReln.getFirst ();
    MvRelation bodyReln = 
      ((SeqMvRelation) funSumReln.getSecond ()).getFirst ();
    MvRelation exitReln = 
      ((SeqMvRelation) funSumReln.getSecond ()).getSecond ();
    
    
    // -- do the bwd images
    MvSet preExit = exitReln.bwdImage (goal.getDst ().getValue ());
    MvSet preBody = bodyReln.bwdImage (preExit);
    
    // -- do the fwd images
    MvSet postCall = callReln.fwdImage (goal.getSrc ().getValue ());
    
    // -- now we need to extract our states
    MvSet postCallPoint = 
      (MvSet) postCall.and (preBody).
      mintermIterator (mvFactory.top (), goal.getValue ()).next ();
    
    // -- find the values
    AlgebraValue postCallValue = 
      postCall.and (postCallPoint).existAbstract (stateCube).getValue ();
    
    AlgebraValue preBodyValue = 
      preBody.and (postCallPoint).existAbstract (stateCube).getValue ();
    
    // -- find all postBody states
    MvSet postBody = bodyReln.fwdImage (postCallPoint);
    
    // -- find postBody point
    MvSet postBodyPoint = 
      (MvSet) postBody.and (preExit).
      mintermIterator (mvFactory.top (), preBodyValue).next ();
    
    // -- get values
    AlgebraValue postBodyValue = 
      postBody.and (postBodyPoint).existAbstract (stateCube).getValue ();
    AlgebraValue preExitValue = 
      preExit.and (postBodyPoint).existAbstract (stateCube).getValue ();
    

    // -- we are ready to create our goals
    // -- extract the edges
    CallEdge callEdge = funSummary.getCallEdge ();
    SummaryEdge bodyEdge = callEdge.getBodySummary ();
    ReturnEdge exitEdge = callEdge.getReturnEdge ();
    
    // -- first the exitEdge, then bodyEdge, then callEdge
    return threeGoal
      (new EdgeProofGoal (goal, new Node (exitEdge.getSrc (), postBodyPoint),
			  goal.getDst (), exitEdge, preExitValue),
       new EdgeProofGoal (goal, new Node (bodyEdge.getSrc (), postCallPoint),
			  new Node (bodyEdge.getDst (), postBodyPoint),
			  bodyEdge, postBodyValue),
       new EdgeProofGoal (goal, goal.getSrc (),
			  new Node (callEdge.getDst (),
				    postCallPoint),
			  callEdge, postCallValue));
			  


					      
    
    
  }
  */  

}
