package edu.toronto.cs.yasm.wp;

import java.util.*;

import edu.toronto.cs.expr.*;
import edu.toronto.cs.tp.SymbolTable;

import static edu.toronto.cs.expr.BoolOp.*;
import static edu.toronto.cs.expr.ComparisonOp.*;
import static edu.toronto.cs.expr.CILIndirectionOp.*;
import static edu.toronto.cs.expr.CILAssignOp.*;
import static edu.toronto.cs.expr.VariableOp.*;
import static edu.toronto.cs.expr.ChoiceOp.*;
import static edu.toronto.cs.expr.SimpleTypeOp.*;
import static edu.toronto.cs.expr.CompoundTypeOp.*;
import static edu.toronto.cs.expr.ArrayOp.*;

import static edu.toronto.cs.expr.CVarDeclOp.*;

/** 
 * A weakest precondition computer based on Morris' general axiom of
 * assignment for conditional variable substitution as presented in a
 * SLAM paper.
 * <p>
 * However, it seems that, according to that axiom, wp (p := e, *p =
 * 0) is (&amp; p = p AND e = 0) OR (&amp; p != p AND *p = 0), which
 * reduces to (p = 0), but the correct wp is *e = 0. There are kludges
 * in the implementation to handle this particular case.
 * 
 * @author Kelvin Ku (kelvin@cs.toronto.edu)
 * @version 
 */
public class MorrisWPComputer extends ExprWPComputer implements WPComputer
{
  // this is in super's varMap, but I'm not supposed to know about that!
  Expr stmt;

  // need a general subst map for pointer expressions (not just from var ->
  // var, but from Expr -> Expr); also, substitution is based on predicates
  // here, not on the statement, so this map needs to be re-initialized for
  // each predicate
  Map<Expr, Expr> substMap;

  SymbolTable st;

  TypeInferencer typeInf;

  SimplifyingRewriter simp;

  public MorrisWPComputer (Expr stmt, SymbolTable st)
  {
    super (stmt);
    this.stmt = stmt;
    this.st = st;
    substMap = new IdentityHashMap<Expr, Expr> ();
    typeInf = new TypeInferencerCached (st);
    simp = new SimplifyingRewriter ();
  }

  /** 
   * Tells if it is possible for two expressions to refer to the same
   * word of memory.
   *
   * @param x
   * @param loc
   */
  private boolean mayAlias (Expr x, Expr loc)
  {
    // Pairs represent (x, loc)
    // int x, y; int [] A, B; int *p, *q;

    if (typeOf (x) != typeOf (loc))
      return false;

    // Consult CIL alias database that we have to build....
    
    // (x, y) ["x" = "y"]
    if ((isVar (loc) || isCVarDecl (loc)) && x == loc)
      return true;

    // (A[i], B[j]) ["A" = "B"]
    if (isArraySel (loc) && isArraySel (x)
        && (getArray(loc) == getArray (x)))
      return true;

    // (A[i], *q) [types match]
    if (isDeref (loc) && isArraySel (x))
      return true;

    // (*p, B[j]) [types match]
    if (isDeref (x) && isArraySel (loc))
      return true;

    // (*p, *q) [types match]
    if (isDeref (loc) && isDeref (x))
      return true;
      
    // (x, *q) [types match]
    if (isDeref (loc) && isVar (x))
      return true;

    // (*p, y) [types match]
    if (isDeref (x) && isVar (loc))
      return true;

    // Really *should* be true by default, with above cases 
    return false;
  }

  public Expr computeWP (Expr pred)
  {
    System.out.println ("\n\n*** wp(" + stmt + "," + pred + ")");
    //System.out.print(pred + ": ");

    // successively apply conditional subst for each loc in pred
    Expr posWP = pred;
    Expr negWP = not (pred);

    // First discharge the array business. They aren't aliased, but 
    // their elements may be aliased through pointers.
    //
    // Mapping to update an array after A[x] := e is (A, A[x |-> e])
    //
    // XXX This may be redundant. Keep it in for now.
    if (stmt.arity () > 0 && ArrayOp.isArraySel (stmt.arg (0))) {
      substMap.clear ();
      substMap.put (ArrayOp.getArray(stmt.arg(0)), 
                    ArrayOp.upd(ArrayOp.getArray(stmt.arg(0)), 
                                ArrayOp.getIndex(stmt.arg(0)),
                                stmt.arg(1)));
      posWP = posWP.guardedSubst (substMap, super.addrGuard);
      negWP = negWP.guardedSubst (substMap, super.addrGuard);
    }

    Set<Expr> locs = getMutables (pred, new HashSet<Expr> ());
    System.out.println("   Locs (" + pred + "): " + locs);
    for (Expr loc : locs)
    {
      substMap.clear ();

      if (mayAlias (stmt.arg (0), loc)) {
        System.out.println (stmt.arg (0) + " may alias " + loc);
        substMap.put (loc, stmt.arg (1));

        posWP = condSubst (posWP, loc);
        negWP = condSubst (negWP, loc);
      }
      else
        System.out.println (stmt.arg (0) + " cannot alias " + loc);
    }
    
    Expr result = lchoice (posWP, negWP);

    System.out.println ("*** = " + result);

    return simp.rewrite (result);
  }

  /** 
   * Returns a set of all the mutables in expression {@code e}.
   * <p>
   * We found the "location" terminology from the SLAM paper 
   * to be confusing; hence, we speak of mutables. Basically, 
   * a mutable is an identifier in an expression which refers to 
   * a word of memory whose contents can change; this can be:
   *   - a scalar variable (including pointers)
   *   - an element of an array
   *   - a dereference of a pointer expression
   * 
   * @param e 
   * @return 
   */
  private Set<Expr> getMutables (Expr e, Set<Expr> l)
  {

    if ((isVar (e) && !isArrayT (typeOf (e)))
        || isCVarDecl (e))
      l.add (e);
    else if (isAddrOf (e) && isArraySel(e.arg (0)))
      // Look for mutables in the array's index
      getMutables (e.arg (0).arg (1), l);
    else if (isDeref (e)) {
      l.add (e);
      getMutables (e.arg (0), l);
    } else if (isArraySel (e)) {
      l.add (e);
      getMutables (e.arg (1), l);
    } else if (e.arity () > 0)
      for (Object o : e.args ())
        getMutables ((Expr) o, l);
    return l;


    /*    if ((isVar (e) && ! isArrayT (typeOf (e)))
        || isDeref (e) || isArraySel (e))
      l.add (e);

    // get pointer arithmetic expressions like (p + 1)
    if (e.op () instanceof NumericOp && isPtrT (typeOf (e)))
      l.add (e);

    if (e.arity () > 0)
      for (Object o : e.args ())
        getMutables ((Expr) o, l);

        return l;*/
  }

  private Expr typeOf (Expr e)
  {
    return typeInf.getType (e);
  }

  /** 
   * Returns the conditional substitution of location {@code loc} in {@code
   * pred} with the RHS of this {@code WPComputer}'s assignment.
   * <p>
   * So, where this {@code WPComputer}'s is {@code x = e} and {@code loc = y},
   * {@code wp (pred, loc) = (&amp;x = &amp;y AND pred[e/y]) OR (&amp;x !=
   * &amp;y AND pred)}.
   * 
   * @param pred 
   * @param loc 
   * @return 
   */
  private Expr condSubst (Expr pred, Expr loc)
  {
    return
      lor (
        land (
          lequal (
            addrOf (stmt.arg (0)),
            addrOf (loc)),
          pred.subst (substMap)),
        land (
          lnot (
            lequal (
              addrOf (stmt.arg (0)),
              addrOf (loc))),
          pred));
  }

  /** 
   * Returns the location affected by an assignment to {@code e}. 
   * <p>
   * If {@code e} is a variable {@literal i}, returns {@literal &amp;i}. If
   * {@code e} is a dereference @{literal *p}, returns {@literal p}.
   * <p>
   * This could go in CILIndirectionOp#laddrOf, but that might cause confusion
   * since one may expect "the address of *p" to be {@literal &amp;*p}, whereas
   * this returns "the address affected by an assignment to *p", which is
   * simply "p".
   * <p>
   * But, where {@literal int *p, affLoc (p) = p}, not {@literal &amp;p}.
   *
   * XXX THIS IS CONFUSING AND INCONSISTENT
   * 
   * @param e 
   * @return 
   */
  private Expr affLoc (Expr e)
  {
    if (isDeref (e))
      return e.arg (0);
    else if (isVar (e) && isPtrT (typeOf (e)))
      return e;
    else
      return addrOf (e);
  }

  // XXX move this to a junit test when there's a convenient way to check
  // result (assertEqual) of computeWP; currently this involves constructing an
  // Expr equivalent to the result, which is tedious
  public static void testBasic ()
  {
    ExprFactory fac = new ExprFactoryImpl ();

    SymbolTable st = new SymbolTable ();

    Expr intT = fac.op (SimpleTypeOp.INT);
    Expr x = fac.var ("x");
    Expr y = fac.var ("y");
    Expr p = fac.var ("p");
    Expr q = fac.var ("q");
    Expr A = fac.var ("A");
    st.put (x, intT); // int x
    st.put (y, intT); // int y
    st.put (p, ptrT (intT)); // int *p
    st.put (q, ptrT (intT)); // int *q
    st.put (A, arrayT (intT, intT)); // int A[];

    System.out.println (st);

    // x := 5
    Expr x_gets_5 = assign (x, fac.intExpr (5));

    // x > 0
    Expr x_gt_0 = gt (x, fac.intExpr (0));

    // y > 3
    Expr y_gt_3 = gt (y, fac.intExpr (3));

    Expr starp = deref (p);
    Expr starq = deref (q);

    // *p > 3
    Expr ptrPred = gt (starp, fac.intExpr (3));

    // *p > *q
    Expr ptrPred2 = gt (starp, starq);

    Expr starPEqZero = eq (starp, fac.intExpr (0));

    // *p := 5
    Expr starp_gets_5 = assign (starp, fac.intExpr (5));

    // A[3] := 2
    Expr A3_gets_2 = assign (sel (A, 3), fac.intExpr (2));

    // x := 1
    Expr x_gets_1 = assign (x, 1);

    // A[x] == 0
    Expr Ax_eq_0 = eq (sel (A, x), 0);

    // A[x] == x
    Expr Ax_eq_x = eq (sel (A, x), x);

    // p = p + 1
    Expr pp1 = assign (p, NumericOp.plus (p, 1));

    // p = &A[x]
    Expr p_eq_addrAx = eq (p, addrOf (sel (A, x)));

    // (p+1) = &A[x]
    Expr pp1_eq_addrAx = eq (NumericOp.plus (p, 1), addrOf (sel (A, x)));

    // &A[x] = &A[y]
    Expr addrAx_eq_addrAy = eq (addrOf (sel (A, x)), addrOf (sel (A, y)));

    // &A[A[x]] == 0
    Expr addr_AAx_eq_0 = eq (addrOf (sel (A, sel (A, x))), 0);

    // WP (x := 5) for x > 0, y > 3, *p > 3, *p > *q
    MorrisWPComputer wpc = new MorrisWPComputer (x_gets_5, st);
    System.out.println ("\n" + x_gets_5 + ":");
    System.out.println (wpc.computeWP (x_gt_0));
    System.out.println (wpc.computeWP (y_gt_3));
    System.out.println (wpc.computeWP (ptrPred));
    System.out.println (wpc.computeWP (ptrPred2));

    // WP (*p := 5) for y > 3
    wpc = new MorrisWPComputer (starp_gets_5, st);
    System.out.println ("\n" + starp_gets_5 + ":");
    System.out.println (wpc.computeWP (y_gt_3));

    // WP (A[3] := 2) for x > 0, *p > 3
    wpc = new MorrisWPComputer (A3_gets_2, st);
    System.out.println ("\n" + A3_gets_2 + ":");
    System.out.println (wpc.computeWP (Ax_eq_0));
    System.out.println (wpc.computeWP (x_gt_0));
    System.out.println (wpc.computeWP (ptrPred));
    System.out.println (wpc.computeWP (p_eq_addrAx));
    System.out.println (wpc.computeWP (addr_AAx_eq_0));

    // WP (x := 1) for A[x] = 0, A[x] = x, p = &A[x]
    wpc = new MorrisWPComputer (x_gets_1, st);
    System.out.println ("\n" + x_gets_1 + ":");
    System.out.println (wpc.computeWP (Ax_eq_0));
    System.out.println (wpc.computeWP (Ax_eq_x));
    // We're getting wp(x :=1, p = &A[x]) = (p = &A[x])
    // This is wrong. addrGuard is stopping us from making the needed 
    // substitution.
    System.out.println (wpc.computeWP (p_eq_addrAx));
    // addrGuard again
    System.out.println (wpc.computeWP (addrAx_eq_addrAy));

    // WP (p++) for p = &A[x], *p > 3, *p > *q
    wpc = new MorrisWPComputer (pp1, st);
    System.out.println ("\n" + pp1 + ":");
    System.out.println (wpc.computeWP (p_eq_addrAx));
    System.out.println (wpc.computeWP (pp1_eq_addrAx));
    System.out.println (wpc.computeWP (ptrPred));
    System.out.println (wpc.computeWP (ptrPred2));

    System.out.println ("=====");
    wpc = new MorrisWPComputer (A3_gets_2, st);
    System.out.println (wpc.computeWP (starPEqZero));
  }

  public static void main (String[] args)
  {
    testBasic ();
  }
}

