package edu.toronto.cs.yasm.wp;

import edu.toronto.cs.expr.*;

import edu.toronto.cs.tp.SymbolTable;

import java.util.Iterator;

/**
 * A factory for creating instances of {@code WPComputer} based on the
 * form of a statement.
 *
 * @author <a href="mailto:arie@kanga.fm.sandbox">Arie Gurfinkel</a>
 * @version 1.0
 */
public class WPComputerFactory
{
  /**
   * Returns a {@code WPComputer} for a statement expressed by {@code e}
   *
   * @param e an <code>Expr</code> value
   * @return a <code>WPComputer</code> value
   */
  public static WPComputer wp (Expr e, SymbolTable st)
  {
    if (e.op () == CILAssignOp.ASSIGN
	|| CStmtOp.isModifyExpr (e))
    {
      if (recNonNull (e)) {
        return new MorrisWPComputer (e, st);
      }

      return unknown (e);  
    }

    if (e.op () == CILListOp.SLIST)
      return new DCWPComputer (e, st);

    throw new
      RuntimeException ("Don't know how to make WP computer for: " + e);  
  }


  /**
   * Creates a {@code WPComputer} for a parallel assignment
   *
   * @param lhs an <code>Expr</code> value
   * @param rhs an <code>Expr</code> value
   * @return a <code>WPComputer</code> value
   */
  public static WPComputer prllAsmt (Expr lhs, Expr rhs)
  {
    return new ExprWPComputer (lhs, rhs);
  }

  /**
   * Creates a {@code WPComputer} for a statement that is known to
   * modify variable {@code e}.
   *
   * @param e an <code>Expr</code> value
   * @return a <code>WPComputer</code> value
   */
  public static WPComputer unknown (Expr e)
  {
    return new UnknownExprWPComputer (e);
  }

  /**
   * Returns a {@code WPComputer} over an statement that does not
   * modify any variables.
   *
   * @return a <code>WPComputer</code> value
   */
  public static WPComputer id ()
  {
    return ExprWPComputer.ID;
  }

  // -- returns true if all children are non-null, false otherwise
  // -- XXX replace with recKnownOp (Expr e) when null ops are eliminated
  private static boolean recNonNull (Expr e)
  {
    if (e == null) return false;

    for (Iterator it = e.args ().iterator (); it.hasNext ();)
    {
      if (!recNonNull ((Expr) it.next ())) return false;
    }

    return true;
  }

}
